# ==============================================================================
# Example script
# ==============================================================================

import sys
import os

# Uncomment if you call this script without X server but want to generate plots
#import matplotlib
#matplotlib.use('Agg')     # Set backend to AGG (Anti-Grain Geometry engine)
#matplotlib.pyplot.ioff()  # turn off matplotlib's interactive mode


# --- Setting paths

# Note: THIS script should not be in the code_TM_app dir
#       to keep your 'analysis' dir separate from the actual code

# Give the location of the "TM app" (as long it's not installed as a package)
sys.path.insert(0, os.path.abspath('../code_TM_app')) # Here: in sub-dir of *parent* dir
# Now the *absolute* path of the "TM app" code is in sys.path[0]
# (also necessary for calling external bash scripts)

from TM.main import *                # Import all
from TM.setupclass import config     # Import (the already instanced) config class instance
from TM.eventops import *
from TM.io import *
from TM.plotting import *
from TM.utils import *


# --- General setup

# -- TM-scan-related paths

# Location of the sequence-related io data (input: catalog *.csv, output: catalog waveforms)
config.dataDir = './data/'          # in subdir
#config.dataDir = '../data/'          # in parent's subdir

# Use FDSN as data source?
#config.FDSN = True
#config.FDSN_server = 'http://arclink.ethz.ch'
#config.FDSN_server = 'BGR'

# Location of the MSEED archive (only if FDSN not used)
config.mseedDir = '/mnt/SED-miniseed/'

# Perhaps you want to save all the detections (mseeds) somewhere else?
# Per default, it's './' and thus saves detections INTO config.directory;
# but you can dump them on another disk:
# config.TMdata = os.path.expanduser("/media/{USER}/{DISKNAME}/TMdata/")  # Dump directory for detections
# (a subfolder named just as 'config.directory' will be created therein)

# -- Network / Station related parameters
config.network = 'CH'

config.region =  'Chamonix'
config.station = 'SALAN';
config.channels = ["EHZ", "EHN", "EHE"];
config.timeTravel = 2.7
config.duration = 7.0

# -- Template parameters
config.templateShiftCorr = False    # Whether to shift templates to align them to the first (maxCC)
config.timeTol = 0.5                # (post-processing) Used to unify detections of templates

# -- Scanning parameters
config.filtConf = [5, 30, 4]; config.stop50Hz = False
config.filterBuff = 1.0             # To avoid filter response in signal:
                                    #    Time to prepend and append before filtering waveform
config.threshold = .40              # The general threshold (as fraction of 1)
config.threshCred = .50             # The credible (more conservative) threshold
config.threshPerChannel = False     # Apply thresh to EACH channel (and not MEAN among channels)

# -- Timing
config.startdate = UTCDateTime("2015-01-01")
#config.enddate = UTCDateTime("2016-01-08")
config.enddate = UTCDateTime(datetime.datetime.now().date()) # Until midnight of last night
#config.enddate = UTCDateTime(datetime.datetime.now().date() + datetime.timedelta(days=1)) # Until next midnight

# -- Resampling
# Get samplerate of original data
#fs = get_waveform(config.network, config.station, config.channels[0],
#                  config.startdate, 1, procType='raw').stats.sampling_rate
from obspy.clients.fdsn import Client
clientFDSN = Client("http://arclink.ethz.ch")
fs = clientFDSN.get_stations(network=config.network, station=config.station,
                             channel=config.channels[0], starttime=config.startdate,
                             level="channel")[0][0][0].sample_rate
#fs = 1000
# Best way to find sufficient sampling rate: fs_new >= 4 * upper BP limit
#   --> still ~half of frequency band until nyquist (not quite half due to anti-alias)
factor = int(fs/config.filtConf[1]/4)  # (has an effect if fs > 8* upper BP limit)
#factor = 1  # Don't do any resampling
config.resampling = fs/max(factor, 1)  # 'max(x, 1)' to avoid zero div

# -- Determine fastest xCorr approach (in time- or frequency-domain)
select_fastest_approach()           # Will benchmark for 2x1 minute
#config.conv_method = 'fftconvolve' # Choose method right away ['convolve', 'fftconvolve']

# -- Misc
config.parallelFilter = False       # (process-based-parallel can take up huge amounts of memory)
config.use_threads = True           # Use threads for certain tasks (e.g., read, calc_std_window)

# Immitate MPI-related config parameters (to avoid 'Value not set' error when NOT using MPI)
config.chunk_len = 0
config.overlap_len = 0
config.chunk_on_disk = 0
config.num_readers = 0

# Define more threshold values
threshs = sorted(set([config.threshold, config.threshCred] +
                     [0.45, 0.5, 0.55, 0.6, 0.8]))

# -- Set overall directory for saving
config.set_dir('results', '', openEnd=True, moreInfo=False)
#config.directory = "results_%s_%s_%s_%s" % (config.region, config.station,
#                                            config.startdate.date, 'openEnd')

# -- Treat template selection ( = making a templateSet)

# * Auto mode
templateSet = []            # Start with empty templateSet
config.templateMagRange = (-2.0, 1.2)  # Magnitude range of events to use for scanning
config.firstTemplateMinMag = 0.5       # Minimum magnitude of first template event
config.threshNewTemplate = .75         # Catalog events that were detected with CC smaller than
                                       #    this will become a new template

#exit()  # Maybe you want to stop here first?

# Run SQL script, and delete the events not needed in the csv
#  (i.e. events before startdate, or AUTOMATIC locations/magnitudes)
#getEventsFromSC3DB()
catalog = readCatalogSEDDB()
# Pre-load and plot
preload_events(catalog, readALLagain=False)
# NOW CHECK IF THERE ARE DOUBLE EVENTS (o qualcosa) ... and exclude them!
# e.g., with the following
#times = [
#    UTCDateTime("2016-09-11T12:19:06.531155"), #double event
#    ]
#for i in range(len(catalog)-1, -1, -1):
#    if (catalog[i][1]) in times:
#       del catalog[i]


# --- Do the Template Matching

# Do the actual TM
foundEvents_full, templateSet = template_matching(templateSet, catalog)

# Reload again
#foundEvents_full = readDetections("progress", config.threshold)
#templateSet = readTemplateSet()

# Write uniqued detections to file
foundEvents = uniqueEventList(foundEvents_full)
fileBase = "detections_%dtps_unique" % len(templateSet)
writeDetections(foundEvents, file=fileBase)


# Make histogram of templates' detection numbers
plot_hist_template_distribution(foundEvents, foundEvents_full)


# --- Determine Magnitudes

# Update the events' amplitude (highpass filtering)
freqMax = 20  # (to remove too-high frequencies that disturb the amplitude comparison)
foundEvents = getAmplitude(foundEvents, "det", freqMax, fullDetStream=True)
# ... and write out w/ updated amplitude (overwrites the old one)
writeDetections(foundEvents, file=fileBase)

# Derive catalog events' amplitude
catalog_amp = getAmplitude(catalog, "catalog", freqMax)

# Determine the events' magnitude using the events' magnitude
magRange = (1.0, 3.5)  # Specify magnitude range of catalog events used for regression
foundEvents = regressMagnitude(foundEvents, catalog_amp, magRange)

# Write detections with magnitudes
for iTresh in threshs:
    writeDetections(foundEvents, iTresh, fileBase)


# --- Plotting: Timeline(s)

for iTresh in threshs:
    plotTimeline(fileBase, iTresh)


# --- Plotting: Waveforms

# Plot each detection seperately
plotWaveforms(foundEvents, 'detections')

# Helicorder plots - grouped by template
#plotHelicorder(foundEvents, config.channels[0], config.threshold, 'raw')
plotHelicorder(foundEvents, config.channels[0], config.threshold, 'bandpass')
plotHelicorder(foundEvents, config.channels[1], config.threshold, 'bandpass')
plotHelicorder(foundEvents, config.channels[2], config.threshold, 'bandpass')
