# -*- coding: utf-8 -*-
"""
Submodule containing the core of template matching / xcorr.

(created) 2016, Marcus Herrmann, Swiss Seismological Service, ETH Zürich
"""

# --- Import modules

from multiprocessing.pool import ThreadPool

# Scientific
import numpy as np
import scipy
import scipy.signal  # To avoid scipy library module api problems
import scipy.ndimage

# ObsPy
from obspy.core import Trace, Stream

# Own modules
from TM.setupclass import config
from TM.waveformops import _process_mask


def calc_windowed_std_trace(trace):
    """
    Calculates a moving std window of a trace.
    Required normalizing cross-correlation-traces.

    This operation is independent on the number of templates
        --> has to be done only once per waveform data element (e.g., a day)

    This function will also make stdWindowed as a carrier for the final mask (as MaskedArray).
    But only in case of present gaps or bad input data --> numerical 'issues' that
    need to be masked and appropriately handled (e.g. extending mask at gaps, etc.)

    FYI: Using a MaskedArray directly in numerical operations is generally slower than
         first separating mask and data, then process the underlying data, and finally
         apply mask later again. (as long as we know which numerical exceptions we can
         expect and treat them accordingly)
    """

    # Calculate window length from template duration
    lenT = int(config.duration * trace.stats.sampling_rate + 1)  # [in samples]

    # To be on the safe side (i.e. reducing numerical errors):
    # Retrieve array masks and overwrite related masked data (=gaps) with zeros
    if np.ma.isMaskedArray(trace.data):
        # Overwrite UNDERYLING data (.data) to maintain array mask
        trace.data.data[trace.data.mask] = 0

    # Create a slice that represents the 'valid' output range of a 'same'-sliced output
    # i.e. disragarding the border area where the mean cannot calculated with a window)
    sl_valid = slice(lenT//2, -lenT//2+1)

    # Compute std for each xCorr sample using an uniform (mean) filter/window/sliding average
    #   (it's using the "stride trick")
    #   Note 1: std = sqrt(mean((x - mean(x))^2)) = ... = sqrt(mean(x^2) - mean(x)^2)
    #       Now, operator E[X] = mu is average/expected value ("mean()"), X = data
    #          -> std^2 = E[(X - E[X])^2]) = ... = E[X^2] - E[X]^2
    #       Here, E[X^2] = E1; E[X]^2 = E2;
    #   Note 2: for the boundaries, it will reflect the data (default)
    #   Note 3: mind the masked data when trace is a MaskedArray
    E1 = scipy.ndimage.filters.uniform_filter1d(trace.data**2, lenT)[sl_valid]
    E2 = scipy.ndimage.filters.uniform_filter1d(trace.data,    lenT)[sl_valid]

    # Checking if there are negative or extrem low values in E1 (which shouldn't be)
    # The following check MIGHT be an indication that uniform_filter1d()
    # didn't produce a proper windowed mean.
    # These are due to strange values in the waveform data (impulses, gaps, ...)
    #    & uniform_filter1d() can apparently NOT deal with them correctly :-(
    # (I guess this is less the case if the waveform was already filtered --> no "impulses"
    #    & the gaps were filled with zeroes. But just to make sure...
    if (E1 <= np.finfo(E1.dtype).eps).any():

        # Commented out printing statement since it will crash the multiprocessing
#        print("Channel %s has strange values (large jumps/impulses?). "
#              "Processing delayed." % trace.stats.channel, flush=True)

        # Ensure proper calculation of windowed mean

        # Method 1: proper but slowest
        #       (~450x slower for 1day 125Hz data than uni_filt1d())
#        E1 = scipy.ndimage.generic_filter(trace.data**2, np.mean, size=lenT)

        # Method 2: like method 1, but differs at borders (lenT//2) --> 'valid' mode
        #       (only ~5x slower)
        E1 = np.convolve(trace.data**2, np.full(lenT, 1.0/lenT), mode='valid')
#        E1 = scipy.signal.convolve(trace.data**2, np.full(lenT, 1.0/lenT), mode='valid')
        # FYI: NEVER use fftconvolve() here! The data is "not suited" for FFT-transformation

        # Method 3: like method 1, but shifted by minus one sample (if lenT is even)
        #       (~15x slower)
#        E1 = scipy.ndimage.filters.convolve(trace.data**2, np.full(lenT, 1.0/lenT))

        # Method 4:  [related to Lewis (1995)]
        # (not faster than uniform_filter1d(); slightly slower [~1.2x])
        # Also: less accurate for bad data (impulses) or veeeery long timeseries (>10mio samples)
        #       (b/c the growing values cannot be numerically well represented anymore)
        #       Can be circumvented by casting trace.data to np.float64
#        cumsum = np.cumsum(np.insert(np.float64(trace.data)**2, 0, 0))
#        E1 = (cumsum[lenT:] - cumsum[:-lenT]) / lenT  # == 'valid' range
##        E1 = np.zeros(len(trace.data))
##        E1[lenT//2:-lenT//2+1] = (cumsum[lenT:] - cumsum[:-lenT]) / lenT  # == 'same' range
#        del cumsum

        # Mask extremely low values or zeros (to avoid quasi zero-division)
        #   (They could eventually lead to CCC > 1.0)
        E1 = np.ma.masked_array(E1, mask=(E1 <= np.finfo(E1.dtype).eps))  # (does not copy)

        # (Total) alternative for stdWindowed:
        #       (but a looooot (x1600) slower)
        # stdWindowed = scipy.ndimage.generic_filter(trace1.data, np.std, size=lenT)

        # (Unchecked: cv2's boxFilter() ... http://stackoverflow.com/a/36266187
        #     (maybe even faster than uniform_filter1d() approach BUT only python2 yet)
        # How about accuracy for bad data?

    # Apply array mask from trace data to avoid numpy warnings (exluding "bad values" i.e. zeros)
    if np.ma.isMaskedArray(trace.data):
        # ... using an extend mask (at gaps)
        mask_ext = _process_mask(trace.data.mask.copy(), 1.0, lenT//2+1)[sl_valid]
        # Note: if E1 is already masked, apply this mask on top
        E1 = np.ma.masked_array(E1, mask=mask_ext)  # (does not copy)

    # FYI: E1 might now be a MaskedArray due to gaps AND/OR impulses

    # Actual calculation of windowed std
    if not np.ma.isMaskedArray(E1):
        stdWindowed = (E1 - E2**2)**.5
    else:
        # (is ~7x faster than above if E1 is MaskedArray)
        # Do a non-masked operation, but assert non-zero values in sqrt
        stdWindowed = np.maximum(np.asarray(E1) - E2**2, 0)**.5
        # Apply previous mask again
        stdWindowed = np.ma.masked_array(stdWindowed, mask=E1.mask)
        # and if stdWindow is quasi zero e.g., due (E1 - E2**2) < 0
        stdWindowed = np.ma.masked_array(stdWindowed, mask=stdWindowed <= np.finfo(E1.dtype).eps)

    # FYI: stdWindowed might now be a MaskedArray if E1 is masked AND/OR if (E1 - E2**2) < 0

    if np.ma.isMaskedArray(stdWindowed):
        # Overwrite masked values (underlying data) with '1'
        # to avoid later numpy warnings when norming xCorrTrace
        stdWindowed.data[stdWindowed.mask] = 1  # Leave as MaskedArray

    return stdWindowed


def do_xcorr_onechannel(trace1, trace2, stdWindowed=None):
    """
    xCorr parallelization level 3:
    Performs cross-correlation on one trace/channel and outputs a 
    cross-correlation trace that is normalized to have [-1, 1] 
    
    """

    # Old approach: in time-domain (slow)
#    xCorr = np.correlate(trace1.data, trace2.data, mode='valid')

    # Template length
    lenT = trace2.count()

    if stdWindowed is None:
        # Get windowed std of trace1
        # (will definitely be masked if trace1 has a mask/gaps)
        stdWindowed = calc_windowed_std_trace(trace1)

    # Use the right (i.e. fastest) approach
    #   but enforce 'time domain' correlation if stdWindowed has a mask,
    #   i.e. the data is bad (gaps, impulses, etc) and doesn't suffice FFT assumptions
    if config.conv_method == 'convolve' or np.ma.isMaskedArray(stdWindowed):
        # In time domain...

        # Do a flipped array convolution, which is a correlation
        # [To save memory, cast data to single precision (float32) ... standard is float(64)]
        xCorrTrace = np.convolve(trace1.data, trace2.data[::-1],
                                 mode='valid').astype(np.float32)

    elif config.conv_method == 'fftconvolve':
        # In frequency domain (FFT)...
        xCorrTrace = scipy.signal.fftconvolve(trace1.data, trace2.data[::-1],
                                              mode='valid').astype(np.float32)
    else:
        raise Warning("Unknown convolution method.")

    # Check if the sample length of template agrees with "theoretical" length
    # (as used in calc_windowed_std_trace())
    if lenT > int(config.duration * trace2.stats.sampling_rate + 1):
        # If actual template data is longer (mostly by one sample), slice accordingly
        if lenT % 2 == 0:  # even
            sl = slice(1, None)  # Will remove first element
        else:  # odd
            sl = slice(None, -1)  # Will remove last element
        stdWindowed = stdWindowed[sl]

    ## To be absolutely correct, do post-zero-mean (subtract means of trace & template)
    ## (for waveforms, it has no significant impact (~eps) on CC
    ##  they are already "quasi zero-mean" (if detrended/filtered) -> ignore this step
    #sl_valid = slice(lenT//2, -lenT//2+1)
    #E2 = scipy.ndimage.filters.uniform_filter1d(trace1.data, lenT)[sl_valid][sl]
    #ET = trace2.data.mean()
    #xCorrTrace -= E2*ET

    # stdWindowed may carry a mask
    # (stems from trace mask (gaps) or certain numerical issues - see calc_windowed_std_trace())
    if np.ma.isMaskedArray(stdWindowed):
        # Fill masked values with zeros
        xCorrTrace[stdWindowed.mask] = 0
        stdWindowed = np.asarray(stdWindowed)  # (masked values are already '1')
        # --> originally masked values will be 0/1 = 0

    # Actual CCC normalization ("covariance" --> "correlation")
    stdT = trace2.data.std()
    xCorrTrace /= stdWindowed*stdT * lenT

    return xCorrTrace


def do_xcorr_onetemplate(stream, template, stdWindowed=None):
    """
    xCorr parallelization level 2: Performs cross-correlation of template on continuous data.

    Makes use of level 3: do_xcorr_onechannel --> multichannel xcorr

    Returns: The events sufficing the CCC threshold (the potential events) with
                - dates/time
                - CCC
                - amplitude
    """

    # Test the type of wavefrom object + the amount of channels
    if isinstance(stream, Trace):
        # Make a Stream object from the Trace object
        stream = Stream(stream)
    if isinstance(template, Trace):
        template = Stream(template)
    elif isinstance(stream, Stream):
        if template.count() < stream.count():
            raise Exception("Template doesn't have enough channels.")

    # Init arrays
    detects = []
    detects_CCC = []
    detects_amp = []
    detects_time = []
    returnList = []
    xCorrAll = []

    # Template length
    lenT = template[0].count()

    # Number of components/channels/traces
    nTraces = stream.count()

    # Serial or parallel (to go through channels)?
    if not config.use_threads:  # Do serially

        # Looping over all provided channels
        for chnlInd in range(nTraces):

            # Calculate moving std window (required for normalizing xCorrTrace) if not yet given
            # (in the presence of gaps, stdWindowed also provides a mask [per trace])
            if stdWindowed is None:
                stdWindowed_chnl = calc_windowed_std_trace(stream[chnlInd])
            else:
                # Take it from the passed list
                stdWindowed_chnl = stdWindowed[chnlInd]

            # Actual cross correlation
            xCorrTrace = do_xcorr_onechannel(stream[chnlInd],
                                             template[chnlInd],
                                             stdWindowed_chnl)

            if config.threshPerChannel:
                if chnlInd == 0:
                    xCorrAll = xCorrTrace.copy()
                else:
                    xCorrAll = np.vstack([xCorrAll, xCorrTrace])  # Here: stack = bundle

            else:  # (= summed CC)
                # FYI: this approach will consume less memory over time (max. 2/3)
                #      b/c we stack (sum) and don't need to keep prev. arrays
                if chnlInd == 0:
                    xCorrSum = xCorrTrace
                else:
                    xCorrSum += xCorrTrace  # Summing cross-correlation traces
                    del xCorrTrace

    else:  # Parallelized xCorr

        if stdWindowed is None:
            with ThreadPool(nTraces) as pool:
                stdWindowed = pool.map(calc_windowed_std_trace, stream)

        # Distribute calculation
        with ThreadPool(nTraces) as pool:
            xCorrTraces = pool.starmap(do_xcorr_onechannel,
                                       zip(stream, template, stdWindowed))

        if config.threshPerChannel:
            xCorrAll = np.vstack(xCorrTraces)  # Bundling arrays
        else:  # (= summed CC)
            xCorrSum = np.sum(xCorrTraces, axis=0)  # Summing arrays

    # Get the indices of the detections
    if config.threshPerChannel:
        # Samples, where *all* traces are >= threshold
        sampDetect = np.where((xCorrAll >= config.threshold).all(axis=0))[0]
    else:
        # Samples, where summed traces are >= threshold*nTraces
        # (now, the threshold criteria is the average among traces)
        sampDetect = np.where(xCorrSum >= config.threshold*nTraces)[0]

    # SampleNr correction for proper timing
#    sampCorr = lenT//2  # If using 'same' mode size output
    sampCorr = 0        # If using 'valid' mode size output (fftconvolve, stdWindowed)

    # If more than one sample exceeds the threshold
    # (e.g. due to several events, high sampling rate, low threshold)
    if sampDetect.size > 0:

        # Group consecutive indices (each group should be a separate event)
        # and account for a little tolerance (to ignore 'side-lobes' exceeding the threshold)
        splitTol = template[0].stats.sampling_rate/10  # == 0.1s
        sampDetect_split = np.split(
            sampDetect, np.where(np.diff(sampDetect) >= splitTol)[0]+1)

        # And for each group ...
        for i in range(len(sampDetect_split)):

            # If a detection is at the very beginning, disregard it (it's likely a filter response)
            if sampDetect_split[i][0] < sampCorr:
                continue

            # ... get the sample of waveform trace where summed CC is highest
            if config.threshPerChannel:
                CCsum = np.sum(xCorrAll[:, sampDetect_split[i]], axis=0)
            else:
                CCsum = xCorrSum[sampDetect_split[i]]
            iMaxCCC = np.argmax(CCsum)
            # (+ correct for sample timing!)
            detects.append(sampDetect_split[i][iMaxCCC] - sampCorr)

            # ... get the representing CCC value
            if config.threshPerChannel:
                # (the minimal CCC among all traces)
                CCCvalue = min(xCorrAll[:, sampDetect_split[i][iMaxCCC]])
            else:
                # (the average CCC among traces)
                CCCvalue = CCsum[iMaxCCC]/nTraces
            detects_CCC.append(CCCvalue)

            # ... and the maximum amplitude (using the last entry in 'detects',
            # which has already corrected time
            detects_amp.append(
                max(abs(stream[0].data[detects[-1]: detects[-1]+lenT])))

        # Finalize
        if len(detects) == 1:
            print("   The following event has been detected:", flush=True)
        else:  # if > 1
            print("   The following %d events have been detected:"
                  % len(detects), flush=True)

        for i in range(len(detects)):

            # Derive the event time(s)
            detects_time.append(
                stream[0].stats.starttime + detects[i]*stream[0].stats.delta)

            print("     %s | CCC: %.4f | Amp: %d" % (detects_time[i],
                                                     detects_CCC[i],
                                                     detects_amp[i]))

            # Reorder the detects to form a list
            returnList.append([detects_time[i],
                               round(detects_amp[i], 1),   # Round already to avoid
                               round(detects_CCC[i], 6)])  # later "mis-rounding"

    return returnList
