# -*- coding: utf-8 -*-
"""
Submodule for updating / modifying / processing the event list.

(created) 2015, Marcus Herrmann, Swiss Seismological Service, ETH Zürich
"""

# --- Import modules

# General
import os
import datetime
import shutil  # e.g. to copy files
import copy    # e.g. to make deepcopies of nested lists
#import glob    # For reading files with wildcard
import collections
from multiprocessing.pool import ThreadPool

# Operands
from operator import itemgetter
from itertools import repeat

# Scientific
import math
import scipy
import numpy as np

# ObsPy
from obspy import UTCDateTime
from obspy.core import read, Stream

# Own modules
from TM.setupclass import config
from TM.waveformops import waveform_processing, get_max_amp
from TM.utils import getChannelIndex
from TM import io
from TM import plotting


def getAmplitude(events, eventType, freqMax=None, rota=0, tdiff=0, fullDetStream=False):
    """
    Derives the peak amplitude from the specified 'events' using all provided channels.
    'eventType' specifies the type of the event list; a special case is "catalog" and "GELcatalog".

    fullDetStream:
        Use the full detection stream instead of slicing detection stream to a part where
        its template has its maximum.
        (necessary when actual locations of maxima don't agree (i.e. once @ P-, once @ S-Phases)

    TODO: be smarter and load events on the same day only once
          (can be ignored if waveforms of all components are present in
           the last entry's element --> then just use this)
    """

    print("\n[getAmplitude()] (of %d) ---------------------" % len(events))

    # Copy events to not touch original list
    events = copy.deepcopy(events)

    if eventType == "catalog":

        # The catalog should have an ID in first column (and then the time)
        if not isinstance(events[0][1], UTCDateTime):
            # Otherwise prepend the continous index of/for each element
            [x.insert(0, ind) for ind, x in enumerate(events)]
            raise Warning("Catalog doesn't have event IDs. Using index.")

        # Load the catalog events into MSEEDs (if they don't exist yet)
        # Initial loading should be using 3 COMPONENTS!
#        preload_events(events)

    elif eventType == "GELevents":

#        channelSuffix = ['xxz', 'xx2', 'xx3']  # MSEEDs' channel names
        # ... for GSE2
        if config.station == 'OTER2':
            channelSuffix = ['EHZ', 'EH2', 'EH3']
        elif config.station == 'MATTE':
            channelSuffix = ['EHA', 'EHB', 'EHC']
        # Get the appropriate channels
        channelSuffix = [channelSuffix[i] for i in getChannelIndex(config.channels)]

        # Make the event list compatible
        for event in events:
            try:
                GEL_Mw = float(event[2])
            except:
                # Magnitude not available ("-.--")
                continue  # ... with next GEL event
            # If using GELevents associated with a foundEvent [using compareWithGELDB()]
#            event = [event[0], event[-3], GEL_Mw, event[-2], event[-1]]
            # ... or read directly
            event = [event[0], event[1], GEL_Mw]

    else:  # (for detections)

        # Retrieve the relative time-of-maximum amplitude of each template
        relTimeOfMax_templates = []
        for iTemplate in range(max([int(row[-2]) for row in events])):
            templateStream = read(config.templateDir + "/templates_mseed/template%02d_" %
                                            (iTemplate+1) + "*_raw.mseed")
            waveform_processing(templateStream, 'highpass', None, freqMax, copy=False)
            relTimeOfMax_templates.append(get_max_amp(templateStream, ret='time'))
#        # Eventually, they should be all similar to the first template
#        templateStream = read(config.templateDir + "/templates_mseed/template01*_raw.mseed")
#        waveform_processing(templateStream, 'highpass', None, freqMax, copy=False)
#        relTimeOfMax_templates = [get_max_amp(templateStream, ret='time') *\
#            len(io.readTemplateSet(config.directory + ""))

    # TODO: Parallelize?
    for idx, event in enumerate(events):

        # Read waveforms (of each channel)
        try:
            if eventType == "catalog":

                eventTime = event[1]
                waveformStream = read(os.path.join(
                                      config.dataDir, 'SED-events_%s_%s_mseed/%03d_' %
                                      (config.region, config.station, event[0]) +
                                      str(eventTime).replace(':', '-') + '.mseed'))
                # Take only the desired channels
#                channel_select = channels[0][:2] + '[' + ''.join([x[-1] for x in channels]) + ']'
#                waveformStream.select(channel=channel_select)
                waveformStream = Stream(
                    [waveformStream[i] for i in getChannelIndex(config.channels)])
#                if rota and len(waveformStream) > 1:
#                    print("Rotated by %d deg." % rota)
#                    waveformStream[-2].data, waveformStream[-1].data = rotate_ne_rt(
#                        waveformStream[-2].data, waveformStream[-1].data, rota)
#                    waveformStream = Stream(waveformStream[-2])

            elif eventType == "GELevents":

#                eventTime = event[3]
#                waveformStream = read(TMdata + directory.replace("_pass2", "")
#                                      + "/detections_mseed/" + str(eventTime).replace(":", "-")
#                                      + "_%.2f" % event[-1] + "_Template%02d" % (event[-2])
#                                      + ".mseed")
                # Read directly
                eventTime = event[1]
                # MSEED
#                waveformStream = Stream()
#                for chnl in channelSuffix:
#                    waveformStream += read(config.TMdata + "Basel_GELevents/MSEED_1000HZ/%s_%s.%s"
#                                           % (event[0], config.station.lower(), chnl))
                # Read GSE
                waveformStream = read(config.TMdata + "Basel_GELevents/GSE2_1000HZ/%s.gse2"
                                      % event[0], verify_chksum=False, apply_calib=False)
                # Select Station
                waveformStream = waveformStream.select(station=config.station)
                # Select channel
                waveformStream = waveformStream.select(
                    channel='*[' + ''.join([x[-1] for x in channelSuffix]) + ']')

            else:  # (if detection)

                eventTime = event[0]
                # Here, some strange thing can happen: bad rounding, e.g. 0.495000 --> 0.49 !!!
                waveformStream = read(config.TMdata + config.directory.replace("_pass2", "") +
                                      "/detections_mseed/" + str(eventTime).replace(":", "-") +
#                                      "_%.2f" % event[-1] + "_Template%02d" % (event[-2]) +
                                      ".mseed")
                waveformStream.detrend(type="linear")
                # If amplitude should depend on template...
                if not fullDetStream:
                    # Slice stream to the part where also its template has maximum amplitude
                    timeMaxAmp = eventTime + relTimeOfMax_templates[event[-2]-1]
                    # Trim to a small part (incl. +/- filter buffer time)
                    waveformStream.trim(timeMaxAmp - 0.1 - config.filterBuff,
                                        timeMaxAmp + 0.1 + config.filterBuff)

        except Exception as err:

            print("\nCannot read waveform of %s event %d." % (eventType, idx+1), err)
            if eventType == "catalog":  # For not-readable catalog events, set amplitude to NaN
                event += [float('NaN')]
#            elif eventType == "GELevents":  # Delete not-readable GELevents
#                del event
            continue  # ... with next event
        else:
            pass

        # Highpass filtering (with no resampling)
        if eventType == "GELevents":
            # Should not *filter* if GEL event is read directly! (b/c it's only 6s)
            # but only detrend stream
            waveformStream.detrend(type="linear")
        else:
            waveform_processing(waveformStream, 'highpass', None, freqMax, copy=False)

        amp = get_max_amp(waveformStream)

        # Some specific amplitude corrections
        if eventTime > UTCDateTime("2012-05-11") and \
           config.station in ['OTER2', 'MATTE']:
            # make comparable using the scale (= total sensitivity [cnt/(m/s)])
            scale_period1 = 1137136837.0
            scale_period2 = 722794747.8
            amp *= scale_period1/scale_period2

        if eventType == "catalog" or eventType == "GELevents":
            # Append amplitude to the catalog's entries
            event += [round(amp, 1)]
        else:
            # Replace current amplitude with new one (rounded)
            event[1] = round(amp, 1)

        # Status report
        if (idx % 100 == 0 and idx > 0):
            print("%d  " % idx, end="", flush=True)

    print("\nDetermined the amplitudes of %d events." % len(events))

    return events


def correct_amplitude_attenuation(events, eventType):
    """
    Corrects the amplitude of based on the distance b/w event and station.
    (= undo attenuation)
    For detections, correct amplitude based on template association.

    The catalog contains the locations events
    """

    events = copy.deepcopy(events)

    # OTER2
    station_coord = [612486.0,  # Lon:  7.60459
                     269838.0,  # Lat: 47.57912
                     2487.39]   # NN (2740 - 252.61)
#    # MATTE
#    station_coord = [609848.0,  # Lon:  7.56945
#                     266749.0,  # Lat: 47.55139
#                     274.0]       # NN (???? - ???)

    # Get the catalog events that contain locations
    if eventType is not "GELevents":
        catalog = io.getNicoCatalog(getAll=True, locations=True)
    else:
        catalog = io.getCatalogGEL(GELIDs=["Main", "Post", "Late"], locations=True)

    # Determine distances (depending on event type)
    if eventType in ["catalog", "GELevents"]:

        # Identify the catalog events
        events_loc = [event for ev in events
                      for event in catalog if event[1] == ev[1]]

        events_dist = [math.sqrt((ev[2]-station_coord[0])**2 +
                                 (ev[3]-station_coord[1])**2 +
                                 (ev[4]-station_coord[2])**2) for ev in events_loc]
        ampInd = -1
    else:
        # Get the current templateSet
        templateSet = io.readTemplateSet(config.directory)

        # Identify which catalog events are templates (and get their magnitudes/locations)
        templateSet_loc = [event for template in templateSet
                           for event in catalog if event[1] == template[0]]

        # Determine template distances
        templateSet_dist = [math.sqrt((ev[2]-station_coord[0])**2 +
                                      (ev[3]-station_coord[1])**2 +
                                      (ev[4]-station_coord[2])**2) for ev in templateSet_loc]
        ampInd = 1

    # Apply amplitude correction
    for idx, event in enumerate(events):

        # Get the distance
        if eventType in ["catalog", "GELevents"]:
            R = events_dist[idx]
        else:
            R = templateSet_dist[event[-2]-1]  # Depending on its template distance

        # Correct for geometrical spreading (spherical decrease)
        event[ampInd] /= 1000/R  # R in [m]
#        event[ampInd] /= (1000/R)**0.59  # Edwards et al. (2010, 2011, 2013)

        # Correct for damping/absorption (for shear wave)
#        Qs = 1200  # [Bethmann 2012]
#        Qs = 1216  # Reference crustal Q0 for vs=3.5km/s [Edwards 2011]
        Qs = 400   # Conservative value
        fq = 30    # [Hz] S-wave frequency
#        fq = 201 + -23.6*math.log10(event[ampInd])  # Relation of amp -> peak amplitude Basel
        vs = 3500  # [m/s] S-wave velocity
        event[ampInd] /= math.exp(-(math.pi*fq*R)/(Qs*vs))  # e.g., see NMSOP Chpt. 2

    return events


#def loglikelihoodRegress(params, x, yObs):
#    """"
#    Definition of the likelihood function that takes initial parameter estimates
#    """
#
#    b, a, sd = params
#
#    # Calculate the predicted values from the initial parameter guesses
#    yPred = a + b*x
#
#    # Calculate the negative log-likelihood as the negative sum of the log of a normal PDF
#    # where the observed values are normally distributed around the mean (yPred)
#    # with a standard deviation of sd
#    logLike = -np.sum(scipy.stats.norm.logpdf(yObs, loc=yPred, scale=sd))
#
#    return logLike


def regressMagnitude(foundEvents, catalog, magRange):
    """
    Determines the magnitde (ML) for the detections (foundEvents) based on a
    regression b/w the catalogs' amplitudes and magnitudes. This relation is
    then applied to the amplitude of the detections.

    'foundEvents' contains the standard entries of the detections:
        [ time, amp, template, CCC ] - only 'amp' and 'CCC' needed
    'catalog' is a list of the events in this format:
        [ ID, time, ML ]

    Returns: an updated 'foundEvents' with inserted magnitudes
    """

    print("\n[regressMagnitude()] ---------------------------")

    # Check if detection list does not already contain a magnitude...
    if len(foundEvents[0]) > 4:
        raise Warning("WARNING: Detection list already contains a magnitude.")
        # Delete magnitudes
#        for row in foundEvents:
#            del row[2]
    # Check if catalog contains amplitudes
    if len(catalog[0]) < 4:
        raise Warning("Catalog doesn't contain amplitudes. Will do nothing.")
    if len(catalog[0]) > 4:
        raise Warning("Catalog has to much values. It should only have ID, time, Mag, amp.")


    # Deepcopy catalog + events to not touch original
    catalog = copy.deepcopy(catalog)
    foundEvents = copy.deepcopy(foundEvents)

    # Remove invalid events from catalog (e.g. not readable, i.e. amp==NaN or amp==0.0)
    catalog = [x for x in catalog if (len(x) == 4 and not (np.isnan(x[-1]) or x[-1] == 0.0))]

    # Use only events withing specified magnitude range
    catalog_minMag = copy.deepcopy([x for x in catalog
                                    if (x[2] >= magRange[0] and x[2] <= magRange[1])])

    # Get amplitude
    catalog_A = np.asarray([x[-1] for x in catalog_minMag])
    # Get catalog MLs
    catalog_ML = np.asarray([x[2] for x in catalog_minMag])

    # Do regression using Linear Least Squares of log10-amp
    b, a, r_value, p_value, std_err = scipy.stats.linregress(np.log10(catalog_A), catalog_ML)
    print("\nLS-regression: ML = %.2f * log(A) + %.2f" % (b, a))

#    # Do regression using maximum likelihood
#    nll = lambda *args: -loglikelihoodRegress(*args)
#    initParams = [b, a, std_err]
#    initParams = [1, 1, 1]
##    catalog_XY = np.sort(np.column_stack((catalog_logA, catalog_ML)), axis=0)
#    result = scipy.optimize.minimize(nll, initParams, args=(catalog_XY[:,0], catalog_XY[:,1]),
#                                     method='nelder-mead')
#    print(result)
#    b_maxL, a_maxL, sd_maxL = result["x"]

    plotting.plot_amp_regression(catalog_ML, catalog_A,
                                 regLineInfos=[b, a, r_value, p_value, std_err, ''])

    # Update foundEvents with REGRESSION ML
    [x.insert(2, float("{0:.2f}".format(b*math.log10(x[1]) + a))) for x in foundEvents]

    # Update catalog with REGRESSION ML
    [x.append(float("{0:.2f}".format(b*math.log10(x[3]) + a))) for x in catalog_minMag]

    # Update catalog with ALL events
    [x.append(float("{0:.2f}".format(b*math.log10(x[3]) + a))) for x in catalog]

    # Write out catalog events USED FOR REGRESSION
    file = config.directory + "/regression_catalog-onlyUsed.dat"
    with open(file, 'w') as f:
        for event in catalog_minMag:
            line = ("%s   %s   %.2f   %.1f   %.2f\n" % tuple(event))
            f.write(line)
    # Write out ALL catalog events
    file = config.directory + "/regression_catalogALL.dat"
    with open(file, 'w') as f:
        for event in catalog:
            line = ("%s   %s   %.2f   %.1f   %.2f\n" % tuple(event))
            f.write(line)

    print("Used %d events." % len(catalog_ML))

    return foundEvents


def ampratioMagnitude(foundEvents, templateSet, catalog_amp):
    """
    Determines the magnitde (ML) for the detections (foundEvents) based on
    the amplitude ratio b/w template event and its detections.

    'foundEvents' contains the standard entries of the detections:
        [ time, amp, template, CCC ] - only 'amp' and 'CCC' needed
    'catalog' is a list of the events in this format:
        [ ID, time, ML ]

    Returns: an updated 'foundEvents' with inserted magnitude
    """

    print("\n[ampratioMagnitude()] ---------------------------")

    # Check if detection list does not already contain a magnitude...
    if len(foundEvents[0]) > 4:
        raise Warning("WARNING: Detection list already contains a magnitude.")

    # Deepcopy catalog + events to not touch original
    foundEvents = copy.deepcopy(foundEvents)

    prevInd = 0  # To speed up operations on a long list
    AmpMag_templates = []
    # Get amplitudes of template events
    for template in templateSet:
        # Get the event index + amplitude of an event that is within the time tolerance
        eventSel = [(x, y[1]) for (x, y) in enumerate(foundEvents[prevInd:])
                    if abs(y[0] - template[0]) <= 0.1]
##        print(eventSel[0][0]+prevInd)
#        print("%s   %9.1f   tp%d   %.6f" % tuple(foundEvents[eventSel[0][0]+prevInd]), end='')
##        print("tp{3:d}: {0:s}  {2:9.1f}  {1:.6f}".format(
##            *foundEvents[eventSel[0][0]+prevInd]), end='')
#        print("   M%.2f" % template[-1])
        AmpMag_templates.append([eventSel[0][1]] + [template[-1]])

    print(AmpMag_templates)
    # Update foundEvents with AMP-RATIO ML
    for event in foundEvents:
        tp_amp, tp_Mag = AmpMag_templates[event[-2]-1]
        event.insert(2, round(tp_Mag - math.log10(tp_amp/event[1]), 2))

    # Make catalog file compatible
    catalog_amp = copy.deepcopy(catalog_amp)
    [x.append(x[2]) for x in catalog_amp];

    # Write out ALL catalog events
    file = config.directory + "/regression_catalogALL.dat"
    with open(file, 'w') as f:
        for event in catalog_amp:
            line = ("%s   %s   %.2f   %.1f   %.2f\n" % tuple(event))
            f.write(line)

    print("Finished determining magnitudes.")

    return foundEvents


def assocMagnitude(foundEvents, catalog, catalogType='SED', position=-1, onlyMissing=False):
    """
    Given a detection list and a catalog, identifies the detections that correspond to
      the catalog events (using cross-correlation) and THEN assignes them its magnitude
      (in a new column).

    If position = -1, automatically determines the best position.
      E.g., in case of already present magnitudes columns, append new column after them.
    """

    print("\n[assocMagnitude()] Adding %s magnitudes..." % catalogType)

    foundEvents = copy.deepcopy(foundEvents)
    numCols = len(foundEvents[0])

    # If desired, determine position to place the new magnitude
    if position == -1:
        position = 2  # Default position
        if (numCols > 4):  # If a magnitude is already present
            position = numCols - 2  # Append after existing magnitudes

    if catalogType == 'SED':

        timeTol = 1.0
        if len(catalog[0]) < 4:
            raise Warning("Catalog doesn't contain amplitudes. Will do nothing.")

        # Insert a new column
        [x.insert(position, float('NaN')) for x in foundEvents]
        numCols += 1

        prevInd = 0  # To speed up operations on a long list
        # Associate each catalog event with a corresponding detection
        for catalogEvent in catalog:
            # Get the event index + amplitude of an event that is within the time tolerance
            eventSel = [(x, y[1]) for (x, y) in enumerate(foundEvents[prevInd:])
                        if abs(y[0] - catalogEvent[1]) <= timeTol]
            if eventSel and onlyMissing is False:
                # Associate to the one with highest amplitude
                eventInd = max(eventSel, key=itemgetter(1))[0]
                foundEvents[eventInd+prevInd][position] = catalogEvent[2]
                prevInd = eventInd
            elif not eventSel:
                # If it's not in the list, take it in!
                #    Create an "empty" event ('-2': disregard time, Amp, and CC)
                eventToAdd = [catalogEvent[1]] + [catalogEvent[-1]] + [float("NaN")] * (numCols-3)
                #    Fill relevant magnitudes
                eventToAdd[position] = catalogEvent[2]
                foundEvents.append(eventToAdd + [float("Inf")])

    elif catalogType == 'GEL':

        # Insert two new columns
        [x.insert(position, float('NaN')) for x in foundEvents]
        [x.insert(position, float('NaN')) for x in foundEvents]
        numCols += 2

        # Try to identify them using xCorr
#        foundGELevents, undetectedGELevents, GELasssoc = compareWithGELDB(
#            foundEvents, channels, 'results_OTER2_2006-12-01_2006-12-13_2comp_decim250',
#            GELIDs=["Main"], removeDouble=True)
        # Unpack the catalog from given results of compareWithGEL()
        foundGELevents, undetectedGELevents = catalog

        # Check if a detection was associated with more than one GEL event
        if onlyMissing is False:
            foundGELevent_uniqTimes = [str(x[3]) for x in foundGELevents]
            foundGEL_double = [x for x, count in
                               collections.Counter(foundGELevent_uniqTimes).items()
                               if count > 1]
            if len(foundGEL_double) > 4:  # Four doublets are known
                print('%d detections were associated with more than one GEL event. '
                      'Used the last associated GEL event in each case.'
                      '(and ignored the others).' % len(foundGEL_double))

            # Assign to foundEvents
            foundEvents_times = [x[0] for x in foundEvents]
            for GELevent in foundGELevents:
                eventInd = foundEvents_times.index(GELevent[3])
    #            eventInd = [i for i, x in enumerate(foundEvents) if x[0] == GELevent[3]]
                foundEvents[eventInd][position] = float(GELevent[2])  # Add original Mw
                foundEvents[eventInd][position+1] = round(
                    float(GELevent[2])*1.58-1.21, 2)  # Add ML

        # Insert undetectedGELevents (assuming it has already a SED magnitude)
        for GELevent in undetectedGELevents:
            # TODO: Check if GELevent is an undetected SED event

            # Create an "empty" event ('-2': disregard time + CC)
            eventToAdd = [GELevent[1]] + [float("NaN")] * (numCols-2)
            # Fill relevant magnitudes
            eventToAdd[position] = float(GELevent[2])  # Add original Mw
            eventToAdd[position+1] = round(float(GELevent[2])*1.58-1.21, 2)  # Add ML
            foundEvents.append(eventToAdd + [float("Inf")])

        # For PostXXXXX, & LateXXXXX GEL events, try to associated only using time
        timeTol = 6.0
        # Get the remaining GEL events in that period
        remainingGELevents = io.getCatalogGEL(GELIDs=["Post", "Late"], useConfigDates=True)
        # Only take the ones that have a magnitude and are before 2012
        remainingGELevents = [x for x in remainingGELevents
                              if x[-1] != '-.--' and x[1] < UTCDateTime('2012-05-20')]
        for GELevent in remainingGELevents:
            eventSel = [(x, y[1]) for (x, y) in enumerate(foundEvents)
                        if abs(y[0] - GELevent[1]) <= timeTol]
            if eventSel and onlyMissing is False:
                # Associate to the one with highest amplitude
                eventInd = max(eventSel, key=itemgetter(1))[0]
            elif not eventSel:
                # If it's not in the list, take it in!
                foundEvents.append([GELevent[1]] + [float("NaN")] * (numCols-2) + [float("Inf")])
                eventInd = -1
            # Fill relevant magnitudes
            foundEvents[eventInd][position] = float(GELevent[2])  # Add original Mw
            foundEvents[eventInd][position+1] = round(float(GELevent[2])*1.58-1.21, 2)  # Add ML
        # Post-Info:
        #        All found & undetect GELevents will contain a magnitude (and are readable).
        #        Post & LateXXXXX events *may* contain a magnitude.
        #        All unavailGELevents will NOT contain a magnitude (and can act. be ignored).
        #        These magnitudes (original and converted) should be treated as strings.

    print("Finished adding %s magnitudes." % catalogType)

    return sorted(foundEvents)


def uniqueEventList(foundEvents, removeSelfDetection=False):
    """
    Unifies the event list in case of multiple detections
       (e.g. several templates detected the same event - with a certain time tolerance).
    In case of a duplicate (or multiplate), take the detection with the highest associated CCC.

    TODO: also take into account 'amplitude' (b/c it's also similar)
    TODO: only unique detections from DIFFERENT templates
    TODO: Implement "removeSelfDetection" (CCC=1.0)
        Actually possible: only remove templates if templateSet is provided
    TODO: Check if it hangs when CCC is the same for one detection
    """

#    # Deepcopy events to not touch original
#    foundEvents = copy.deepcopy(foundEvents)

    # Get the times and their indexes
    idx, times = zip(*enumerate([x[0] for x in foundEvents]))

    # Note: there is certainly a more elegant way...

#    # Return the index of the first event of multiple
#    multiplets = np.where(abs(np.diff(times)) <= config.timeTol)[0]

    # Group the multiples together by index using the times
    multiplets = np.split(idx, np.where(abs(np.diff(times)) >= config.timeTol)[0]+1)

    foundEvents_unique = []
    if len(multiplets) > 0:

#        # Group the multiples together
#        multiplets = np.split(multiplets, np.where(np.diff(multiplets) != 1)[0]+1)

        # Generate the indices of the multiples that should be excluded
        for i in range(len(multiplets)):
#            # Add the index of the remaining duplicate
#            # (because .diff() misses the last of each multiplet)
#            multiplets[i] = np.append(multiplets[i],
#                                         multiplets[i][-1]+1).tolist()

            # Derive the CCCs of a multiplet
            CCCs = [foundEvents[j][-1] for j in multiplets[i]]
#            CCCs = [(ind, foundEvents[j][-1]) for ind, j in enumerate(multiplets[i])]

            idx_highestCCC = multiplets[i][CCCs.index(max(CCCs))]
#            idx_highestCCC = multiplets[i][max(CCCs, key=itemgetter(1))[0]]  # Avoiding .index()

            # Excludes the event with the maximum CCC
            # ... and returns the events to be removed
            # NOTE: When a template detects another template,
            #       it will only keep the CCC 1.0 self detection
#            del idx_highestCCC

            # Append the event with the maximum CCC
            foundEvents_unique += [foundEvents[idx_highestCCC]]

#        # Flatten this list
#        multiplets = [item for sub in multiplets for item in sub]

#        # Finally return the uniqued list
#        # (by removing the duplicate events with lower CCC)
#        # (go backwards to not mess up the indexing; it changes when deleting elements)
#        for i in range(len(multiplets)-1, -1, -1):
#            del foundEvents[multiplets[i]]

    print("\n[uniqueEventList()] Removed %d duplicates." % (len(foundEvents) - len(multiplets)))

    # Now test if there could still be potential multiplets (using the amplitude similarity)
    amplitude = [x[1] for x in foundEvents_unique]
    amp_multiplets = np.where(abs(np.diff(amplitude)) <= 0.1)[0]
    amp_multiplets = np.split(amp_multiplets,
                              np.where(np.diff(amp_multiplets) != 1)[0]+1)

    print("But there are still %d potential multiplet pairs." % len(amp_multiplets))
    print("(Or these are double events with only ~1-2 seconds time difference)")
#    for i in range(len(amp_multiplets)):
#        print("%d, " % amp_multiplets[i][0], end="")

    return foundEvents_unique


#def correctTimes(foundEvents, templateSet):
#    """
#    Correct the detection times to have the detections of different templates at roughly
#    the same time sample.
#    Approach: get the maximum of each template and correct detection times accordingly.
#              This way, the times will be shifted relative to the templates' S-wave maximum.
#    """
#
#    # Deepcopy events to not touch original
#    foundEvents = copy.deepcopy(foundEvents)
#
#    templateCorrection = []
#
#    # Loop over templates and determine the temporal correction
#    for idx, template in enumerate(templateSet):
#
#        templateNo = idx+1
#
#        # Get the stream / waveform
#        templateStream = read(directory + "/templates_mseed/template%02d_" % templateNo +
#                              str(template[i][0]).replace(":", "-") + ".mseed")
#
#        for chnlInd in range(templateStream.count()):
#            # Build up the sum of the squared waveform
#            if chnlIndex == 0:
#                waveformPow2 = templateStream[0].data**2
#            else:
#                waveformPow2 += templateStream[chnlInd].data**2
#
#        # Calculate time correction (maximum minus time-until-maximum)
#        t_corr = np.argmax(waveformPow2**0.5) * templateStream[0].stats.delta - timeToMaxAmp
#        # And append
#        templateCorrection.append(t_corr)
#
#    # Loop over detections and apply time correction
#    for detection in foundEvents:
#        detection[0] += templateCorrection[int(detection[-2].lstrip("tp"))-1]
#
#    return foundEvents


# --- Misc


def load_detections(foundEvents, station=None):
    """
    This function retrieves the waveforms for given detection infos.

    Ideally, this crawling can work as a separate process - indepentently from the core
    template matching routine to speed up xCorr (but receives its detections).

    It can also be used to retrieve data again or from another station.
    """

    if station is None:
        station = config.station
        tdiff = 0.0

    else:
        # Get station group
        groups = ['OTER2', 'OTER1', 'HALTI', 'JOHAN', 'MATTE', 'RIEH2']
        group = groups.index(station)*2+1

        # Determine arrival time pattern of S-phase
        # Approximation here: additional arrival time to OTER2
        #   1. Read the triggers and picks
        filename = "../Info_BaselEGS/Dyer_BaselEventData_Okt2010/TriggerTimes_All.evt"
        with open(filename, "r") as file:
            GELtriggerCat = [line.strip().split() for line in file if not line[0] == "#"]
        #   2. extract the Ts times
        GEL_Ts_OTER2 = [float(x[12])/1000 for x in GELtriggerCat]
        GEL_Ts_station = [float(x[7+group*5])/1000 for x in GELtriggerCat]
        #   3. build Tp differences (equals ~ travel time differences)
        Ts_diff = [y-x for x, y in zip(GEL_Ts_OTER2, GEL_Ts_station) if x and y]
        #   4. Derive a representive time difference
#        hist = plt.hist(Ts_diff, bins=np.arange(0.0, 6.0, 0.01))
#        tdiff = hist[1][hist[0].argmax()]
#        tdiff = scipy.stats.mode(Ts_diff)[0][0]
        tdiff = np.median(Ts_diff)

    # Set up the "big-data folder (but don't create a pass2 folder there)
    TMdata_now = config.TMdata + config.directory.replace("_pass2", "")
    os.makedirs(TMdata_now + "/detections_mseed_%s" % station, exist_ok=True)

    # Group foundEvents by day
    dates = [x[0].date for x in foundEvents]
    dateSplitIdx = np.where(np.diff(dates) == datetime.timedelta(1))[0]
    foundEventsSplit = np.split(foundEvents, dateSplitIdx+1)

    # Go through foundEvents_daily (day-wise)
    for foundEventSet in foundEventsSplit:

        date = UTCDateTime(foundEventSet[0][0].date)

        # Load the daily Stream
        try:
            # Read files in parallel using threads
            with ThreadPool(config.numChannels) as pool:
                dailyStream = Stream(pool.starmap(
                    io.get_waveform, zip(repeat(config.network), repeat(station),
                                         config.channels, repeat(date), repeat(86400),
                                         repeat('standard'))))
        except Exception as err:
            print("Cannot read daily waveform (or any component).", err)

        # For each event in this set
        for idx, event in enumerate(foundEventSet):

            t = event[0] + tdiff

            detectionStream_RAW = dailyStream.slice(t - config.filterBuff,
                                                    t + config.filterBuff + config.duration)

            # .. and save detection waveform (with the original time)
            detectionStream_RAW.write(
                TMdata_now + "/detections_mseed_%s/" % station +
                str(event[0]).replace(":", "-") +
#                "_%.2f" % detection[3] + "_Template%02d" % (iTemplate+1) +
                ".mseed",
                format='MSEED', encoding="FLOAT32")

        # TODO: routine to receive detections and process them
    return tdiff


def cleanUpMSEEDs(foundEvents):
    """
    Cleans up MSEEDS (the ones which are not in the uniqued list anymore).
    (because they can take up A LOT of space)
    """

    print("\n[cleanUpMSEEDs()]")

    # Build the uniqued list
    foundEvents_unique = uniqueEventList(foundEvents)
    # .. and the copy the original list (will be modified)
#    foundEvents_nonUnique = copy.deepcopy(foundEvents)

    lenAll = len(foundEvents)
    lenUniq = len(foundEvents_unique)
    # Check that the given foundEvent list is indeed the UN-uniqued one
    if lenAll == lenUniq:
        raise Warning("Attention: Cannot clean, list is already uniqued.")
        return None

    TMdata_now = config.TMdata + config.directory.replace("_pass2", "")
    seed_directory = TMdata_now + "/detections_mseed/"

    print("Cleaning up %d *.mseed files (%.fperc)." % (lenAll-lenUniq, (1-lenUniq/lenAll)*100))

    # Version1: Move all unique event mseeds to a temp dir, then delete all old, then move back
    # (preferred method to clean, assuming that fewer files will kept than removed)
    errorCount = 0
    tmp_directory = TMdata_now + "/detections_mseed_unique/"
    os.makedirs(tmp_directory, exist_ok=True)
    for idx, event in enumerate(foundEvents_unique):
        fileToBeSaved = (seed_directory +
                         str(event[0]).replace(":", "-") +
#                         "_%.2f" % event[-1] + "_Template%02d" % (event[-2]) +
                         ".mseed")
        try:
            shutil.move(fileToBeSaved, tmp_directory)
        except:
            errorCount += 1
            continue
        # Status report
        if (idx % 10000 == 0):
            print("%d  " % idx, end="", flush=True)
    # Delete directory with never-to-be-used mseed files
    shutil.rmtree(seed_directory)
    # Move temporary directory to the old location --> cleaned
    shutil.move(tmp_directory, seed_directory)

    print("Cleaned up %d *.mseed files (%.f%). Encountered %d errors."
          % (lenAll-lenUniq, (1-lenUniq/lenAll)*100, errorCount))

#    # Version2: Remove all unneccesary mseed files IN-PLACE
#
#    # Removing the uniqued events from ALL events
#    # (go backwards to not mess up the indexing; it changes when deleting elements)
#    for i in range(len(foundEvents_nonUnique)-1, -1, -1):
#        # Check if they are the same...
#        if foundEvents_nonUnique[i] == foundEvents_unique[-1]:
#            del foundEvents_nonUnique[i]  # ... and delete the uniqued one here
#            del foundEvents_unique[-1]    # ... and here
#            if len(foundEvents_unique) == 0:
#                break
#
#    callstr = ""
#    for idx, event in enumerate(foundEvents_nonUnique):
#
#        filename = (seed_directory
#                    + str(event[0]).replace(":", "-") + "_%.2f" % event[-1]
#                    + "_Template%02d" % (event[-2]) + ".mseed")
#
#        # Variant 2.1 (multiple rm call)
#        callstr += filename + " "
#
#        # With enough files as arguments, delete themm in one run
#        if (idx % 1000 == 0):
#            print("%d  " % idx, end="", flush=True)
#            try:
#                subprocess.check_call("rm -f %s" % callstr, shell=True)
#                callstr = ""
#            except:
#                callstr = ""
#                print("X", end="", flush=True)
#                continue
#        # Variant 2.2 (single rm call)
#        try:
#            os.remove(filename)
#        except:
#            continue
#        # Status report
#        if (idx % 1000 == 0):
#            print("%d  " % idx, end="", flush=True)
