# -*- coding: utf-8 -*-
"""
Submodule containing input / output functions.

(created) 2015, Marcus Herrmann, Swiss Seismological Service, ETH Zürich
"""

# --- Import modules

import copy    # e.g. to make deepcopies of nested lists
import os      # OS specific
import glob    # For reading files with wildcard
from multiprocessing.pool import ThreadPool

# Operands
from itertools import repeat

# Scientific
import numpy as np
import matplotlib.pyplot as plt

# ObsPy
from obspy import UTCDateTime
from obspy.core import read, Stream
from obspy.clients.fdsn.client import Client as clientFDSN
from obspy.clients.filesystem.sds import Client as clientSDS
#from obspy.clients.arclink.client import Client as clientArcLink

# Own modules
from TM.setupclass import config
from TM.core import calc_windowed_std_trace
from TM.waveformops import (waveform_processing, waveform_processing_parallel,
                            getTimeShiftOfMaxSimilarity)
from TM.utils import getChannelIndex, limit_numpy_fft_cache

# --- Waveform-related


def get_waveform(*args, **kwargs):
    """
    Choose the right function to derive waveforms
    """

    if config.FDSN:
        return get_waveform_FDSN(*args, **kwargs)
    else:
#        return get_waveform_localMiniSEED(*args, **kwargs)
        return get_waveform_SDS(*args, **kwargs)


def get_waveform_localMiniSEED(
        network, station, channel, t, duration, procType='raw', aTrim=False):  # , **kwargs
    """
    Returns a Trace object (not Stream object) containing the waveform for specified parameters.
    This function doesn't access the global Setup class; it can be used independently.

    Options (kwargs):
     - aTrim: If "True", the function (tries to) select the event waveform
              of interest (e.g. in case "t" doesn't provide a good time clue)
              Attention: may not work properly in case of several events in short
                         time (especially an event with larger amplitude)
     - waveProc: Type of waveform processing (is passed to waveform_processing()
    """

    # Define options and their default values
#    aTrim = kwargs.get('aTrim', False)
#    fType = kwargs.get('fType', 'bandpass')

#    network = "CH"
#    station = "OTER2"
#    channel = "DP1"
#    t = UTCDateTime("2006-12-04T00:04:16.500000Z")
#    duration = 4

    if (os.name == "posix"):
        localPath = config.mseedDir
    else:
        localPath = "Z:/"

    year = UTCDateTime(t).year
    julday = UTCDateTime(t).julday

    # Specify file
    file = localPath + str(year) + "/" + network + "/" + station + "/" + \
        channel + ".D/*." + str(julday).zfill(3)

    # Define exact start-/end-times (not for daily waveforms)
    if aTrim:   # If auto-trim, select a wider time range
        starttime = t - 12
        endtime = t + 24
    else:       # At least extend slice to remove filter response later
        starttime = t - config.filterBuff
        endtime = t + duration + config.filterBuff

    # Read the file
    try:
        # If whole day is requested, ...
        if (duration == 86400):
            # read the COMPLETE appropriate file, i.e. incl. accounting for
            #   the strange cutting of the MSEED files (not at midnight)
            # TODO : Do it properly and append the remaining waveform data
            #        of that day
            stream = read(file)
        else:
            # otherwise read only the desired snippet
            # The following can lead to wrong selection (sample wise - e.g. one sample too late)
#            stream = read(file, starttime=starttime, endtime=endtime)
            # Bugfix (slower read-in):
            stream = read(file)
            stream.trim(starttime=starttime, endtime=endtime)

        # Print gap info (only for the first channel)
        numGaps = stream.count() - 1
        if numGaps > 0 and stream[0].stats.channel == config.channels[0]:
            print('(has %d gaps)' % numGaps)
            print(stream.__str__(extended=True))  # All traces in stream
#            stream.print_gaps(min_gap=86400/4)    # Big gaps

        # Make sure to get only the data that we need (in case of unrelated traces in stream)
        stream.trim(starttime - 1*3600, endtime + 1*3600)
        stream = stream.select(station=config.station, channel=config.channelSelect)

        # Merge traces of stream (if necessary)
        # Gap filling: 'None' (data as masked arrays)
        # Note: Obspy (by default) does NOT MAINTAIN masks when processing waveform!
        stream.merge(fill_value=None)

        trace = stream[0]  # This could raise an IndexError if stream is empty

#        # Check if it has gaps and merge
#        if stream.getGaps():
#            # ... so join them while filling with zeros
#            for i in range(1, len(stream)):
#                trace = trace.__add__(stream[i], fill_value=0)

    # Exception: If 1. the stream is empty
    #            or 2. the read waveform shorter than expected  (rare case)
    #            or 3. the file doesn't exist
    except (IndexError, Warning, Exception) as err:

        # # 1. For daily waveforms:
        # If file not found (or file empty)
        if (duration == 86400 and isinstance(err, Exception)):
            raise Exception("Daily Waveform not available. ", err)
#        else:
            #TODO: read the other parts, and append

        # # 2. For short waveforms / templates:
        # Maybe it's in another file?
        # (due to strange file saving: files are not split at exactly midnight)
        juldayMod = 0
        midnightTol = 3600  # [s]

        # 1. Check the case of a recent midnight
        if (t - UTCDateTime(t.date))/midnightTol <= 1:
            juldayMod = -1
        # 2. Check the case of an upcoming midnight
        elif (UTCDateTime(t.date)+86400 - t)/midnightTol <= 1:
            juldayMod = 1

        fileNew = localPath + str(year) + "/" + network + "/" + station \
            + "/" + channel + ".D/*." + str(julday+juldayMod).zfill(3)

        # Now, searching for another file only makes sense if close to midnight
        if juldayMod:
            try:
                stream = read(fileNew, starttime=starttime, endtime=endtime)
                traceNew = stream[0]  # This could raise another IndexError

            except IndexError:  # If also this waveform doesn't exist...
                raise Exception("Waveform not available.\n 1.", err,
                                "\n 2. Also not found in the next mseed file.")

            else:
                # If initial error was an incomplete waveform, ...
                if isinstance(err, Warning):
                    # ... add (append/prepend) the trace to the existing one
                    trace = trace.__add__(traceNew, fill_value='interpolate')
                # Otherwise (initially an empty stream or non existing file)...
                else:
                    # ... set the new trace as the new one
                    trace = traceNew
        else:  # If the waveform just doesn't exist in the specified file...
            raise Exception("Waveform not available. ", err)
    else:
        pass

    # Apply signal processing
    # (if procType=='raw' (default) will return same waveform without any processing)
    # Otherwise (if processed) INCLUDES TRIMMING filter response margin on default
    trace = waveform_processing(trace, procType)

    if aTrim:
        # Get absolute time relative to time of peak/maximum (minus 0.5s)
        t = starttime + trace.times()[abs(trace.data).argmax()] - 0.5
        trace.trim(t, t + duration)

    return trace


def get_waveform_SDS(network, station, channel, t, duration, procType='raw'):
    """
    Returns a Trace object (not Stream object) for specified parameters from a
    SeisComP Data Structure (SDS) directory stored locally.
    """

    # Specify SDS client
    client = clientSDS(config.mseedDir, fileborder_seconds=20*60)

    starttime = t - config.filterBuff
    endtime = t + duration + config.filterBuff

    # Form request
    request = [network, station, "*", channel, starttime, endtime]
#    print(request)

    try:
        stream = client.get_waveforms(*request)

        # Print gap info (only for the first channel)
        numGaps = stream.count() - 1
        if numGaps > 0 and stream[0].stats.channel == config.channels[0]:
            print('(has %d gaps)' % numGaps)
            print(stream.__str__(extended=True))  # All traces in stream
#            stream.print_gaps(min_gap=86400/4)    # Big gaps

        # Merge traces of stream (if necessary)
        # Gap filling: 'None' (data as masked arrays)
        # Note: Obspy (by default) does NOT MAINTAIN masks when processing waveform!
        stream.merge(fill_value=None)

        # Get the one (and only) trace object
        trace = stream[0]

    except Exception as err:
        raise Exception("Data could not be read. ", err)

    # Apply signal processing
    # (if procType=='raw' (default) will return same waveform without any processing)
    # Otherwise (if processed) INCLUDES TRIMMING filter response margin on default
    trace = waveform_processing(trace, procType)

    # Make sure to get the appropriate time range
    trace.trim(starttime, endtime)

    return trace


def get_waveform_FDSN(network, station, channel, t, duration, procType='raw'):
    """
    Returns a Trace object (not Stream object) for specified parameters from a FDSN server.
    """

    # Specify FDSN client
    client = clientFDSN(config.FDSN_server)
#    client = clientArcLink(user='sc31007arc', host='arclink.ethz.ch')

    starttime = t - config.filterBuff
    endtime = t + duration + config.filterBuff

    # Form request
    request = [network, station, "*", channel, starttime, endtime]
#    print(request)

    try:
        stream = client.get_waveforms(*request)

        # Print gap info (only for the first channel)
        numGaps = stream.count() - 1
        if numGaps > 0 and stream[0].stats.channel == config.channels[0]:
            print('(has %d gaps)' % numGaps)
#        print(stream.__str__(extended=True))  # All traces in stream

        # Merge traces of stream (if necessary)
        # Gap filling: 'None' (data as masked arrays)
        # Note: Obspy (by default) does NOT MAINTAIN masks when processing waveform!
        stream.merge(fill_value=None)

        # Get the one (and only) trace object
        trace = stream[0]

    except Exception as err:
        raise Exception("Data could not be read. ", err)

    # Apply signal processing
    # (if procType=='raw' (default) will return same waveform without any processing)
    # Otherwise (if processed) INCLUDES TRIMMING filter response margin on default
    trace = waveform_processing(trace, procType)

    # Make sure to get the appropriate time range (FDSN may return more data)
    trace.trim(starttime, endtime)

    return trace


# --- Catalog-related


def readCatalogSEDDB(filename=None, timecol=1, magcol=9):
    """
    Returns a catalog / list of events (incl. Ml)

    You may specify the columns for time and mag if you have a custom SQL
    """

    if filename is None:
        filename = os.path.join(config.dataDir, 'seddb_' + config.region + '.csv')
    else:
        filename = os.path.join(config.dataDir, filename)

    catalog = []
    with open(filename, "r") as file:
        next(file)    # leave out column header
        for line in file:
            if line[0] == "#":  # Remove comment lines
                continue
            catalog.append(line.strip().split(','))

    for idx, entry in enumerate(catalog):

        entry[0] = idx+1  # ID
        entry[1] = UTCDateTime(entry[timecol]) + config.timeTravel
        entry[2] = float(entry[magcol])
        del entry[3:]  # remove everything beyond

    return catalog


def getNicoCatalog(getAll=False, locations=False, useConfigDates=False, getOldVer=False):
    """
    Returns a Basel catalog with relatively located events (incl. ML)
    """

    if getOldVer:
        filename = "../Info_BaselEGS/daNico/relocu_OLD.all"
    else:
        filename = "../Info_BaselEGS/daNico/relocu_reloclin_all.out"

    catalog = []
    with open(filename, "r") as file:
        for line in file:
            if line[0] == "#":  # Remove comment lines
                continue
            catalog.append(line.strip().split())

#    catalog = [[int(x[1]),  # ID (event no)
#                UTCDateTime(" ".join(x[2:7])+" 00") +
#                int(float(x[7]) / 60)*60 + float(x[7]) % 60 + config.timeTravel,
#                float(x[-1])] for x in catalog]

#    catalog = catalog[:-3]
    if locations:
        catalog = [x for x in catalog if not x[9] == '---.---']

    for entry in catalog:

        entry[0] = int(entry[1])  # ID (event no)
        entry[1] = UTCDateTime(" ".join(entry[2:7])+" 00")
        entry[1] += int(float(entry[7]) / 60)*60 + float(entry[7]) % 60 + config.timeTravel
        if not locations:
            entry[2] = float(entry[-1])  # Ml
            del entry[3:]   # remove everything beyond 5th column of list
        else:
            entry[2] = float(entry[9])*1000   # X coord
            entry[3] = float(entry[11])*1000  # Y coord
            entry[4] = float(entry[13])*1000  # Z coord
            entry[5] = float(entry[-1])  # Ml
            del entry[6:]

    # Exclude events that are bad (start from the end!) Index = SED-Nr -(minus) 1
    if not getAll:
        del catalog[170]  # A double event, will bias the magnitude-amplitude determination
        del catalog[65]   # A double event
        del catalog[64]   # A double event, will make it difficult to unify the detection list

    # Trim to time of interest
    if useConfigDates:
        catalog = [x for x in catalog if x[1] >= config.startdate and x[1] < config.enddate]

    return catalog


def getCatalogGEL(GELIDs=["PRE", "Main", "Post", "Late"],
                  onlySED=False, locations=False, useConfigDates=False):
    """
    Returns a catalog / list of events (incl. Ml)
    """

    filename = "../Info_BaselEGS/GEL_DB.dat.new"

    catalog = []
    with open(filename, "r") as file:
        for line in file:
            if line[0] == "#":  # Remove comment lines
                continue
            catalog.append(line.strip().split())

    # Filter / list comprehension: only SED events
    if onlySED:
        catalog = [event for event in catalog if "SED" in event[14]]
    else:
        # Return only events with specified GEL-IDs
        catalog = [event for event in catalog if any(x in event[1] for x in GELIDs)]

    for entry in catalog:

        entry[0] = entry[1]  # GEL-ID
        entry[1] = UTCDateTime(entry[2] + "T" + entry[3])  # Time
        # entry[2] = np.mean([entry[9], entry[11]])  # Mw

        if onlySED:
            Mag = float(entry[13])  # ML
        else:
            Mag = entry[9]  # mw(fq) as string

        if not locations:
            entry[2] = Mag
            del entry[3:]   # remove everything beyond 5th column of list
        else:
            entry[2] = float(entry[4])  # X coord (lon)
            entry[3] = float(entry[5])  # Y coord (lat)
            entry[4] = float(entry[8])  # Z coord
            entry[5] = Mag  # Mag
            del entry[6:]

    # Trim to time of interest
    if useConfigDates:
        catalog = [x for x in catalog if x[1] >= config.startdate and x[1] < config.enddate]

    return catalog


def getEventsFromSC3DB():
    """
    Queries events from a POSTGRES DB using an external SQL file.
    """

    import psycopg2
    import csv

    SQLfile = os.path.join(config.dataDir, 'seddb_%s.sql' % config.region)
    outfile = SQLfile.replace('.sql', '.csv_new')

    SEDsc3db = psycopg2.connect(host="seddba.ethz.ch", dbname="sc3dba",
                                port=5433,
                                user="wwwuser", password="")
    cursor = SEDsc3db.cursor()

    # Execute SQL file
    cursor.execute(open(SQLfile, "r").read())

    colnames = [desc[0] for desc in cursor.description]
#    queryResults = cursor.fetchall()

    # Output
    with open(outfile, 'w') as f:
        writer = csv.writer(f, delimiter=',')
        writer.writerow(colnames)
        for row in cursor:
#            if UTCDateTime(row[1]) >= config.startdate:
            writer.writerow(row)


# --- Custom-file-related


#def readTemplates(filename):
#    """
#    Read identified template file and return as useful array
#    """
#
#    # Make it possible to read from other directories, if filename contains a (relative) path
#    if os.sep not in filename:
#        filename = config.directory + filename
#
#    templateSet = []
#    with open(filename, "r") as file:  # file will be closed automatically
#        for line in file:
#            if line[0] == "#":  # Remove commented events
#                continue
#            templateSet.append(line.strip().split())
#
#    for entry in templateSet:
#
#        # Create ObsPy date format and account for time-"prefix"
#        entry[0] = UTCDateTime(entry[0] + "T" + entry[1]) + float(entry[2])
#        entry[1] = float(entry[3])  # Duration
#        entry[2] = float(entry[4])  # Mag
#        del entry[3:]   # remove everything beyond 5th column of list
#
#    return templateSet


def readTemplateSet(directory=None):
    """
    Reads the list of templates from a file.
    """

    # Default value for directory
    if directory is None:
        directory = config.directory

    files = glob.glob(directory + "/_templateSet.dat")
    if len(files) != 1:
        return None

    templateSet = []
    with open(files[0], "r") as file:
        for line in file:
            templateSet.append(line.strip().split())

    # Reformat each entry
    for template in templateSet:

        template[0] = UTCDateTime(template[0])
        if len(template) == 4:                  # backwards compatible to older templateSet files
            template[1] = float(template[1])    # Shift to template 1
        template[-2] = float(template[-2])      # Duration
        template[-1] = float(template[-1].lstrip("ML"))

    return templateSet


def writeTemplateSet(templateSet, verbose=1):
    """
    Writes the list of templates to a file (incl. time, duration).
    """

    file = config.directory + "/_templateSet.dat"

    with open(file, 'w') as f:
        for template in templateSet:
            if len(template) == 4:
                line = ("%s   % .3f   %.1f   ML%.1f\n" % tuple(template))
            else:
                line = ("%s   %.1f   ML%.1f\n" % tuple(template))
            f.write(line)

    if verbose:
        print("\n[writeDetections()] Wrote %d templates to file %s." % (len(templateSet), file))


def readDetections(filename, threshold=None):
    """
    Reads a list of events from a file.
    """

    # Make it possible to read from other directories, if filename contains a (relative) path
    if os.sep not in filename:
        filename = config.directory + os.sep + filename

    # Use the threshold notation if given
    if threshold is not None:
        filename += "_thresh%.2f" % threshold

    filename += ".dat"

    foundEvents = []
    with open(filename, "r") as file:
        for line in file:
            foundEvents.append(line.strip().split())

    i = 0
    # If it contains a magnitude
    if foundEvents:
        if len(foundEvents[0]) == 5:
            i = 1
    else:
        return []

    # Reformat each entry
    for event in foundEvents:

        event[0] = UTCDateTime(event[0])
        if i == 1:  # If includes magnitude
            event[2] = float(event[2])             # (Magnitude)
        event[2+i] = int(event[2+i].lstrip("tp"))  # Template-number
        try:
            event[1] = float(event[1])             # Amplitude
            event[3+i] = float(event[3+i])         # CCC
        except:
            event[1] = float("NaN")                # Amplitude
            event[3+i] = float("Inf")              # CCC

    return foundEvents


def writeDetections(foundEvents, threshold=None, file='detections',
                    prefMag=False, verbose=1, append=False):
    """
    Writes the list of detections to a file  (incl. time, CCC, amp, and template Nr#.

    You have to make sure to have a reference of associated templates.

    (prefMag): Use the preferred Magnitude only if more than 1 Magnitude provided.
    """

    # Default value for threshold
    if threshold is None:
        threshold = config.threshold

    # Make it possible to write to other directories, if filename contains a (relative) path
    if os.sep not in file:
        file = config.directory + os.sep + file

    baseStr = ['%s', '%9.1f', 'tp%d', '%.6f\n']

    numMags = len(foundEvents[0]) - 4
    if foundEvents:

        # ... if list contains at least one magnitude (the regressed one)
        if (numMags >= 1):
            baseStr.insert(2, '% .2f')
            # Append string to filename
            if 'Mag' not in file:
                file += '_Mag'

        # ... if list contains more than one magnitude
        if (numMags >= 2):
            # Depcopy list b/c it will be modified
            foundEvents = copy.deepcopy(foundEvents)
            # 1. insert another column (for the magnitudes of the catalog)
            baseStr.insert(3, '% .2f')
            # 2. Provide alternative columns if NaN (amp, Mregr, Mcat, templ, CC)
            baseStr_alt = copy.copy(baseStr)
            baseStr_alt[1:] = ['  -----.-', ' -.--', ' -.--', 'tp0', '-.------\n']

        # ... if list even contains more than two magnitudes (special GEL case)
        if (numMags == 4):

            # insert string columns for GEL magnitudes
            baseStr.insert(3, '% .2f')
            baseStr.insert(3, '% .2f')
            baseStr_alt.insert(3, ' -.--')
            baseStr_alt.insert(3, ' -.--')

    file += "_thresh%.2f" % threshold + ".dat"

    if append:
        writeMode = 'a'
    else:
        writeMode = 'w'

    numLines = 0
    with open(file, writeMode) as f:
        for event in foundEvents:
            if event[-1] >= threshold:
                # When other mags are present and only the best magnitude desired ...
                if prefMag and numMags > 1:
                    # In case a GEL mag is available, take them (the ML conversion)
                    if numMags == 4 and not np.isnan(event[4]):
                        event[2] = event[4]
                    # But always prefer SED ML if available
                    if not np.isnan(event[-3]):
                        event[2] = event[-3]
                # Format all fields directly when other magnitudes are present
                if numMags > 1:
                    for i in range(1, numMags+4):
                        if np.isnan(event[i]) or np.isinf(event[i]):
                            # Process (remaining) NaN fields
                            event[i] = baseStr_alt[i]
                        else:
                            # Directly format non-NaN fields
                            event[i] = baseStr[i] % event[i]
                    # Also convert time to string
                    event[0] = baseStr[0] % event[0]
                    if prefMag:
                        line = '   '.join(event[0:3] + event[-2:])
                    else:
                        line = '   '.join(event)
                else:
                    line = '   '.join(baseStr) % tuple(event)
                f.write(line)
#                ("{0:s}   {2:9.1f}   tp{3:d}   {1:.6f}".format(tuple(event)))
                numLines += 1

    if verbose:
        print("\n[writeDetections()] Wrote %d detections to file %s." % (numLines, file))


def writeToFile(file, eventlist, lineFormat):
    """
    Write anything to file given the list and line formatting.
    """

    with open(file, 'w') as f:
        for event in eventlist:
            f.write(lineFormat % tuple(event))


# --- Waveform-snippet-related


def preload_events(catalog, readALLagain=False, plot=True):
    """
    Loads the waveforms of the events (e.g. potential templates) into MSEEDs to save time.
    # Initial loading should be using 3 COMPONENTS!

    Determines already available files and retrieves only new ones.
    """

    print("\n[preload_events()] (of %d) ---------------------" % len(catalog), flush=True)

    numSeedsAvail = 0
    outDir = os.path.join(config.dataDir,
                          'SED-events_%s_%s_mseed' % (config.region, config.station))
    if not os.path.exists(outDir):
        os.makedirs(outDir, exist_ok=True)
    elif not readALLagain:  # (Then TRY to load ONLY NEW events)
        # Check if there are new, yet unloaded, events
        mseed_list = glob.glob("%s/*.mseed" % outDir)
        if len(mseed_list) > 0:
            # Get last read event
            lastSeedTime = UTCDateTime(
                sorted(mseed_list)[-1]
                .split('/')[-1]
                .split('_')[-1]
                .split('.mseed')[0])
            # Which events have been read already?
            numSeedsAvail = sum([x[1] <= lastSeedTime for x in catalog])
            # Get the remaining events
            catalog = catalog[numSeedsAvail:]

    if not catalog:
        print("All MSEEDs already exist.")
        return False

    # Do the following only if there are still remaining events to be read

    if numSeedsAvail > 0:
        print("%d catalog events exist already. Reading in the remaining %d" %
              (numSeedsAvail, len(catalog)), flush=True)

    for event in catalog:

        # Read waveforms (of each channel)
        try:
            if config.use_threads:
                # Read files in parallel using threads
                with ThreadPool(3) as pool:
                    waveformStream = Stream(pool.starmap(
                        get_waveform,
                        zip(repeat(config.network), repeat(config.station), config.channels,
                            repeat(event[1]), repeat(config.duration))))
            else:
                # Serial read in
                waveformStream = Stream()
                for chnl in config.channels:
                    waveformStream += get_waveform(
                        config.network, config.station, chnl, event[1], config.duration)

            # Save the event waveform (unfiltered) for later fast retrieval
            waveformStream.write(outDir + "/%03d_" % event[0] +
                                 str(event[1]).replace(":", "-") + ".mseed",
                                 format='MSEED')  # , encoding="FLOAT32")
            if plot:
                if 'fig' not in locals():
                    # Create a matplotlib figure instance to prevent a memory leak
                    # of matplotlib. Also, determine default figure size
                    fig = read()[:waveformStream.count()].plot(handle=True)

                fig.clear()  # Reuse the same figure

                waveform_processing(waveformStream, 'none', None).plot(
                    outfile=outDir + "/%03d_" % event[0] +
                    str(event[1]).replace(":", "-") +
                    "_ML%.1f" % event[-1] + ".png", fig=fig)

        # If ANY channel cannot be read...
        except Exception as err:
            print("Cannot read waveform of catalog event %d." % event[0], err)
            continue  # ... and ignore it
        else:
            pass

    if plot:
        plt.close("all")

    print("Loaded %d new MSEEDs." % len(catalog))
    return True


def read_chunk(chunkId):
    """
    Read and return a chunk of waveform(s) previously stored to disk with
    preload_chunk_from_archive().

    Return the chunk, or None if it doesn't exist
    """

    filename = os.path.join(config.chunkCacheDir, "chunk_%s" % chunkId)  # +
#                           str(chunkTime).replace(":", "-") + "_%d" % chunkLength)

    # 1. the (processed) waveform data (in float32 dtype/encoding)
    try:
        chunkStream = read(filename + ".mseed")
    except:
        return None, None

    # 2. the windowed standard deviation arrays (possibly incl. masks)
    try:
        stdWindowed = np.load(filename + "_extra.npy")  # is now an array of arrays
    except:
        return chunkStream, None  # Will have to calculate it later...

    return chunkStream, stdWindowed


def get_ready_chunk(chunkTime, chunkLength):
    """
    Read a chunk of waveform(s) from the archive, process it accordingly,
    and make it ready to use for xcorr (incl. calculation of windowed standard deviation).

    Returns:
        - the loaded (processed) chunk
        - the loaded (unprocessed) chunk - e.g., for later cutting out original data
        - the windowed standard deviation of each trace (possibly incl. mask)
    """

    try:
        if config.use_threads:
            # Read chunks in parallel using threads
            with ThreadPool(config.numChannels) as pool:
                chunkStream = Stream(pool.starmap(
                    get_waveform, zip(repeat(config.network),
                                      repeat(config.station),
                                      config.channels,
                                      repeat(chunkTime),
                                      repeat(chunkLength))))
        else:
            chunkStream = Stream()
            # Read chunks serially
            for idx, channel in enumerate(config.channels):
                # Create a stream object containing all components/traces
                chunkStream += get_waveform(
                    config.network, config.station, channel, chunkTime, chunkLength)

        # Determine starts & /ends of all traces
        starts = [x.stats.starttime for x in chunkStream]
        ends = [x.stats.endtime for x in chunkStream]
        if min(ends) <= max(starts):
            raise Exception("No common data among the traces. Skip it.")

    except Exception:
        raise  # To pass the exception to the calling function
    else:
        pass

    # Trim the stream to the max start/end among all traces
    chunkStream.trim(max(starts), min(ends))
#    # Trim rest to 0 (get most out of it)
#    dailyStream.trim(min(starts), max(ends), pad=True, fill_value=0)

    # Create a bandpassed stream (perhaps resampled)
    # (not necessarily parallel - depends on config.parallelFilter)
    chunkStream_BP = waveform_processing_parallel(chunkStream, 'bandpass')  # (keeps original)

    # Clean up FFT cache originating from FFT resampling
    # In this way we keep more memory free for xcorr :-)
    if chunkStream[0].stats.sampling_rate != chunkStream_BP[0].stats.sampling_rate:
        # (Workaround for numpy issue where many FFTs of various lengths fill
        #  a cache that never gets cleared, effectively creating a memory leak)
        limit_numpy_fft_cache()
#        # Alternatively clear scipy's cache (if used for resampling)
#        scipy.fftpack._fftpack.destroy_rfft_cache

    # Trim again (since some short data chunks may have been removed)
    starts = [x.stats.starttime for x in chunkStream_BP]
    ends = [x.stats.endtime for x in chunkStream_BP]
    chunkStream_BP.trim(max(starts), min(ends))

    for idx, tr in enumerate(chunkStream_BP):
        # Check if there are "unnecessary" masks == all mask entries are False
        # This might happen when a former gappy trace was trimmed and now
        # doesn't contain gaps anymore.
        if np.ma.isMaskedArray(tr.data) and np.ma.is_masked(tr.data) is False:
            print("You found a trace (%s) that has gaps where others have no data "
                  "at all. No big deal; the gap(s) were cut away anyway." %
                  tr.stats.channel)
            tr.data = np.asarray(tr.data)  # removes mask

    # Calculates the moving std window (required for normalizing xCorrTrace)
    # ASSUMPTION: if using several templates, they all need to be the SAME LENGTH
    # Note: in the presence of gaps, stdWindowed also provides an (extended) mask [per trace])
    #       (to be able to remove 'untrustworthy' calculated valus at the borders of gaps)
    if config.use_threads:
        with ThreadPool(config.numChannels) as pool:
            stdWindowed = pool.map(calc_windowed_std_trace, chunkStream_BP)
    else:
        stdWindowed = []
        for tr in chunkStream_BP:
            stdWindowed.append(calc_windowed_std_trace(tr))

    return chunkStream, chunkStream_BP, stdWindowed


def delete_chunk(chunkId):
    """
    Delete preloaded chunk waveform(s) from disk
    """
    basefilename = os.path.join(config.chunkCacheDir, "chunk_%s" % chunkId)
    try:
        os.remove(basefilename + ".mseed")
        os.remove(basefilename + "_extra.npy")
    except:
        pass


def preload_chunk(chunkId, chunkTime, chunkLength):
    """
    Read a chunk of waveform(s) from the archive, process it accordingly,
    and store it to disk (together with the windowed standard deviation
    arrays) for a subsequent fast loading (see read_chunk()).

    If the file(s) already exists, skip the retrieval but return the chunk
    anyway.

    The extra arrays (stdWindowed) may also contain/carry masks.

    Return the pre-loaded (and processed) chunk and the windowed standard
    deviation (and mask), or None if it wasn't possible to load the chunk
    """

    # First, try to load the chunk + extra file, maybe they exists already
    chunkStream, stdWindowed = read_chunk(chunkId)
    if chunkStream is not None and stdWindowed is not None:
        return chunkStream, stdWindowed

    # Get processed chunk from archive and calculate windowed std
    try:
        chunkStream, stdWindowed = \
            get_ready_chunk(chunkTime, chunkLength)[1:]
    except:
        return None, None

    filename = os.path.join(config.chunkCacheDir, "chunk_%s" % chunkId)

    # Save to disk
    # 1. the waveform data (now processed, not raw)
    for tr in chunkStream:
        # Ensure float32 dtype/encoding & remove possible mask
        tr.data = np.asarray(tr.data, dtype=np.float32)

    chunkStream.write(filename + ".mseed", format='MSEED', encoding='FLOAT32')

    # 2. the windowed standard deviation arrays (possibly incl. masks)
    np.save(filename + '_extra.npy', stdWindowed)
    # FYI: np.savez() and np.savez_compressed makes writing and loading NOT faster, but slower
    # FYI: a *list* of np.arrays will become an *np.array* of np.arrays

    return chunkStream, stdWindowed


def read_template(templateId, verbose=1):
    """
    Read and return a template waveform(s) previously stored to disk with
    preload_template_from_archive().

    Return the template waveform or None if the file doesn't exist
    """

    # Check if template already exists
    try:
        templateStream = read(config.templateDir + "/templates_mseed" +
                              "/template%02d_processed.mseed" % (templateId+1))
        # Select desired channels
        # (can also account for changing channel names of a station)
        streamChannels = [templateStream[x].stats.channel
                          for x in range(len(templateStream))]
        channel_idx_want = getChannelIndex(config.channels, streamChannels)

    except Warning as err:
        print("Not enough channels present in preloaded template %d." %
              (templateId+1), err.args)
        return None

    except FileNotFoundError as err:
        if verbose:
            print("Cannot use template %d. It was not preloaded." %
                  (templateId+1), err.args)
        return None

    # (Actually any exception is so fatal that TM should exit and let
    #  the user explore what's the cause)

    else:
        return Stream([templateStream[i] for i in channel_idx_want])


def delete_template(templateId):
    """
    Delete preloaded template waveform(s) from disk
    """
    filename = config.templateDir + "/templates_mseed" + \
        "/template%02d_processed.mseed" % (templateId+1)
    try:
        os.remove(filename)
    except:
        pass


def preload_template(templateId, templateTime, templateLength, magnitude):
    """
    Read a template waveform from the archive and store it to disk for a
    subsequent fast loading (see read_template()).

    If the file already exists, skip the retrieval but return the template
    anyway.

    It may apply a template shift to align it to the first template.

    Return 1. the preloaded template, or None if loading wasn't possible,
       and 2. the templateShift
    """

    # First, try to load the event, maybe it exists already
    templateStream = read_template(templateId, verbose=0)
    if templateStream is not None:
        return templateStream, None

    # Prepare
    if config.templateShiftCorr and templateId > 0:
        # The buffer to give for aligning template waveform to template 1
        buff = 1.0  # [s]
    else:
        buff = 0.0

    try:

        if config.use_threads:
            # Read files in parallel using threads
            with ThreadPool(config.numChannels) as pool:
                templateStream = Stream(pool.starmap(
                    get_waveform,
                    zip(repeat(config.network), repeat(config.station), config.channels,
                        repeat(templateTime-buff), repeat(templateLength+2*buff))))
        else:
            # Read in serially
            templateStream = Stream()
            for channel in config.channels:
                # Create a stream object containing all components/traces
                templateStream += get_waveform(config.network, config.station,
                                               channel, templateTime-buff,
                                               templateLength+2*buff)

    except:
        print("Cannot read waveform of template %d (or any component)." %
              (templateId+1))
        return None, None

    # If template waveform aligning to first template desired ...
    if config.templateShiftCorr and templateId > 0:

        # Read first template again
        template1_stream = read(config.templateDir +
                                "/templates_mseed/template01_*_raw.mseed")

        # Get time shift (account for additional buffer)
        templateShift = getTimeShiftOfMaxSimilarity(template1_stream,
                                                    templateStream) - buff
        # The shift should not be > buffer
        if abs(templateShift) > buff:
            print("The shift to the first template is larger than 1s."
                  "Unreasonable. Set to -0.0s.")
            templateShift = -0.0  # The minus denotes this error

        # Trim template accordingly (account for buffer!)
        templateStream.trim(templateStream[0].stats.starttime + templateShift + buff,
                            templateStream[0].stats.endtime + templateShift - buff)
    else:
        templateShift = 0.0

    # Ensure accurate (sample-)length of traces
    # (should the template ever differ by one sample in rare cases compared to
    # 'theoretical' (desired) length)
    samps_expected = int((config.duration + 2*config.filterBuff) *
                         templateStream[0].stats.sampling_rate + 1)
    if not all(samps_expected == len(x.data) for x in templateStream):

        print("Some traces of template %d don't have the expected length"
              " of %d samples" % (templateId+1, samps_expected))

        for trace in templateStream:

            samps = len(trace.data)
            if np.isclose(abs(samps_expected-samps), 1):

                if len(trace.data) > samps_expected:
                    # Remove last sample
                    trace.data = trace.data[:samps_expected]
                elif len(trace.data) < samps_expected:
                    # Duplicate last sample
                    trace.data = np.append(trace.data, trace.data[-1])
                print("Fixed length of %s by one sample (%d -> %d)." %
                      (trace.stats.channel, samps, samps_expected))

            else:
                print("Couldn't fix it. Sample difference is quite a lot."
                      " Please inspect yourself.")
                return None, None

    # .. save original template waveform
    templateStream.write(
        config.templateDir + "/templates_mseed/template%02d_%s_raw.mseed" %
        (templateId+1, str(templateTime).replace(":", "-")), format='MSEED')

    # Process template waveform
    waveform_processing(templateStream, 'bandpass', copy=False)

    # .. save processed template waveform (for later faster xcorr)
    templateStream.write(
        config.templateDir + "/templates_mseed/template%02d_processed.mseed" %
        (templateId+1), format='MSEED')

    # ... plot processed template
    templateStream.plot(
        outfile=config.directory + "/_template%02d_" % (templateId+1) +
        str(templateTime).replace(":", "-") +
        "_ML%.1f" % magnitude + ".png")

    # Close figure instance to not show them in the IPython console
    plt.close('all')

    return templateStream, templateShift
