# -*- coding: utf-8 -*-
"""
Program for finding low-magnitude earthquakes via template matching

A template is defined by
    - the recording station (and its network) on a specific channel
    - the starttime
    - the duration

(created) 2015, Marcus Herrmann, Swiss Seismological Service, ETH Zürich
"""

# --- Import modules

# General
from sys import exit
import time
import copy    # e.g. to make deepcopies of nested lists
import psutil  # Memory utilities
import os      # OS specific

# Scientific
import numpy as np
#import statsmodels.api as sm
#import scipy.fftpack

# ObsPy
from obspy import UTCDateTime

# Own modules
#import TM.setupclass                     # Setup class (this also makes an instance)
from TM.setupclass import config         # This instance is 'config' and imported here
from TM.core import do_xcorr_onetemplate
from TM import io
#from TM.waveformops import waveform_processing
from TM.utils import limit_numpy_fft_cache


def do_xcorr_oneday(iDate, scannedDays, templateSet, catalog=None):
    """
    xCorr parallelization level 1: Performs cross-correlation at one day.

    Makes use of level 2: do_xcorr_onetemplate --> multi-template xcorr
    """

    print("xCorr for %s (%s, %s)" % (iDate.date, config.station, "+".join(config.channels)))

    # Determine the template not yet used for that day
    # Works okay for only one active 'oneday' process
    iTemplate = scanned_days_manager('checkday', scannedDays, iDate, templateSet)
    if iTemplate is None:
        # Day is already complete. No need to scan again
        return [], templateSet

    # Create a matplotlib figure instance to prevent memory leaks
#    fig = read()[:config.numChannels].plot(handle=True)  # To have to default figure size

    # Derive daily waveform (unprocessed and processed), and the
    # windowed standard deviation arrays (which may also contain/carry masks)
    print("Loading waveform(s)...", flush=True)
    try:
        dailyStream, dailyStream_processed, stdWindowed = \
            io.get_ready_chunk(iDate, 86400)
    except Exception as err:
        print("Waveform data input error.", err)
        # Declare scanned day as '-'
        scannedDays = scanned_days_manager('update', scannedDays, iDate, '-')
        return [], templateSet

    foundEvents = []

    # As long as there are still unused templates in the set, ...
    while (iTemplate < len(templateSet)):

        print("   ## Taking Template #%02d" % (iTemplate+1), flush=True)

        # Get the current template's waveform (already processed)
        templateStream = io.read_template(iTemplate)
        if templateStream is None:
            print("Exiting.\n")  # The cause was printed by read_template()
            exit()

        # Trim to agree with actual template duration (only if shorter than pre-loaded)
#        templateStream.trim(endtime=templateStream[0].stats.starttime +
#                            int(config.duration * config.resampling + 1))

        # Do cross-correlation (multichannel)
        xCorrResults = do_xcorr_onetemplate(dailyStream_processed, templateStream,
                                            stdWindowed)

        # If an event (or more) has/have been found ...
        if xCorrResults:

            # To support saving detections...
            for idx, tr in enumerate(dailyStream):
                # The following test is stricter than np.ma.is_masked(tr.data)
                if np.ma.isMaskedArray(tr.data):
                    tr.data = np.asarray(tr.data)  # removes mask

            # TODO: outsource to run in background as worker that receives these events
            # TODO: If event-rate is too high, don't write out bazillion files, but the whole day
            for detection in xCorrResults:

                # ... insert its templateIndex before the last column (CCC)
                detection.insert(2, iTemplate+1)

                t = detection[0]
                dur = (templateStream[0].count()-1) / templateStream[0].stats.sampling_rate

                # ... get the raw (unfiltered) waveform incl. 1s more at the start and end
                #     to allow later remove of the filter response
                detectionStream_RAW = dailyStream.slice(t - config.filterBuff,
                                                        t + config.filterBuff + dur)

                # .. and save detection waveform
                detectionStream_RAW.write(
                    config.TMdata + config.directory.replace("_pass2", "") + "/detections_mseed/" +
                    str(detection[0]).replace(":", "-") +
#                    "_%.2f" % detection[3] + "_Template%02d" % (iTemplate+1) +
                    ".mseed",
                    format='MSEED')  # , encoding="FLOAT32")

                # ... and also plot it now (to see progress while running)
                # ATTENTION: this slows down the template matching quite considerably
                #            (e.g. 20'000 [3comp] take ~5 hours)
#                fig.clear()
#                waveform_processing(detectionWaveform_RAW, "bandpass").plot(
#                    outfile=TMdata_now + "/detections_plots/detect_" +
#                    str(detection[0]).replace(":", "-") +
#                    "_%.2f" % detection[3] +
#                    "_Template%02d" % (iTemplate+1) + ".png", fig=fig)
#                # Alternative: close matplotlib figure
#                #plt.close(fig)

            # Extend the daily detection list
            foundEvents.extend(xCorrResults)

        # Increase template index to continue with the next template in the set
        iTemplate += 1

        # If end of templateSet reached, and a catalog provided
        # check if templateSet can be updated with the an UNDETECTED event from catalog
        # TODO: outsource to new function?
        if (iTemplate == len(templateSet)) and catalog:
            # (!) For this, only take the detections that have a high CCC
            #   --> ignore detections with a low CCC
            #      --> try to find a possibly better event
            #   (a higher value leads to more templates)
            foundEvents_times_highCCC = [x for x in foundEvents
                                         if x[3] >= config.threshNewTemplate]
            # Get indices of (possibly) yet undetected events in catalog
            # Take the time of the dailyTrace
            #   (which can differ from "midnight-to-midnight")

            missed = missingEvents(foundEvents_times_highCCC, catalog,
                                   dailyStream[0].stats.starttime,
                                   dailyStream[0].stats.endtime)
            # If a catalog event was missed (or more), ...
            if missed:
                # ... append only the the first missed catalog event
                #  --> update templateSet step-by-step
                # TODO: Take the one missing event that is most dissimilar
                #        from the current templates
                templateSet.append([missed[0][0], 0.0, config.duration, missed[0][1]])
                print("--- Updated template set ---")
                print(templateSet[-1])
                # Preload this template
                templateSet = preload_templates(templateSet, iTemplate)
            elif missed is []:
                print("   All catalog events detected with template matching.")
            else:  # (if missed is None)
                pass

    # Keep track of the scanned days + number of used templates
    scannedDays = scanned_days_manager('update', scannedDays, iDate, templateSet)

    if config.conv_method == 'fftconvolve':
        # Finally clean up numpy's FFT cache (if over limit)
        # (Workaround for numpy issue where many FFTs of various lengths fill
        #  a cache that never gets cleared, effectively creating a memory leak)
        limit_numpy_fft_cache()

    # Close matplotlib figure
#    plt.close(fig)

    return foundEvents, templateSet


def template_matching(templateSet=None, catalog=None):
    """
    The function that sets up the Template Matching. This includes
        template selection mode, directory setup, organizing to continue a previous scan, etc.

    Input:
        'templateSet' - a list of eventtimes & durations
        'catalog' - a reference event list.
            If provided, the TM routine will automatically update the
            template set with events that have NOT yet been detected: Unless all
            catalog events have been detected for a certain day, TM is repeated on
            the day with a newly included catalog event (updates template set).

      Based on the provided input, determine the template selection mode (auto or manual)

    Returns:
      1) A list of found events with
        - Event time
        - Amplitude
        - Cross-correlation coefficient
        - Template identification.
      2) The templateSet list with
        - time
        - possible shift of templates to the first template
        - duration
        - Magnitude (from catalog)
    """

    # Check if all required parameters are set
    print("Checking config class...")
    config.all_set_check()

    # Determine template selection mode
    # 1. Manual mode - when specifying a templateSet, but no catalog
    #   For example a selection from clustered catalog events, or simply ALL known catalog events
    if templateSet and not catalog:

        tpSelMode = 'manual'
        print("Template selection mode: 'manual': predefined templates.\n")
        nextpass = False

    elif catalog:

        # 2. 'Auto' mode
        #    This is useful for real-time applications: when we don't know what follows and
        #    have to selectively choose template events from the catalog in an automatic fashion.
        #    (of course only if we don't want to use ALL events)
        if not templateSet:
            tpSelMode = 'auto'
            print("Template selection mode: 'auto': "
                  "taking templates automatically from catalog.\n")

        # 3. 'manual' & 'auto' mode
        else:
            tpSelMode = 'man+auto'
            print("Template selection mode: 'man+auto': "
                  "specified templates plus taking templates automatically from catalog.\n")

        # Three parameters required
        #   templateMagRange - magntiude range of catalog events to consider as templates
        #   firstTemplateMinMag - minimum magnitude of first template
        #   threshNewTemplate - Catalog events detected with CC smaller than this will
        #                       become a new template
        missingVals = []
        for var in ['templateMagRange', 'firstTemplateMinMag', 'threshNewTemplate']:
            if str(var) not in vars(config):
                missingVals += [str(var)]

        if missingVals:
            raise ValueError("Not all parameters for 'auto' mode were set. Missing \n - %s" %
                             '\n - '.join(missingVals))

        # Adjust canditate catalog list
        #  1. If first column is not the time, remove it (most likely an event ID)
        if not isinstance(catalog[0], UTCDateTime):
            catalog = [x[1:] for x in catalog]
        #  2. Use only events in given range
        catalog = [x for x in catalog
                   if config.templateMagRange[0] <= x[1] <= config.templateMagRange[1]]
        if not catalog:
            raise Warning("Couldn't find any candidate events that suffice"
                          "'config.templateMagRange' = %.1f" % config.templateMagRange)

    else:
        raise Warning("Cannot start template matching without template information."
                      "You may provide a template set and/or a catalog.")

    # If custom templates specified, make sure it's a list of list
    if 'man' in tpSelMode and not isinstance(templateSet[0], list):
            templateSet = [templateSet]

    foundEvents = []
    scannedDays = []

    print("Checking directory...")
    # Check if this run was active before
    if os.path.exists(config.directory):

        print("    Directory already exists. Continue with Template Matching where we left off.\n")

        # 1. Load the templates from last run
        templateSet_fromDir = io.readTemplateSet(config.directory)  # This one can have new entries
#        print('ATTENTION: Using the templateSet + waveforms from the directory...')

        if templateSet and templateSet != templateSet_fromDir:
            raise Warning("Sorry, specifying a template set different from the existing run "
                          "is not considered to be safe. You might implement it? ;-)")
            # TODO: implement: add new templates into templateSet
        elif not templateSet:
            templateSet = templateSet_fromDir

        # 2. load progress from last run
        if os.path.isfile(config.directory + "/progress_thresh%.2f.dat" % config.threshold):
            foundEvents = io.readDetections("progress", config.threshold)

        # 3. read already scanned days (incl. last used =highest template)
        scannedDays = scanned_days_manager('read')

    else:  # Otherwise, create everything necessary

        # Create main and template_mseed dir
        os.makedirs(config.directory + "/templates_mseed", exist_ok=True)

        # Write config to file
        with open(config.directory + "/config.txt", 'w') as f:
            for (key, val) in sorted(vars(config).items()):
                f.write("%s = %s\n" % (key, repr(val)))

        print("    Directory created. Start with a new Template Matching.\n")

    # Check if a reference eventlist was specified (but not a templateSet)
    if tpSelMode == 'auto' and not templateSet:

        # Let the first template be large enough
        firstEvent = [x for x in catalog if x[1] >= config.firstTemplateMinMag][0]
        if firstEvent is []:
            raise Warning("Couldn't find a template that suffices "
                          "'config.firstTemplateMinMag'")
        print("Taking %s (M%.1f) as the initial template." % (firstEvent[0], firstEvent[1]))

        templateSet = [[firstEvent[0], 0.0, config.duration, firstEvent[1]]]  # (list of list)

    # Check for the "big-data folder, but don't create a pass2 folder there
    TMdata_now = config.TMdata + config.directory.replace("_pass2", "")
    if not os.path.exists(TMdata_now + "/detections_mseed"):
        os.makedirs(TMdata_now + "/detections_mseed", exist_ok=True)

    # Preload the templates (if in 'auto' mode, only the one and only)
    templateSet = preload_templates(templateSet)

    # Stop the time
    tX = time.time()

    # Go through time...
    iDate = copy.copy(config.startdate)  # Start with the first day

    # If templates have to be selected automatically (over time)
    if 'auto' in tpSelMode:
        # Determine last scanned day
        if scannedDays:
            iDate = scannedDays[-1][0]
            print("Already scanned until %s. Continuing with the next day.\n" % iDate.date)
            # Continue scanning with the next day
            iDate += 86400
        # (Otherwise it is the config.startdate)

        # Do day by day
        # (Cannot be parallelized since the templateSet will change over time / is not constant)
#        numOfDays = int(math.ceil((config.enddate-config.startdate)/86400))
#        for iDate in (config.startdate + day*86400 for day in range(numOfDays)):
        while iDate < config.enddate:

            # Doing actual Template Matching
            # TODO: outsource template updater (and call only if 'auto')
            xCorrResults, templateSet = do_xcorr_oneday(iDate, scannedDays, templateSet, catalog)

            # Extend the overall detection list
            foundEvents.extend(xCorrResults)

            # At the "end of the day"...
            # 1. Write out ALL detections (only if anything was found on that day)
            if xCorrResults:
                io.writeDetections(sorted(foundEvents), file="progress", verbose=0)
                # TODO: partially update (append) the progress_file
                #       (a problem during the second pass)

            # 2. Make sure to get to the next day in next iteration
            iDate += 86400  # Go to next day

        # Determine if first pass finished (events from catalog detected and used as templates)
        #      Optional: check if reached enddate and
        #                if current templateSet differs from the one used at the beginning
#        if iDate == config.enddate and int(scannedDays[0][1]) < len(templateSet):
        print("First run is finished. Revisit whole period with newer templates again...")
        nextpass = True
        # And start from beginning again
        iDate = copy.copy(config.startdate)  # Copy! Otherwise config variable will be touched
        # --> now re-scan from beginning (using later templates not used at the beginning)
        # This second pass can be done in parallel day-wise

    # We have all templates, go through the whole period once (more)
    # (perfectly parallelizable)
    # Here: a serial solution
    # If you want a parallized version, use the MPI approach
    if tpSelMode == 'manual' or nextpass:

        while iDate < config.enddate:
            xCorrResults, _ = do_xcorr_oneday(iDate, scannedDays, templateSet)
            if xCorrResults:
                foundEvents.extend(xCorrResults)
                io.writeDetections(sorted(foundEvents), file="progress", verbose=0)
            iDate += 86400

    print("Elapsed time: %.3fs" % (time.time() - tX))

    # Return sorted (otherwise: ordered daily by templates)
    return sorted(foundEvents), templateSet


def scanned_days_manager(operation, scannedDays=None, date=None, templateSet='-'):
    """
    Helper to maintain information about scanned days
    and the number of templates used for that day.

    TODO: implement as class
    TODO: make capable of being accessed and updated by several processes (using a queue?)
    """

    if operation == 'read':
        if os.path.isfile(config.directory + "/scannedDays.dat"):
            with open(config.directory + "/scannedDays.dat", "r") as f:
                scannedDays = [[UTCDateTime(line.split()[0]), str(line.split()[1])] for line in f]
        # TODO: 3.1 If last scanned day was not completely scanned (24h not available back then)
        # for now: delete last day in 'scannedDays.dat' and perhaps in 'progress_thresh*.dat'
#        if (scannedDays[-1][0] == ???):
#            del scannedDays[-1]  # ... and delete the day from the "index" (i.e., redo that day)
            # (Attention: there will be double entries in the foundEvents list of the last day,
            #  but they will be later removed with (uniqueEventList())
            # Remember, not all mseeds start exectly at 00:00:00.0!
#            foundEvents = [x for x in foundEvents if x[0] <= iDate]
            return scannedDays
        else:
            return []
    elif operation == 'checkday':
        try:
            # Try to access that day
            scannedDayIndex = [x[0] for x in scannedDays].index(date)
    #        scannedDayIndex = [i for i, x in enumerate(scannedDays) if x[0] == iDate]
        except:
            # If this day is yet unscanned...
            scannedDayIndex = len(scannedDays)
            scannedDays.append([date, 0])
            # Start with template #1
            iTemplate = 0
        else:
            try:
                # For scanned days, start with the next unscanned template
                # == the python-index of the next template
                # Try to convert to an integer
                iTemplate = int(scannedDays[scannedDayIndex][1])
            except:  # (if the value is a string)
                if (scannedDays[scannedDayIndex][1] == '-'):
                    print("  Day could not be read before... Now?")
                    iTemplate = 0
            else:
                # But if the day was already scanned with all templates, go to next day
                # (should quickly go until the enddate)
                if (iTemplate == len(templateSet)):
                    print("  Day was already scanned with all templates.")
#                    date += 86400  # Go to next day
#                    continue
                    return None
                else:
                    print("  Day was already scanned with %d templates. Do %d more."
                          % (iTemplate, len(templateSet)-iTemplate))
        return iTemplate

    elif operation == 'update':
        # Get the index of  that day (this day entry should exist!)
        scannedDayIndex = [x[0] for x in scannedDays].index(date)
        # Update
        if templateSet == '-':
            char = '-'
        else:
            char = str(len(templateSet))
        scannedDays[scannedDayIndex][1] = char
        # Write out the scanned days + numer of templates used so far
        with open(config.directory + "/scannedDays.dat", 'w') as f:
            for scannedDay in scannedDays:
                f.write("%s   %s\n" % (scannedDay[0].date, scannedDay[1]))
        return scannedDays


def missingEvents(eventList, referenceList, starttime, endtime):
    """
    Returns a list of the 'referenceList's events that are missing in 'eventList'
    (referring to the time and some tolerance).

    Checks only between 'start-' and 'endtime'. The first column has to be the eventtime.
    """

    # Define time tolerance
    # If too small, perhaps catalog events count as missed altough they were detected
    #    (b/c they may have a large time shift)
    # If too large, perhaps catalog events count as detected altough only an other
    #    event is within that timeTol that is NOT THAT catalog event
    timeTol = 1.0

    # Make sure the reference list is a list of list
    # (e.g. if reference list is only single list of times)
    if not isinstance(referenceList[0], list):
        referenceList = [[x] for x in referenceList]

    # Make sure the two lists are sorted
    eventList.sort()
    referenceList.sort()

    missed = []
    # Get the times, between 'start-' and 'endtime'
    times = [x[0] for x in eventList
             if (x[0] >= starttime and x[0] < endtime)]
    refList_period = [x for x in referenceList
                      if (x[0] >= starttime and x[0] < endtime)]

    if not refList_period:
        return None

    idx = 0
    # For each event in the reference list, ...
    for i in range(len(refList_period)):
        # If eventList is empty, it's definitely missed
        if len(times) == 0:
            missed.append(refList_period[i])
        else:
            # Otherwise, go through the eventList, and ...
            # (btw: don't check everytime from the beginning - lists are sorted)
            for j in range(idx, len(times)):
                #  ... check if reference event is in the event list
                # If found, ...
                if (abs(refList_period[i][0] - times[j]) <= timeTol):
                    # ... update index, to not start at the beginning again
                    idx = j
                    break  # Continue with next event in reference list
                # Otherwise, if at end (i.e. nothing found), ...
                elif (j == len(times)-1):
                    # ... count as missed reference event
                    missed.append(refList_period[i])

    # Return list
    return missed


def preload_templates(templateSet, templateId=0):
    """
    Preloads template waveform(s).
    ... and may apply a template shift to align them to the first. --> returns templateSet

    If 'templateId' specified, load only that (most useful when templateSet was updated),
    otherwise (e.g., if '0') check all and load the ones not yet existing (though it will
    only work for missing templates that are "greater" than the last existing template!).

    When finished, writes templateSet to file.
    """

    if not templateSet:  # Nothing to preload
        raise Warning("No templates specified.")

    if templateId == 0:
        print("Checking preloaded templates...")

    if not isinstance(templateSet[0], list):
        templateSet = [templateSet]  # Pack into a list

    # Determine number of preloaded templates
    templateDir = config.templateDir + '/templates_mseed'
    if os.path.exists(templateDir):
        preloadedTemplates = len(os.listdir(templateDir))//2  # (we have raw and processed)
        if templateId == 0:
            if len(templateSet) != preloadedTemplates:
                print("   %d new templates found in the templateSet file.\n" %
                      (len(templateSet)-preloadedTemplates))
            else:
                print("   All (%d) template waveforms already exist.\n" % len(templateSet))
                return templateSet
    else:
        preloadedTemplates = 0

    while (templateId < len(templateSet)):  # (Remember: indexing of templateId starts at '0')

        # If this template is not yet preloaded (as mseed), ...
        if (templateId+1 > preloadedTemplates):

            # Preload
            templateStream, templateShift = io.preload_template(
                    templateId, templateSet[templateId][0],
                    config.duration, templateSet[templateId][-1])

            if templateStream is None:
                raise Exception("Something is wrong with template %d." % (templateId+1))
            else:
                # Update timeshift entry in templateSet (even if still '0.0')
                templateSet[templateId][1] = templateShift

                preloadedTemplates += 1  # Mark as preloaded
                templateId += 1

        else:  # (template is already preloaded)
            templateId += 1  # Go to next template (-> don't load again)

    # ... and write the whole (new) template set to file
    io.writeTemplateSet(templateSet, verbose=0)

    return templateSet


# --- Main


def main():
    """Main function"""

    RAMfree = psutil.virtual_memory().available
    samplingRate = 1000
    numChannels = 3
    numStations = 1
    RAMperTrace = samplingRate*86400*np.array(0, dtype=np.float32).nbytes
    RAMrequired = RAMperTrace * 3 * numChannels * numStations  # *3 for BP & xCorr trace
    if (RAMfree < RAMrequired):
        print("\n---\nI guess you don't have enough memory to run this script, "
              "especially not for a %dHz sampling rate. Sorry. Goodbye.\n"
              "Hint: Try restarting the python kernel.\n---\n" % samplingRate)
        exit()


if __name__ == '__main__':
    # Execute only if run as a script
    main()
