# -*- coding: utf-8 -*-
"""
Submodule for plotting.

(created) 2015, Marcus Herrmann, Swiss Seismological Service, ETH Zürich
"""

# --- Import modules

import time
import datetime
import os      # OS specific
import subprocess
import glob    # For reading files with wildcard

# Scientific
import math
import scipy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm

# ObsPy
from obspy import UTCDateTime
from obspy.core import read, Stream

# Own modules
from TM.setupclass import config
from TM.waveformops import waveform_processing
from TM.utils import getChannelIndex
from TM import io


def plotWaveforms(events, eventType, sortAfter='time', procType='bandpass'):
    """
    Plots the waveforms in 'events' to a .PNG file

    Eventtype can be ['templates', 'detections', 'falseEvents'].
    This mainly determines the plot location.

    The filenames can be generated so that they appear 'sortAfter' [time, CCC, template]
    """

#    if eventType not in ['templates', 'detections', 'falseEvents']:
#        raise Exception("Not a valid event type.")

    # Stop the time
    tX = time.time()

    if not eventType == 'templates':
        TMdata_now = config.TMdata + config.directory.replace("_pass2", "")
        # The standard plot directory
        plots_dir = TMdata_now + "/" + eventType + "_plots_" + procType
        # Create appropriate subdirectories for sorted output
        if (sortAfter == 'CCC') or (sortAfter == 'template'):
            plots_dir += "/sorted_" + sortAfter
        print("[plotWaveforms()] Plotting %d %s... Stay tight!"
              % (len(events), eventType), flush=True)

    # Check if this directory exists, and create if necessary
    if not os.path.exists(plots_dir):
        os.makedirs(plots_dir)

    # Create a matplotlib figure instance to prevent a memory leak of matplotlib
    fig = read()[:config.numChannels].plot(handle=True)  # To have to default figure size

    # Loop over every entry in the given list
    for i in range(len(events)):

        timeStr = str(events[i][0]).replace(":", "-")

        if (eventType == 'templates'):

            # Get the stream / waveform
            stream = read(plots_dir + "/templates_mseed/template%02d_" % (i+1) +
                          timeStr + ".mseed")

            filename = ("/_template%02d_" % (i+1) + timeStr)

            # Reuse the same figure
            fig.clear()
            waveform_processing(stream, procType).plot(
                outfile=plots_dir + filename + ".png", fig=fig)

        else:

            # Get the stream / waveform
            stream = read(TMdata_now + "/detections_mseed/" + timeStr +
#                          "_%.2f" % events[i][-1] + "_Template%02d" % (events[i][-2]) +
                          ".mseed")

            # Modify filename so that it will be sorted by ...
            # 1) the time (DEFAULT)
            if (sortAfter == 'time'):
                filename = ("/detect_" + timeStr + "_%.2f" % (events[i][-1]) +
                            "_Template%02d" % (events[i][-2]))
            # 2) the CCC
            elif (sortAfter == 'CCC'):
                filename = ("/detect_%.2f_" % (events[i][-1]) + "Template%02d_" % (events[i][-2]) +
                            timeStr)
            # or 3) the template #
            elif (sortAfter == 'template'):
                filename = ("/detect_Template%02d_" % (events[i][-2]) +
                            "%.2f_" % (events[i][-1]) +
                            timeStr)

            # Reuse the same figure
            fig.clear()
            waveform_processing(stream, procType).plot(
                outfile=plots_dir + filename + ".png", fig=fig)

    print("[plotWaveforms()] ... finished after %.2fmin." % ((time.time() - tX)/60))
    plt.close("all")


def plotHelicorder(foundEvents, channel, threshold, procType='bandpass'):
    """
    Plot all detections found by a template into one plot -- sorted by CC.
    """

    TMdata_now = config.TMdata + config.directory.replace("_pass2", "")

    directory_heli = TMdata_now + "/" + "helicorder_plots"

    if procType == 'bandpass':
        directory_heli += '_BP'
        resampling = config.resampling
        linewidth_template = 2
    else:
        resampling = 500
        linewidth_template = 1

    # Check if this directory exists, and create if necessary
    if not os.path.exists(directory_heli):
        os.makedirs(directory_heli)
        os.makedirs(directory_heli + "/PDF")

    numTemplates = int(max([float(x[-2]) for x in foundEvents]))

    print("[plotHelicorder()] Plotting helicorders for %d templates... "
          % (numTemplates), flush=True)

    for iTemplate in range(numTemplates):

        # Get template-relevant detections higher than threshold
        detects = [x for x in foundEvents if x[-2] == iTemplate+1 and x[-1] > threshold]
        # ... and sort after CC
#        detects_CC = [x[-1] for x in detects]
#        detects = sorted(zip(detects, detects_CC),
#                         key=lambda pair: pair[-1], reverse=True)
#        detects = [x[0] for x in detects]  # Remove the CC column
        detects = sorted(detects, key=lambda x: x[-1], reverse=True)

        nDetects = len(detects)
        if nDetects == 0:
            continue

        # Load template waveform & get desired channel
        templateStream = read(glob.glob(config.templateDir + "/templates_mseed/template%02d_"
                                        % (iTemplate+1) + "*.mseed")[0])
#        try:
#            templateTrace = templateStream.select(channel=channel)[0]
#        except:
#            print("Template %d doesn't have this channel, only %s" %
#                  (iTemplate+1, ' '.join(tr.stats.channel for tr in templateStream)))
#            continue
#        else:
#            pass
        templateChann = [tr.stats.channel for tr in templateStream]
        channel_idx_want = getChannelIndex(channel, templateChann)
        templateTrace = Stream([templateStream[i] for i in channel_idx_want])

        # Process
        templateTrace = waveform_processing(templateTrace, procType, resampling)

        # Normalize
        templateTrace.data = (templateTrace.data/max(abs(templateTrace.data))) * 0.8

        # Load all events into a stream
        heliStream = Stream()
        for event in detects:
            detectStream = read(TMdata_now + "/detections_mseed/" +
                                str(event[0]).replace(":", "-") +
                                # "_%.2f" % event[-1] + "_Template%02d" % (event[-2]) +
                                ".mseed")
#            detectTrace = detectStream.select(channel=channel)[0]
            detectChann = [tr.stats.channel for tr in detectStream]
            channel_idx_want = getChannelIndex(channel, detectChann)
            detectTrace = Stream([detectStream[i] for i in channel_idx_want])

            detectTrace = waveform_processing(detectTrace, procType, resampling)
            detectTrace.data = (detectTrace.data/max(abs(detectTrace.data))) * 0.8
            heliStream += detectTrace

#        chnlName = heliStream[0].stats['channel']

        # Plot!
        # Split up
        splitUp = 250
        numOfSplits = nDetects // splitUp
        if nDetects % splitUp > 0:
            numOfSplits += 1
        for iSplit in range(numOfSplits):

            detects_split = detects[iSplit*splitUp:(iSplit+1)*splitUp]
            heliStream_split = heliStream[iSplit*splitUp:(iSplit+1)*splitUp]

            fig = plt.figure(num=None, figsize=(20, len(detects_split)),
                             dpi=72, facecolor='w', edgecolor='k')

            levelYaxis = 1  # Starting level for Y-axis
            t = heliStream[0].times()

            # Plot the template waveform
            plt.plot(t, templateTrace.data, color='k', linewidth=linewidth_template)

            for i in range(len(detects_split)):

                plotLabel1 = "M%.1f  |  CC%.2f" % (detects_split[i][2], detects_split[i][-1])
                plotLabel2 = detects_split[i][0]

                plt.plot(t, heliStream_split[i].data + levelYaxis, color='k')
                plt.text(0, levelYaxis-0.5, plotLabel1, fontsize='large',
                         color='w', backgroundcolor='#808080', va='top')
                plt.text(t[-1], levelYaxis-0.5, plotLabel2, fontsize='x-large',
                         color='w', backgroundcolor='#808080', va='top', ha='right')

                levelYaxis += 1

            # Plot the template waveform again
            plt.plot(t, templateTrace.data + levelYaxis, color='k', linewidth=linewidth_template)

            plt.title(channel, fontsize='xx-large')
            plt.xlabel("Time [s]", size='large')
            plt.ylabel("Detect #", size='large')
            yTick = np.arange(0, i+3, 1)
            yTick_label = np.arange(iSplit*splitUp+1, iSplit*splitUp+len(detects_split)+1, 1)
            yTick_label = yTick_label.astype('str')
            yTick_label = np.insert(yTick_label, 0, 'Template')
            yTick_label = np.append(yTick_label, 'Template')
            plt.yticks(yTick, yTick_label)
            plt.tick_params(axis='both', which='major', labelsize='x-large')
            plt.ylim([-0.95, i+2.95])
            plt.gca().invert_yaxis()

            # Save figures
            if numOfSplits > 1:
                splitStr = '_%d' % (iSplit+1)
            else:
                splitStr = ''
            plt.savefig(directory_heli + "/%s_template%02d" % (channel, iTemplate+1) +
                        "%s_%ddetects.png" % (splitStr, i+1), bbox_inches='tight')
            plt.savefig(directory_heli + "/PDF/%s_template%02d" % (channel, iTemplate+1) +
                        "%s_%ddetects.pdf" % (splitStr, i+1), bbox_inches='tight')
            plt.close(fig)

        file = "/detections_template%02d" % (iTemplate+1)
        io.writeDetections(detects, threshold, directory_heli + file, verbose=0)


def plot_hist_template_distribution(foundEvents, foundEvents_full=[], threshold=None):
    """
    Make histogram of templates' detection numbers
    """

    # Default value for threshold
    if threshold is None:
        threshold = config.threshold

    suffix = '_unique'

    if foundEvents_full:
        templateDistrib = [x[-2] for x in foundEvents_full if x[-1] >= threshold]
        numTemplates = max(templateDistrib)
        plt.hist(templateDistrib, bins=np.arange(0.5, numTemplates+1.5),
                 color='0.85', label='raw')
        suffix = '_+full'

    templateDistrib = [x[-2] for x in foundEvents if x[-1] >= threshold]
    numTemplates = max(templateDistrib)
    plt.hist(templateDistrib, bins=np.arange(0.5, numTemplates+1.5),
             color='#4292c6', label='after unifying')

    plt.xlabel("Template #")
    plt.ylabel("Detections")
    plt.xlim([0.5, numTemplates+0.5])
    plt.xticks(np.arange(1, numTemplates+1))
    plt.tick_params(axis='x', which='major', labelsize=6)

    plt.legend()

    plt.savefig(config.directory + "/_templateSet_hist_distribution%s_thresh%.2f.png"
                % (suffix, threshold),   dpi=150)
    plt.close()


def plot_hist_template_counts(foundEvents, threshold=None):
    """
    Histogram of how many templates detected an event (should be non-uniqued detections)!
    """

    # Default value for threshold
    if threshold is None:
        threshold = config.threshold

    foundEvents = [x for x in foundEvents if x[-1] >= threshold]

    # Get only the times
    times = [x[0] for x in foundEvents]
    # Group the multiples together by index using the times
    multiplets = np.split(times, np.where(abs(np.diff(times)) >= config.timeTol)[0]+1)

    # Make histogram of templates per event
    templatesCounts = [len(x) for x in multiplets]
    maxCount = max(templatesCounts)
    maxRange = max([x[-2] for x in foundEvents])

    plt.figure()
    plt.hist(templatesCounts, bins=np.arange(0.5, maxRange+1.5), color='#4292c6')
    plt.axvline(x=maxCount+0.5, linestyle='--', color='k')

    plt.xlabel("Number of detecting templates")
    plt.ylabel("Detections")
    plt.xlim([0.5, maxRange+0.5])
    plt.xticks(np.arange(1, maxRange+1))
    plt.tick_params(axis='x', which='major', labelsize=6)
    plt.savefig(config.directory + "/_templateSet_hist_counts_thresh%.2f.png" % threshold,
                dpi=150)
    plt.close()


def plotTimeline(infile, threshold=None, numTemplates=0, colorBy='templateNumber',
                 startdate=None, enddate=None):
    """
    Plots the timeline using Toni's GMT script.
    """

    infile = config.directory + os.sep + infile
    if 'Mag' not in infile:
        infile += '_Mag'  # (to allow other 'Mags' (pref*, all*, ...))
    infile += '_thresh%.2f' % threshold

    # Default values
    if threshold is None:
        threshold = config.threshold
    if startdate is None:
        startdate = config.startdate
    if enddate is None:
        enddate = config.enddate

    suffix = ''
#    suffix = '_x'

    outfile = infile.replace('detections', 'timeline' + suffix).replace('_Mag', '')

    foundEvents = io.readDetections(infile)

    if len(foundEvents[0]) < 5:
        raise Exception('No magnitude present.')

    # --- Make some early calculations (difficult to do in gmt-script)
    templateSet = io.readTemplateSet(config.directory)
    # 1. For timeline
    if numTemplates == 0:
        numTemplates = len(templateSet)
#        numTemplates = max([float(row[3]) for row in foundEvents])
#    minMag = math.floor(min([float(row[2]) for row in foundEvents])-0.1)
#    maxMag = math.ceil(max([float(row[2]) for row in foundEvents])+0.1)
    minMag = math.floor(min([float(row[2]) for row in foundEvents])*10)/10-0.2
    maxMag = math.ceil(max([float(row[2]) for row in foundEvents])*10)/10+0.2

    # 2. For histogram
    from collections import Counter
    maxDailyRate = Counter([x[0].date for x in foundEvents]).most_common()[0][-1] * 1.05
    maxDailyRate_Order = max([np.floor(np.log10(maxDailyRate)-0.05), 0])
    maxDailyRate = np.ceil(maxDailyRate/10**maxDailyRate_Order)*10**maxDailyRate_Order
    dailyRate_tick = 10**maxDailyRate_Order
    if maxDailyRate/dailyRate_tick > 10:
        dailyRate_tick *= 2
    elif maxDailyRate/dailyRate_tick < 4:
        dailyRate_tick /= 2

    # 3. For Cumulative number
    maxCumRate = len(foundEvents)
    maxCumRate_Order = max([np.floor(np.log10(maxCumRate)-0.2), 0])
    maxCumRate = np.ceil(maxCumRate/10**maxCumRate_Order)*10**maxCumRate_Order
    cumRate_tick = 10**maxCumRate_Order
    if maxCumRate/cumRate_tick > 10:
        cumRate_tick *= 2
    elif maxCumRate/cumRate_tick < 4:
        cumRate_tick /= 2

    # --- Deal with catalog info
    # (will be plotted as gray circles)
    catalog = []
    with open(config.directory + "/regression_catalogALL.dat", "r") as file:
        for line in file:
            catalog.append(line.strip().split())

    # --- Deal with template information
    # (if loaded, will plot over detecions)
    templateSetX = []
    catalogTimes = [UTCDateTime(x[1]) for x in catalog]
    # Identify which catalog events are templates (and get their magnitudes)
    for template in templateSet:
        eventIndex = catalogTimes.index(template[0])
        templateSetX.append(catalog[eventIndex])
    # Write out template events + number + regression magnitude
    with open(config.directory + "/regression_templates.dat", 'w') as file:
        for template in templateSetX:
            if UTCDateTime(template[1]) >= startdate and \
               UTCDateTime(template[1]) <= enddate:
                file.write("%s   %s   %s\n" % (template[1], template[2], template[-1]))
            else:
                # To not plot templates that are out-of-bounds
                file.write("%s   NaN   NaN\n" % (template[1]))

    # --- Handle the event coloring
    if colorBy == 'templateDistance':  # (by distance to the well base)

        # Inject a string into the outfile string
        strpos = outfile.find('_thresh')
        outfile = outfile[:strpos] + '_distance' + outfile[strpos:]

        catalog = io.getNicoCatalog(getAll=True, getOldVer=False, locations=True)

        # Identify which catalog events are templates (and get their magnitudes/locations)
        templateSet_loc = [event for template in templateSet
                           for event in catalog if event[1] == template[0]]

        center = [611663, 270611, 4354.0]  # WGS84: 7.593662 47.586091, 4401.8

        # Determine template distances
        templateSet_dist = [round(math.sqrt((ev[2]-center[0])**2 +
                                            (ev[3]-center[1])**2 +
                                            (ev[4]-center[2])**2)) for ev in templateSet_loc]

        if not templateSet_dist:
            raise Warning("You seem to not use the right catalog")

        # Sort
        idx = np.asarray(templateSet_dist).argsort().argsort()  # Twice to get the rank
        for detection in foundEvents:
            detection[-2] = idx[detection[-2]-1] + 1
#        writeDetections(foundEvents, threshold, file=infile[:infile.find('_thresh')])
        io.writeToFile(config.directory + '/timeline_template_order.dat',
                       [(0, 0, '---')] +  # to deal with undetected (tp0) events
                       list(zip(range(1, len(templateSet)+1), idx+1, templateSet_dist)),
                       "tp%d   tp%d   %sm\n")

    else:  # (by template number)

        # 1:1
        templateNum = range(1, len(templateSet)+1)
        io.writeToFile(config.directory + '/timeline_template_order.dat',
                       [(0, 0, '---')] +  # to deal with undetected (tp0) events
                       list(zip(templateNum, templateNum, templateNum)), "tp%d   tp%d   Tpl.%s\n")

    # --- Generate script call + arguments
    command = os.path.join(config.appDir, 'scripts/plotTimeline.sh')
    args = [infile + '.dat', numTemplates, startdate.date, enddate.date,
            minMag, maxMag,                # Timeline: mag limit
            maxDailyRate, dailyRate_tick,  # Histogram: maximum, major tick
            maxCumRate, cumRate_tick,      # Histogram: maximum, major tick
            outfile]
    strArgs = [str(x) for x in args]
    strArgsJoint = ' '.join(strArgs)

    # Set environment path for GMT
#    environ = {'PATH': os.environ['PATH'] + ':/usr/lib/gmt/bin',
#               'HOME': os.environ['HOME']}

    print("\n[plotTimeline()] Calling: %s %s\n" % (command, strArgsJoint))

#    # Open .ps output-file
#    f = open(outfile + '.ps', "w")
#    # Execute GMT script (and write its standard output into specified file f)
##    subprocess.call(command + strArgsJoint, stdout=f, env=environ, shell=True)
#    ps = subprocess.Popen([command] + strArgs, stdout=f)  # Avoiding shell + calling subprocess
#    ps.wait()  # Wait for subprocess to terminate
#    f.close()
#
#    # Convert PS to PNG and PDF
#    #  (stderr=subprocess.STDOUT also captures stderr in stdout, i.e. the python console)
#    subprocess.call([os.path.join(config.appDir, 'scripts/make_PDF+PNG.sh'), outfile],
#                    stderr=subprocess.STDOUT)

    # Execute GMT script
    subprocess.call([command] + strArgs, stderr=subprocess.STDOUT)


def plotLocations(templateSet, threshold=None):
    """
    Plots the template locations + catalog events using Toni's GMT script.
    """

    # Default value for threshold
    if threshold is None:
        threshold = config.threshold

    catalog = []
    numTemplates = len(templateSet)

    print(config.region)

    if (config.region == "Basel"):

        catalog = io.getNicoCatalog(getAll=True, locations=True)

        templateFileLineStr = "%s   %3.3f   %3.3f   %1.3f   %.1f   % d\n"

        command = os.path.join(config.appDir, 'scripts/plotLocations_Basel.sh')
        args = [config.directory, numTemplates]

    else:  # (for individual sequences)

        filename = os.path.join(config.dataDir, 'seddb_' + config.region + '.csv')

        # Read the events derived from DB = catalog events
        with open(filename, "r") as file:
            next(file)    # leave out column header
            for line in file:
                if line[0] == "#":  # Remove comment lines
                    continue
                entry = line.strip().split(',')
                catalog.append([0,  # ID
                                UTCDateTime(entry[1]) + config.timeTravel,  # Time + traveltime
                                float(entry[3]),   # X coord (lon)
                                float(entry[2]),   # Y coord (lat)
                                float(entry[4]),   # Z coord
                                float(entry[9])])  # ML
        # Write out the catalog events with location
        io.writeToFile(config.directory + "/loc_catalog.dat",
                       [x[1:] for x in catalog], "%s   %2.5f   %2.5f   %1.2f   %.2f\n")

        # Find the catlog event's template association
        # Load a file that contains template detections (only unified and with Mag)
        foundEvents = io.readDetections(config.directory +
                                        '/detections_%dtps_unique_Mag_thresh%.2f' %
                                        (len(templateSet), threshold))
        catalog_assoc = []
        # For every catalog event, ...
        for event in catalog:
            # ... try to find the corresponding detection (with tolerance)
            for detection in foundEvents:
                if (abs(detection[0] - event[1]) <= 3.0):
                    # Event found. Append event data + template association + CC
                    catalog_assoc.append(event[1:] + detection[3:5])
                    break  # (don't add another event that fulfills this critera)
        # Write out the catalog events with location, template association, and CC
        io.writeToFile(config.directory + "/loc_catalog_assoc.dat",
                       catalog_assoc, "%s   %2.5f   %2.5f   %1.2f   %.2f   % d   %.2f\n")

        templateFileLineStr = "%s   %2.5f   %2.5f   %1.2f   %.2f   % d\n"

        command = os.path.join(config.appDir, 'scripts/plotLocations.sh')
        args = [config.region, config.directory, numTemplates]

    # Identify which catalog events are templates (and get their magnitudes)
    templateSet_LOC = []
    for idx, template in enumerate(templateSet):
        # Find the corresponding entry
        catalogEvent = [x for x in catalog if template[0] == x[1]][0]
        # Get coordinates of the used templates
        templateSet_LOC.append(catalogEvent[1:] + [idx+1])

    # Write out the located templateSet (only for local plots)
    file = config.directory + "/loc_templateSet.dat"
    with open(file, 'w') as f:
        for templateLoc in templateSet_LOC:
            f.write(templateFileLineStr % tuple(templateLoc))

    # Set environment path for GMT 4
#    environ = {'PATH': os.environ['PATH'] + ':/usr/lib/gmt/bin',
#               'HOME': os.environ['HOME']}

    strArgs = [str(x) for x in args]

    print("\n[plotLocations()] Calling: %s %s\n" % (command, ' '.join(strArgs)))

    # Execute GMT script
    subprocess.call([command] + strArgs, stderr=subprocess.STDOUT)


def plotCHlocationOverview(directory):
    """
    Plots an overview map of all scanned sequences using an GMT script
    """

    filenames = glob.glob(os.path.join(config.dataDir, 'seddb_' + '*' + '.csv'))

    sequenceNames = []
    catalog = []
    for filename in filenames:
        with open(filename, "r") as file:
            next(file)    # leave out column header
            for line in file:
                if line[0] == "#":  # Remove comment lines
                    continue
                entry = line.strip().split(',')
                catalog.append([0,   # ID
                                UTCDateTime(entry[1]) + config.timeTravel,  # Time + traveltime
                                float(entry[3]),    # X coord (lon)
                                float(entry[2]),    # Y coord (lat)
                                float(entry[4]),    # Z coord
                                float(entry[9])])   # Ml
        sequenceNames.append(catalog[-1][2:4] + [filename.split('_')[1].split('.')[0]])

    # Write out the catalog events with location
    io.writeToFile(directory + "/loc_catalog.dat",
                   [x[1:] for x in catalog], "%s   %2.5f   %2.5f   %1.2f   %.2f\n")

    # Write out the sequence names + locations
    io.writeToFile(directory + "/loc_sequences.dat", sequenceNames, " %2.5f   %2.5f   %s\n")

    # Set environment path for GMT
    environ = {'PATH': os.environ['PATH'] + ':/usr/lib/gmt/bin',
               'HOME': os.environ['HOME']}

    command = os.path.join(config.appDir, 'scripts/plotLocations_all.sh')
    strArgs = [directory]

    print("\n[plotLocations()] Calling: %s %s\n" % (command, ' '.join(strArgs)))

    # Execute GMT script
    subprocess.call([command] + strArgs, env=environ, stderr=subprocess.STDOUT)


def plot_amp_regression(magSet, ampSet, regLineInfos=None, legendPos='upper left'):
    """
    Plots the regression for given magnitude <-> amplitude pairs.

    Function is capable of treating several sets.
    """

    # Determine how many different sets we have to plot
    if isinstance(magSet, list) or isinstance(ampSet, list):
        numSets = len(magSet)
        if len(ampSet) != numSets:
            raise Warning("Number of magnitude and amplitde sets doesn't agree.")
    else:
        numSets = 1
        # Pack the arrays into a list (with one entry)
        magSet = [magSet]
        ampSet = [ampSet]

    # Set colors and other things
    if numSets == 1:
        colors = ['k']
        legFontSize = 'small'
    elif numSets > 1 and numSets < 5:
        colors = ['c', 'orange', 'crimson', 'black']  # Useful for different FPS types
        legFontSize = 'x-small'
    else:
        colors = cm.Set1(np.linspace(0, 1, numSets))  # Give a looot more colors (last one is gray)
        legFontSize = 'xx-small'

    plt.figure()

    for iSet, mags in enumerate(magSet):

        amps = ampSet[iSet]

        # Plotting
    #    plt.xlabel("log(A) @%s" % (station[1]))
        plt.xlabel("max. amplitude @%s" % (config.station))
        plt.ylabel("ML (SED)")

        # Variant 1: scatter with density coloring (only useful for ONE set)
        if numSets == 1 and len(mags) > 100:
            # Calculate the point density
            xy = np.vstack([np.log10(amps), mags])
            density = scipy.stats.gaussian_kde(xy)(xy)
            # Sort points by density; densest points are plotted last
            idx = density.argsort()
            # Plot
            plt.scatter(amps[idx], mags[idx], s=25, alpha=0.5,
                        c=density[idx], cmap=plt.get_cmap('gist_heat'), edgecolor='')
            plt.gca().set_xscale('log')

        # otherwise Variant 2: simple one-colored scatter (per set)
        else:
            plt.semilogx(amps, mags, '.', c=colors[iSet])

    # min/max among all sets
    minAmp_log10 = math.log10(min([min(x) for x in ampSet]))
    maxAmp_log10 = math.log10(max([max(x) for x in ampSet]))
#    minAmp_log10 = math.log10(1e4)
#    maxAmp_log10 = math.log10(2e7)
    minMag = min([min(x) for x in magSet])
    maxMag = max([max(x) for x in magSet])

    # Plot (possible) regression line(s) + infos
    if regLineInfos:
        # Check if it is only set
        if not isinstance(regLineInfos[0], list):
            regLineInfos = [regLineInfos]  # Pack into a list

        for idx, regLine in enumerate(regLineInfos):
            # Unpack values
            b, a, r_value, p_value, std_err, label = regLine
            # Create x vector
            if len(regLineInfos) == numSets:
                x_plot = np.array([math.log10(min(ampSet[idx]))-0.1,
                                   math.log10(max(ampSet[idx]))+0.1])
            else:
                x_plot = np.array([minAmp_log10-0.1, maxAmp_log10+0.1])
            # Plot
            plt.plot(10**x_plot, a + b*x_plot, '-', c=colors[idx], lw=1,
                     label=('%s\n' % label +
                            'ML = %.2f * log(A) + %.2f' % (b, a) +
                            '\n$R^2$: %.4f    $\sigma_{slope}$: %.4f' %
                            (r_value**2, std_err)))
            label += '_'
        #    plt.plot(10**x_plot, a_maxL + b_maxL*x_plot, 'r-', lw=1)

    #    plt.axis('scaled')  # Doesn't work with semilog axis
    plt.axis([10**(minAmp_log10-0.2), 10**(maxAmp_log10+0.2),
              minMag-0.2, maxMag+0.2])
#    plt.axis([10**(4.9893-0.2), 10**(7.0858+0.2), 0.7-0.2, 3.4+0.2])  # SED events for OTER2
#    plt.axis([10**(3.8516-0.2), 10**(6.7256+0.2), 0.7-0.2, 3.4+0.2])  # SED events for MATTE

    # Legend
    l = plt.legend(loc=legendPos, fontsize=legFontSize, fancybox=True)
    # Color even the text of each entry with the appropriate color
    for idx, legTexts in enumerate(l.get_texts()):
        legTexts.set_color(colors[idx])

    plt.savefig(config.directory + "/regression_mag-amp_%s%devents.png" %
                (label, sum(len(x) for x in magSet)), dpi=150, bbox_inches='tight')
#    plt.savefig(config.directory + "/regression_mag-amp_%devents.pdf" % len(catalog_ML),
#                dpi=150, bbox_inches='tight')
    # Determine current time as ID... to not overwrite possible previous file
#    crtTime = time.strftime("%Y-%m-%dT%H%M%S", time.localtime())
#    plt.savefig("plots/regression_" + crtTime + ".png", dpi=150)
    plt.close()

    # Plot residuals
    if numSets == 1:
        amps = ampSet[0]
        mags = magSet[0]

        resis = mags - a - b*np.log10(amps)
        plt.plot(amps, resis, 'k.')
        plt.axhline(color='k', linestyle='--')
        plt.gca().set_xscale('log')

        plt.annotate('sum(residuals²): %.2f' % sum(resis**2),
                     xy=(0, 1), xycoords='axes fraction', fontsize=12,
                     xytext=(4, 4), textcoords='offset points',
                     ha='left', va='bottom', bbox=dict(boxstyle="square", fc="w"))

        plt.xlabel("max. amplitude @%s" % (config.station))
        plt.ylabel("residuals")
        plt.axis([10**(minAmp_log10-0.2), 10**(maxAmp_log10+0.2), -1, 1])

        plt.savefig(config.directory + "/regression_mag-amp_resid_%s%devents.png" %
                    (label, len(mags)), dpi=150, bbox_inches='tight')
        plt.close()


def plotRate(foundEvents, variant='adaptive-interval', xscale='log', start=None, para=None):
    """
    Plots the rate over time in a log-log or log-lin-plot (e.g. to investigate Omori decay)

    Variants:
        'daily':
            Simply ploting the daily rate
        'adaptive-interval'
            Adaptive window length - depending on rate
        'log-interval':
            Rate in log-normalized windows (increasing length over time)
    """

    print("[plotRate()] ... ")

    # Define exact starting time for plot
    if start is None or not isinstance(start, UTCDateTime):
        start = UTCDateTime(foundEvents[0][0])
    print('   Start day/time is %s.' % start)

    startingDay = start.date
    endDay = foundEvents[-1][0].date
    numOfDays = (endDay - startingDay).days + 1

    # VERSION 1
    if 'daily' in variant:

        dates = [x[0].date for x in foundEvents if x[0] >= UTCDateTime(start.date)]

        days = []
        dailyRates = []
        # Determine rate of each day
        for i in range(numOfDays):
            day = startingDay + datetime.timedelta(days=i)
            days.append(day)
            dailyRates.append(dates.count(day))

        # Rates without empty days
#        from collections import Counter, deque
#        dailyRates = list([list(x) for x in Counter(dates).items()]).sort()

        # Cumulative rate
#        dailyRates_cum = np.cumsum(dailyRates)
#        plt.plot_date(days, dailyRates_cum)
#        plt.plot(list(range(numOfDays)), dailyRates_cum)
#        plt.xscale('log')

        # Create xtick dates + labels
        if xscale == 'log':

            # Define the first ticklabel (and it's location)
            xtick_days = [1]  # Where to place the first label
            xtick_labels = [str(start.date)]  # What it reads

            if config.region == 'Basel':
                xtick_labels += ['2007', '2008', '09', '10', '12', '16']
            else:
                # TODO
                xtick_labels += []
        else:
            # TODO
            xtick_days = []  # Placing of first label not clear yet
            xtick_labels = [str(x) for x in list(range(startingDay.year, endDay.year+1))]

        for label in xtick_labels[len(xtick_days):]:
            # Try different time formats for converting the label
            for timestr in ['%Y', '%y', '%Y-%m-%d']:
                try:
                    # Get the day difference b/w desired x-axis labels and starting date
                    daydiff = datetime.datetime.strptime(str(label), timestr).date() - startingDay
                    xtick_days += [daydiff.days + 1]
                    break
                except:
                    pass

        plt.figure()
        plt.axis([0.7, numOfDays*1.1,
                 0.5, 10**math.ceil(math.log10(max(dailyRates)))])
        plt.gca().set_autoscale_on(False)
        plt.semilogy(list(range(1, numOfDays+1)), dailyRates, 'k.')

        if xscale == 'log':
            plt.xscale('log')

        plt.xticks(xtick_days, xtick_labels)
        plt.tick_params(axis='x', which='major', labelsize=10)
    #    plt.xlabel("Days after %s" % str(startingDay - datetime.timedelta(days=1)))

        plt.ylabel("Daily rate")

        plt.savefig(config.directory + '/rate_start%s_%s_%s.png' %
                    (str(start.date), variant, xscale), dpi=150)

    # VERSION 2 (using adaptive intervals)
    elif 'interval' in variant:

        # Determine relative times to starting time [in days]
        timesRelative = [(x[0]-start)/60/60/24 for x in foundEvents if x[0] >= start]

        if 'adaptive' in variant:

            # Treat the windowing parameter
            if para is None or para <= 1:
                nEQ = 100  # Number of earthquakes that define one interval
            else:
                nEQ = para
            strPara = 'nEQ%d' % nEQ
            print('   Earthquake window size set to %d.' % nEQ)

            # Create the adaptive intervals
            intervals = timesRelative[0::nEQ]  # Take every n-th time
            # Treat last interval separately
            intervals += [timesRelative[-1]]

            # Length of intervals (for normalization)
            norm = np.diff(intervals)

            # Count and normalize (to get average daily rate)
            rate = nEQ/norm
            # Treat (again) last interval separately (to fix it)
            rate[-1] = (len(timesRelative) % nEQ)/norm[-1]

        else:
            # Treat the windowing parameter
            if para is None or para >= 1:
                logBinWidth = 0.05  # Number of earthquakes that define one interval
            else:
                logBinWidth = para
            strPara = 'logbin%.2f' % logBinWidth
            print('   log-binning set to %.2f.' % logBinWidth)

            # Create log intervals
            intervals = 10**np.arange(-4, np.ceil(np.log10(numOfDays)/logBinWidth) * logBinWidth +
                                      logBinWidth, logBinWidth)
            intervals[-1] = numOfDays

            # Length of intervals (for normalization)
            norm = np.diff(intervals)

            # Count and normalize (to get average daily rate)
            hist = np.histogram(timesRelative, bins=intervals)
            rate = (hist[0]/norm)

        # The center of the intervals
        intervalsCenter = [intervals[i]+norm[i]/2 for i in range(len(intervals)-1)]

        # Create xtick dates + labels
        if xscale == 'log':
            # Determine first rate and its occurence time
            firstRateIdx = next((i for i, x in enumerate(rate) if x), None)
            firstRateTime = (start + intervalsCenter[firstRateIdx]*86400) \
                .datetime.replace(microsecond=0)

            # Define the first ticklabel (and it's location)
            xtick_days = [intervalsCenter[firstRateIdx]]  # Where to place the first label
            xtick_labels = [str(firstRateTime)]  # What it reads

            if config.region == 'Basel':
                xtick_labels += ['2006-12-09', '2007', '2008', '10', '12', '16']
            else:
                # TODO
                xtick_labels += []

        else:  # (linear axis)
            # TODO
            xtick_days = []  # Placing of first label not clear yet
            xtick_labels = [str(x) for x in list(range(startingDay.year, endDay.year+1))]

        for label in xtick_labels[len(xtick_days):]:
            # Try different time formats for converting the label
            for timestr in ['%Y', '%y', '%Y-%m-%d', '%Y-%m-%d %H:%M:%S']:
                try:
                    # Get the day difference b/w desired x-axis labels and starting date
                    daydiff = datetime.datetime.strptime(label, timestr) - start.datetime
                    xtick_days += [daydiff.total_seconds()/86400 + intervalsCenter[0]]
                    break
                except:
                    pass

        # Plotting
        plt.figure()
        plt.axis([intervalsCenter[0], numOfDays*1.1,
                 10**math.floor(math.log10(np.min(rate[np.nonzero(rate)]))),
                 10**math.ceil(math.log10(max(rate)))])
        plt.gca().set_autoscale_on(False)

        plt.semilogy(intervalsCenter, rate, 'k.-')
        # for non-adaptive: remove last value
        #   (b/c it's certainly not "complete" (not enough data for last bin))
#        plt.semilogy(intervalsCenter[:-1], rate[:-1], 'k.-')

        if xscale == 'log':
            plt.xscale('log')

        plt.xticks(xtick_days, xtick_labels)
        plt.tick_params(axis='x', which='major', labelsize=8)
        plt.tick_params(axis='x', which='minor', bottom='off')
#        plt.xlabel("Time after Mainshock [days]")

        plt.ylabel("Normalized daily rate")

        plt.savefig(config.directory + '/rate_start%s_%s_%s_%s.png' %
                    (str(start.date), variant, xscale, strPara), dpi=150)

    plt.close()


def plotFMD(foundEvents, suffix='', otherCatalogs=None, otherBinnings=None, Mc=None,
            BIG=False, ani=False,
            starttime=None, endtime=None):
    """
    Plots the frequency-magnitude-distribution for a given event list with magnitude.
    """

    # Check if detection list does not already contain a magnitude...
    if len(foundEvents[0]) < 5:
        raise Warning("WARNING: Detection list doesn't contain a magnitude.")

    if suffix is not '' and suffix[0] is not '_':
        suffix = '_' + suffix

    # Default values
    if starttime is None:
        starttime = UTCDateTime('1900-01-01')
    else:
        suffix += '_' + str(starttime).replace(":00.000000Z", "").replace(":", "-") + '-'
    if endtime is None:
        endtime = UTCDateTime('2900-01-01')
    else:
        suffix += '_-' + str(endtime).replace(":00.000000Z", "").replace(":", "-")

    if BIG:
        plt.figure(num=None, figsize=(8, 6), dpi=150, facecolor='w', edgecolor='w')
        suffix = '_BIG' + suffix

    # Guess CC threshold
    thresholdGuess = min(np.floor([x[-1]*100 for x in foundEvents])/100)

    # Get relevant magnitudes
    magLists = [[x[2] for x in foundEvents if (x[0] >= starttime and x[0] < endtime)]]
    binnings = [0.01]  # Magnitude binning
    Mc_adds = [0.2]  # Mc correction

    numEv = len(magLists[0])
    if numEv == 0:
        print('\n[plotFMD()] No events in that period. Skipping FMD.')
        plt.close()
        return None
    else:
        print('\n[plotFMD()] Using %d events.' % numEv)
        print('    thresholdGuess: %.2f' % thresholdGuess)
        print('    suffix: %s' % suffix)

    # Color definition
    colors = ['#4292c6']
    color_incs = ['#FF0000']
    color_bline = str(0.1)

    # If other magnitude vectors given
    if otherCatalogs:
        # If only one catalog provided (as a list)
        if not isinstance(otherCatalogs[0][0], list):
            # ... pack it into another list (of catalogs)
            otherCatalogs = [otherCatalogs]
        if not isinstance(otherBinnings, list):
            otherBinnings = [otherBinnings]
        if len(otherCatalogs) is not len(otherBinnings):
            raise Warning('You have to specify the binning for each additional magnitude vector')

        binnings = binnings + otherBinnings
        Mc_adds_factor = [2, 2]

        timeindex = 1  # The column of time
        # Join all magnitude vectors
        for idx in range(len(otherCatalogs)):
            # If the time is in the first column
            if isinstance(otherCatalogs[idx][0], UTCDateTime):
                timeindex = 0
            magindex = timeindex+1
            magLists += [[x[magindex] for x in otherCatalogs[idx]
                          if (x[timeindex] >= starttime and x[timeindex] < endtime)]]
            Mc_adds += [Mc_adds_factor[idx] * binnings[idx+1]]

#        Mc_adds += [2*binnings[1], 2*binnings[2]]

        colors += ['#32CD32', '#DC143C']
        # Use cumulative dot color also for non-cumulative dots
        color_incs = colors

    for idx, mags in enumerate(magLists):

        binning = binnings[idx]      # Set the binning
        Mc_add = Mc_adds[idx]        # ... the Mc correction
        color = colors[idx]          # ... the color
        color_inc = color_incs[idx]  # ... the color for non-cumulative FMD
        if idx > 0:
            color_bline = color

        # Round magnitudes correctly (especially if they were converted)
#        mags = np.round(np.array(mags)/binning)*binning  # THIS CAN INTRUDUCE A SHIFT (e.g 0.079)
#        mags = np.round(np.array(mags)/0.001)*0.001

        # Determine the magnitude range
        maxMag = max(mags)
        minMag = min(mags)
        bins = np.arange(minMag, maxMag+2*binning, binning)

        # Calc Distribution
        hist, binEdges = np.histogram(mags, bins=bins)
        histCum = np.flipud(np.cumsum(np.flipud(hist)))

        # Determine b- and a-value
        # Calculate Mc (based on max. curvature)
        if Mc is None or idx > 0:  # (the None statement also allows Mc=0.0 to be true)
            Mc = binEdges[hist == max(hist)][0] + Mc_add
        else:
            Mc_add = 0.0
        #    Calculate mean magnitude & length of catalog
        magsC = [x for x in mags if x >= Mc]
        nLenC = len(magsC)
        try:
            meanMag = np.mean(magsC)
        except:
            print('No events above Mc. Skipping FMD.')
            plt.close()
            return None
        else:
            print('    Mc: %.2f' % Mc)
        #    Calculate b-value (maximum likelihood)
        bValue = np.log10(np.exp(1)) / (meanMag-(Mc-(binning/2)))
        #    Calculate standard deviation
        std = sum((magsC-meanMag)**2) / (nLenC*(nLenC-1))
        std = 2.30 * np.sqrt(std) * bValue**2
        #    Calculate a-value
        aValue = np.log10(nLenC) + bValue * Mc

        nLen = len(mags)

        # Plotting
#        plt.title("%d events (%d > Mc)" % (len(magList), len(magsC)), color='#4292c6')
        plt.semilogy(bins[:-1], histCum, '.', color=color)
        plt.semilogy(bins[:-1], hist, 'x', color=color_inc, markersize=2+idx)
#        x_plot = np.array([Mc, maxMag])
        x_plot = np.array([Mc, 4])
        plt.semilogy(x_plot, 10**(aValue - bValue*x_plot), '-', color=color_bline,
                     label="b: %.2f±%.2f (Mc %.2f+%.2f)" % (bValue, std, Mc-Mc_add, Mc_add))
#                     label="b: %.2f , a: %.2f\nfor Mc = %.2f" % (bValue, aValue, Mc))

        # Event summary
#        left_mag = plt.gca().get_xlim()[0]
        left_mag = np.floor(min(magLists[0]))
        plt.text(left_mag+0.05, nLen*0.90, "%d events\n (%d > Mc)" % (nLen, nLenC),
                 ha='left', va='top', color=color)
        # Legend
        plt.legend()

    plt.ylabel("Cumulative number of events")
    plt.xlabel("ML")

#    plt.tick_params(axis='both', which='major', labelsize='large')
    plt.xticks(np.arange(left_mag, np.ceil(maxMag)+1, 1))  # Major x-ticks
    plt.minorticks_on()
#    plt.gca().set_xticks(bins, minor=True)  # Minor x-ticks

    plt.gca().set_ylim(bottom=0.6)
#    plt.gca().set_ylim(top=10**4)
    plt.gca().set_xlim(left=-3)
#    plt.gca().set_xlim(right=np.ceil(maxMag))
    plt.gca().set_xlim(right=3.5)
#    plt.axis([-2.4, 1.2,
#              0.6, 2000])

    if ani:
        # To make looping animations with 'convert -delay 10 -loop 0 FMD*.png FMD_ani.gif'
        plt.savefig(config.directory + '/FMD_i%03d.png' % ((Mc*10)+100), dpi=150)
        plt.savefig(config.directory + '/FMD_i%03d.png' % (100-(Mc*10)), dpi=150)
    else:
        plt.savefig(config.directory + '/FMD_thresh%.2f%s_Mc%.2f.png' %
                    (thresholdGuess, suffix, Mc), dpi=150, bbox_inches='tight')
#        plt.savefig(config.directory + '/FMD_thresh%.2f%s_Mc%.2f.pdf' %
#                    (thresholdGuess, suffix, Mc), dpi=150)
    plt.close()
