# -*- coding: utf-8 -*-
"""
Setup class. Which should be instantiated to have a global "config" container.

(created) 2015, Marcus Herrmann, Swiss Seismological Service, ETH Zürich
"""

import sys
import os


class Setup:
    """ Configuration class; includes all the specifications needed for TM.
        Most functions that are used for the workflow have access to its parameters
        ... and depend on it!
    """

    region = None
    directory = None            # Output directory
    templateDir = None          # Template directory (it could be external)
    chunkCacheDir = None        # Chunk cache directory (it could be external)
    nNodes = 1                  # Number of available nodes on cluster
    conv_method = None          # Fastest convolution approach (sample length dependent)

    # Network / station related
    network = None
    station = None
    numStations = 1
    channels = None
    numChannels = None          # Will be autom. calculated from 'channels'
    channelSelect = None        # Will be autom. calculated from 'channels'

    # Template parameters
    timeTravel = None           # Approx. travel time of P-waves (minus some extra time)
    duration = None             # Duration of template waveform
    templateShiftCorr = None    # Whether to align templates to the first

    # Scanning parameters
    resampling = None           # maximum sampling rate (to save memory)
    filtConf = None             # Upper & Lower bounds & Filter order
    stop50Hz = None             # Whether to remove the 50Hz

    threshold = None
    threshCred = None           # Credible threshold
    threshPerChannel = None     # Apply thresh to EACH channel (and not MEAN among channels)

    startdate = None
    enddate = None              # The day that will not be scanned anymore (Pythonic?)

    chunk_len = None            # chunk length in seconds
    overlap_len = None          # chunks overlapping in seconds
    num_readers = None          # maximum number of reader processes
    chunk_on_disk = None        # maximum number of cached chunk on disk

    use_threads = None

    # Post processing
    timeTol = None              # The time tolerance to unique detections of different templates

    # Paths & sources
    dataDir = None              # Location of the sequence-related io data
    FDSN = None                 # Whether to use FDSN as waveform source
    FDSN_server = None          # The FDSN server
    mseedDir = None             # Location of the MSEED archive
    # Default dump directory of detection snippets (mseeds). Subdir will be 'config.directory'
    TMdata = os.path.expanduser("./")  # Like this, they will be written into the same directory

    # Absolute path of the location of this application (useful if not yet installed in system)
    appDir = os.path.abspath(sys.path[0])  # Location of the application

#    def __init__(self):

    def __setattr__(self, name, value):

        # Set number of channels for easier access
        if name == 'channels':

            if isinstance(value, str):
                value = [value]  # 'channels' has to be a list!
            super().__setattr__(name, value)
            self.numChannels = len(self.channels)
            self.channelSelect = '*[' + ''.join([x[-1] for x in self.channels]) + ']'

        # When setting directory, also set default template directory (it could be external)
        elif name == "directory":

            super().__setattr__(name, value)

            if 'templateDir' not in vars(self):
                self.templateDir = value

            if 'chunkCacheDir' not in vars(self):
                self.chunkCacheDir = value

        else:
            super().__setattr__(name, value)

    def set_dir(self, prefix, suffix='', openEnd=False, moreInfo=False):
        """
        Create a suitable directory name:
            {prefix}_[startdate]_{[enddate]}{_suffix}

            [] is taken from config variable, {} can be changed.

        :type prefix: str
        :param prefix: The first string of the directory name.
        :type suffix: str, optional
        :param suffix: The last string of the directory name.
        :type openEnd: bool, optional
        :param openEnd: replace enddate with 'openEnd' string to allow
                        easy continuing of an older scan
        :type moreInfo: bool, optional
        :param moreInfo: append more info b/w [enddate] and {_suffix}:
        """

        if suffix != '' and suffix[0] is not '_':
            suffix = '_' + suffix

        if openEnd:
            end = 'openEnd'
        else:
            end = self.enddate.date

        if moreInfo:
            suffix = (
                  "_%dcomp_resamp%d" % (self.numChannels, self.resampling) +
                  "_dur%.1fs_BP%d-%d_%dO" % tuple([self.duration] + self.filtConf) +
                  suffix)

        self.directory = (prefix +
                          "_%s_%s_%s_%s" % (self.region, self.station,
                                            self.startdate.date, end) +
                          suffix)

    def all_set_check(self):
        "Check if all required parameters were set"

        # Retrieve all possible custom parameters of this class
        missingVals = [x for x in dir(self) if self.__getattribute__(x) is None and
                       x[0] != '_' and 'FDSN' not in x]

#        # Determine which were not set yet
#        missingVals = []
#        for var in parameters:
#            if str(var) not in vars(self):
#                missingVals += [str(var)]

        if missingVals:
            raise ValueError('Not all parameters were set. Missing \n - %s' %
                             '\n - '.join(missingVals))
        else:
            print('   All set! Ready to go.\n')


#def loadConfig():
#    """
#    Loads the configuration for the current folder inot the class
#    """
#
#    filename = config.directory + os.sep + "config.txt"
#    with open(filename, "r") as file:
#        configLines = [line.strip().split(" = ") for line in file if not line[0] == "#"]
#
#    # TODO: finish
#    #   Remember that it should initialize the config class,
#    #   and then import all the values from the file
#
#    # TODO: use a standard format (INI, JSON, or YAML)
#    #   http://stackoverflow.com/a/27891283
#    #   https://martin-thoma.com/configuration-files-in-python/


# Instantiate the Setup class to create a global config object
if "config" not in globals():
    config = Setup()

