# -*- coding: utf-8 -*-
"""
Submodule containing some useful fuctions.

(created) 2015, Marcus Herrmann, Swiss Seismological Service, ETH Zürich
"""

# --- Import modules

# General
import os      # OS specific
import time

# Scientific
import math
import scipy
import scipy.signal  # To avoid scipy library module api problems
import numpy as np
import matplotlib.pyplot as plt

# ObsPy
from obspy import UTCDateTime
from obspy.core import read, Stream

# Own modules
from TM.setupclass import config
from TM.waveformops import waveform_processing
from TM import io


class Timer:
    """
    Own timer.
    Usage:
        with Timer() as t:
            DO SOMETHING
    """

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.interval = self.end - self.start
        print('This took %f s' % self.interval)


def getChannelIndex(channels, refChannels=None):
    """"
    For a list of channel names, return a list of channel indices. (Index starts at 0)
    """

    # 'channels' has to be a list!
    if isinstance(channels, str):
        channels = [channels]
    if isinstance(refChannels, str):
        refChannels = [refChannels]

    # Do some recursive magic
    if refChannels:

        refChannels_idx = getChannelIndex(refChannels)
        chnlIndices = np.intersect1d(getChannelIndex(channels), refChannels_idx)

        chnlIndices -= np.min(refChannels_idx)  # "Normalize" to actual existing channels

        if len(chnlIndices) < len(channels):
            raise Warning("Channels don't fit (i.e. no suitable reference channels).")

        return chnlIndices

    channelDB = np.array([["Z", "N", "E"],
                          ["A", "B", "C"],
                          ["Z", "2", "3"],
                          ["1", "2", "3"]])

    # Determine the refChannel rows that could be a candidate
    rows = np.where(channelDB == channels[0][-1])[0]

    # Try all possible candidates
    for row in rows:
        chnlIndices = []
        try:
            for chnl in channels:
                col = np.where(channelDB[row, ] == chnl[-1])[0][0]
                chnlIndices.append(col)
        except:
            # On error, try next refChannel row
            continue
        else:
            # If no error occured, it's the right one. Exit for loop
            break

    if len(chnlIndices) < len(channels):
        raise Warning("Unknown channel constellation.")

    return chnlIndices


def clusterTemplates(catalog):
    """
    Given a set of events (catalog), find common clusters using their waveforms.

    Returns: a list that can act as a new templateSet
    """

    print("\n[clusterTemplates()] ---------------------------")

    from scipy.cluster import hierarchy
    from scipy.spatial import distance

    # Check if this directory exists, and create if necessary
    if not os.path.exists(config.directory):
        os.makedirs(config.directory, exist_ok=True)

    # The catalog should have an ID in first column (and then the time)
    if not isinstance(catalog[0][1], UTCDateTime):
        # Otherwise prepend the continous index of/for each element
        [x.insert(0, ind) for ind, x in enumerate(catalog)]
        raise Warning("Catalog doesn't have event IDs. Using index.")

    # Load the catalog events into MSEEDs (if they don't exist yet)
    io.preload_events(catalog)

    # Use only "small" events
    catalog = [x for x in catalog if x[2] <= 1.2]

    print("(first, reading in waveforms of %d catalog-events)" % len(catalog))

    eventList = []

    for event in catalog:
        waveformStream = Stream()
        # Read waveforms (of each channel)
        try:
            waveformStream = read(os.path.join(
                                  config.dataDir,
                                  'SED-events_%s_mseed/%03d_' % (config.station, event[0]) +
                                  str(event[1]).replace(':', '-') + '.mseed'))
            # Do BP-filter
            waveformStream = waveform_processing(waveformStream, 'bandpass', resampling=None)
            # Imitate the same conditions as in the "real-time xCorr"
            t_start = waveformStream[0].stats.starttime
            waveformStream.trim(t_start + 0.25, t_start + 3.0 + 0.25)
            # Append the Stream object to the list
            eventList.append([event, waveformStream])
        # If ANY channel cannot be read...
        except Exception as err:
            print("Cannot read waveform of catalog event %d." % event[0], err)
            continue  # ... and ignore it
        else:
            pass

    # Create empty distance matrix
    dissimilarityMatrix = np.zeros((len(eventList), len(eventList)))

    # Pairwise calculate dissimilarity (using cross-correlation)
    for idx1 in range(len(eventList)-1):
        for idx2 in range(idx1+1, len(eventList)):

            for chnlInd in range(config.numChannels):

                trace1 = eventList[idx1][-1][chnlInd]
                trace2 = eventList[idx2][-1][chnlInd]

                # Do xCorr
                xCorr_comp = scipy.signal.fftconvolve(trace1.data,
                                                      trace2.data[::-1],
                                                      mode='same')
                # Normalize CCC
                xCorr_comp /= (trace1.data.std() *
                               trace2.data.std() *
                               trace2.count())

                if chnlInd == 0:
                    xCorr = xCorr_comp
                else:
                    # Stacking cross-correlation traces
                    xCorr += xCorr_comp

            # Compute dissimilarity ("distance")
            dissimilarityMatrix[idx1, idx2] = config.numChannels - max(xCorr)

    # Fill the lower triangle of matrix
    dissimilarityMatrix += dissimilarityMatrix.T

    # Make a condensed matrix
    dissimilarityMatrix_SYM = distance.squareform(dissimilarityMatrix)

    # Plot similarity matrix
    plt.figure()
    plt.imshow(config.numChannels - dissimilarityMatrix, interpolation="nearest")
    plt.savefig(config.directory + "/clustering_similMat_%devents_6s.png"
                % (len(eventList)), dpi=150)
    plt.close()

    # Do clustering
    dist_threshold = 1.2
    linkage = hierarchy.linkage(dissimilarityMatrix_SYM, method="average")
#    linkage = hierarchy.linkage(dissimilarityMatrix, method="ward")
    clusters = hierarchy.fcluster(linkage, dist_threshold, criterion="distance")
#    clusters = hierarchy.fcluster(linkage, 13, criterion="maxclust")
#    clusters = hierarchy.fcluster(linkage, 1.10, criterion="inconsistent")

#    # Create a dendrogram
    plt.figure()
    hierarchy.dendrogram(linkage, color_threshold=dist_threshold)
    plt.tick_params(axis='x', which='major', labelsize=3)
    plt.savefig(config.directory + "/clustering_dendrogram_%dclust_%devents_6s.png"
                % (max(clusters), len(eventList)), dpi=300)
    plt.close()

#    # Verify linkage
#    linkage_inconsistency = hierarchy.inconsistent(linkage)
#    linkage_maxinconsts = hierarchy.maxinconsts(linkage,linkage_inconsistency)
    print("Cophenet coefficient: %.4f" % hierarchy.cophenet(linkage, dissimilarityMatrix_SYM)[0])
#    print(linkage_maxinconsts)
#    plt.hist(linkage_maxinconsts)
    # Verify clustering
    print("Clusters: %d" % max(clusters))

    # Write used events and their assigned cluster
    file = (config.directory + "/clustering_eventList_%dclust_%devents.dat"
            % (max(clusters), len(eventList)))
    with open(file, 'w') as f:
        for idx, event in enumerate(eventList):
            line = ("%s   ML%.1f   %d\n"
                    % (event[0][1], event[0][2], clusters[idx]))
            f.write(line)

    # Plotting waveforms, sorted by cluster
    if 'fig' not in locals():
        # Create a matplotlib figure instance to prevent a memory leak of matplotlib
        fig = read()[:config.numChannels].plot(handle=True)  # To have to default figure size

    for i in range(len(eventList)):
        timeStr = str(eventList[i][0][1]).replace(":", "-")
        filename = "plots_SEDevents_3comp_cluster/%d_%d_%s.png" % (clusters[i],
                                                                   eventList[i][0][0], timeStr)
        fig.clear()
        eventList[i][-1].plot(outfile=filename, fig=fig)

    # Sort eventList by clusters
    events_sort = sorted(zip(eventList, clusters), key=lambda pair: pair[-1])
    # Form a nested list, split by clusters
    events_clustered = np.split([x[0] for x in events_sort],
                                np.where(np.diff([x[-1] for x in events_sort]) == 1)[0]+1)

    # Create templateSet
    templateSet = []
    # Get a representative event out of each cluster
    for clusterInd in range(max(clusters)):
        # e.g., take the first event
        evtInd = 0
        # Populate the templateSet
        templateSet.append([events_clustered[clusterInd][evtInd][0][1],   # Time
                            3.0,                                          # Duration
                            events_clustered[clusterInd][evtInd][0][2]])  # Magntitude

    # Sort + save
    templateSet = sorted(templateSet)
    io.writeTemplateSet(config.directory, templateSet, verbose=0)

    return templateSet


def get_template_distances(catalog, templateSet, coordCenter):
    """
    Returns the distances of the templates to the coordinate center
    """

    # Identify which catalog events are templates (and get their magnitudes/locations)
    templateSet_loc = [event for template in templateSet
                       for event in catalog if event[1] == template[0]]

    # Determine template distances
    templateSet_dist = [math.sqrt((ev[2]-coordCenter[0])**2 +
                                  (ev[3]-coordCenter[1])**2 +
                                  (ev[4]-coordCenter[2])**2) for ev in templateSet_loc]

    return templateSet_dist


def limit_numpy_fft_cache(max_size_in_mb_per_cache=100):
    """
    NumPy's FFT implementation utilizes caches to speedup subsequent FFTs of
    the same size. This accumulates memory when run for various length FFTs
    as can readily happen in seismology.

    This utility function clears both, full and read-only caches if their
    size is above the given threshold.

    The default 100 MB is fairly generous but we still want to profit from
    the cache where applicable.
    """
    for cache in ["_fft_cache", "_real_fft_cache"]:
        # Guard against different numpy versions just to be safe.
        if not hasattr(np.fft.fftpack, cache):
            continue
        cache = getattr(np.fft.fftpack, cache)
        # Check type directly and don't use isinstance() as future numpy
        # versions might use some subclass or what not.
        if type(cache) is not dict:
            continue
        # Its a dictionary with list's of arrays as the values. Wrap in
        # try/except to guard against future numpy changes.
        try:
            total_size = sum([_j.nbytes for _i in cache.values() for _j in _i])
        except:
            continue
        if total_size > max_size_in_mb_per_cache * 1024 * 1024:
            cache.clear()
            print("Cleared numpy's FFT cache.")


def select_fastest_approach():
    """
    Given data chunk size, template duration, and sampling rate, determine
    the fastest convolution approach (either time or frequency domain)

    This operation is used for cross-correlation in main.do_xcorr_onechannel()
    """

    fs = config.resampling
    # Sizes in samples
    a_len = int(fs)*60*60*24  # data chunk size (here: 1 day)
    t_len = int(fs*config.duration) + 1  # template length

    print("Determining fastest convolution/cross-correlation method for this case.")
    print("Benchmarking...")
    convtime, fftconvtime = benchmark_computimes_conv(a_len, t_len, out='times')

    # Predict the computation time using machine-specific coefficients
    # (but more inaccurate / only an estimate)
#    convtime, fftconvtime = predict_computimes_conv(t, a)

    if convtime < fftconvtime:
        config.conv_method = 'convolve'
    else:
        config.conv_method = 'fftconvolve'

    print("%s() is fastest here." % config.conv_method)


def predict_computimes_conv(tr, tp):
    """
    Predict running times of convolution in time and frequency domain.
    Input:
        either the length of trace & template or the arrays itself
    Returns:
        *estimated* running times (per run)
        (!! only plausible if you ran the benchmark on THIS system do dereive the coefficients !!)
    """

    if not np.isscalar(tr) and not np.isscalar(tp):
        tr = tr.size
        tp = tp.size

    # Machine specific coefficients
    #  (approx. calculation time per operation);
    #  (they don't account for arry copying, overhead, etc.) - only an ESTIMATE!
    # Home machine
#    k_conv = 2.6709592572922964e-10  # std: 6.379228128623962e-11
#    k_fft = 9.2717687893665909e-09   # std: 6.9225722192937631e-10
    # Work machine
    k_conv = 2.9824722617224554e-10  # std: 4.9248580738007472e-11
    k_fftc = 7.4641259524719957e-09  # std: 1.4075530486498362e-09

    # Determine coefficients once
#    k_conv, k_fftc = benchmark(tr, tp, maxtime=60, out='coeffs'))

    # spatial convolution time scales with the total number of operations
    convtime = k_conv*(tr*tp)

    # Fourier convolution time scales with N*log(N) ... (N: total number of elments)
    outdims = tr + tp
    fftconvtime = k_fftc*(np.prod(outdims) * np.log(np.prod(outdims)))

    print("Predicted running time 'conv':  %.6e\n\t\t\t\t\t   'fftconv':  %.6e" %
          (convtime, fftconvtime))

    return convtime, fftconvtime


def benchmark_computimes_conv(tr, tp, maxtime=60, out='times'):
    """
    Benchmarks convolution in time and frequency domain.
    Input:
        either the length of trace & template or the arrays itself
    Returns:
        both running times (normed by number of runs)
        and their coefficients (approx. calculation time per operation)
    """

    if np.isscalar(tr) and np.isscalar(tp):
        # Create random arrays
        tr = np.random.randn(tr)
        tp = np.random.randn(tp)

    print('  [np.convolve()] :: calculating %d x %d  for min. %ds' % (len(tr), len(tp), maxtime))
    convreps = 0
    tic = time.time()
    toc = time.time()
    while (toc-tic) < maxtime:
        np.convolve(tr, tp, mode='valid')
        convreps += 1
        toc = time.time()
    convtime = (toc-tic)/convreps
    print("  Mean running time: %.3fs" % convtime)

    print('  [fftconvolve()] :: calculating %d x %d for min. %ds' % (len(tr), len(tp), maxtime))
    reps = 0
    tic = time.time()
    toc = time.time()
    while (toc-tic) < maxtime:
        scipy.signal.fftconvolve(tr, tp, mode='valid')
        toc = time.time()
        reps += 1
    fftconvtime = (toc-tic)/reps
    print("  Mean running time: %.3fs" % fftconvtime)

    if out == 'times':
#        print("Runningtime conv vs. fftconv:  %.3fs vs. %.3fs" % (convtime, fftconvtime))
        return convtime, fftconvtime

    elif out == 'coeffs':
        N = len(tr) * len(tp)
        k_conv = convtime/N
        N = len(tr) + len(tp)
        k_fft = fftconvtime/(N*np.log(N))

        print('k_conv: %e; k_fft: %e' % (k_conv, k_fft))
        return k_conv, k_fft
