# -*- coding: utf-8 -*-
"""
Submodule for operations / processings regarding waveforms (streams and traces).

(created) 2015, Marcus Herrmann, Swiss Seismological Service, ETH Zürich
"""

# --- Import modules

# General
from multiprocessing import Pool

# Operands
from itertools import repeat
from future.utils import native_str

# Scientific
import numpy as np
import scipy
import scipy.fftpack
import scipy.signal  # To avoid scipy library module api problems
from scipy.signal import get_window
try:
    from scipy.fftpack.helper import next_fast_len  # scipy >= v1.18
except ImportError:
    from scipy.signal.signaltools import _next_regular as next_fast_len
import matplotlib.pyplot as plt
    
# Monkey patch a 3rd party library
# Here: pyFFTW - a pythonic wrapper around FFTW 3, the speedy FFT library
#   https://github.com/pyFFTW/pyFFTW
# Install python3-pyfftw (with apt-get or pip3)
# Requires: libfftw3-dev, cython, and maybe python3-dev
#import pyfftw
#import scipy
#import scipy.fftpack
#scipy.fftpack.rfft = pyfftw.interfaces.scipy_fftpack.rfft
#scipy.fftpack.irfft = pyfftw.interfaces.scipy_fftpack.irfft
#pyfftw.interfaces.cache.enable()  # Turn on the cache for optimum performance

# ObsPy
from obspy.core import Trace, Stream

# Own modules
from TM.setupclass import config


def _process_mask(mask, factor, extend_gap=0):
    """
    Process an array mask of a trace (required when decimating/resampling related data).

    TODO: implement UPsampling
    TODO: beautify with np.ma.clump_unmasked, but requires the mask ON DATA
            --> np.ma.clump_unmasked(np.ma.array(mask, mask=mask))

    """

    if not float(factor).is_integer() or factor < 1:
        raise Warning("Sorry, this mask resampling factor is not supported")

    if extend_gap:

        extend_gap = int(extend_gap*factor)  # [in samples of original waveform]

#        # Find the False->True / True->False indices
#        mask_change = np.where(np.diff(mask))[0]+1

        masked_clumps = np.ma.clump_masked(np.ma.masked_array(mask, mask=mask))

#        if mask_change.any():
#            if mask[0] is np.bool_(False):  # If the first value is False (mask off) - most likely
#                mask_changeFT = mask_change[::2]   # A change from False to True (mask on)
#                mask_changeTF = mask_change[1::2]  # A change from True to False (mask off)
#                if len(mask_changeFT) > len(mask_changeTF):  # If the mask reaches until the end
#                    mask_changeTF = np.append(mask_changeTF, len(mask))
#            else:
#                mask_changeFT = np.append([0], mask_change[1::2])  # Prepending 0
#                mask_changeTF = np.append(mask_change[::2], len(mask))  # Appending last element

        if masked_clumps:

            # Extend each consecutive mask by specified length
            sls = []
#            for idx, _ in enumerate(mask_changeFT):
#
#                start = mask_changeFT[idx]
#                end = mask_changeTF[idx]

            for masked_clump in masked_clumps:

                start = masked_clump.start
                end = masked_clump.stop
                # Check if we can actually extend the gap
                if start >= extend_gap:
                    start -= extend_gap
                if end <= len(mask)-extend_gap:
                    end += extend_gap
                sls.append(slice(start, end))
                mask[start:end] = np.bool_(True)
        else:
            print("Whoops... a masked array without gaps? (o.o)")
            return mask  # Return original mask

#        # Extend also at beginning and end
#        # (b/c it's certainly rubbish: filter & resample zero pad windowing)
#        mask[0:extend_gap] = np.bool_(True)
#        mask[-extend_gap:] = np.bool_(True)

    # Decimate array mask
    mask = mask[::int(factor)]

    return mask


def separate_masked_trace(waveobject):
    """
    A stream previously merged with 'fill_value'=None will contain a masked array.
    This function separates each continous waveform clump (the valid data) into
    into a stream to facilitate further processing.

    Removes mask of array!
    """

    if isinstance(waveobject, Stream):
        if waveobject.count() == 1:
            waveobject = waveobject[0]
        else:
            Warning("Too much traces. Can process only one trace")
    elif not isinstance(waveobject, Trace):
        raise Warning("I need a single trace!")

    if not np.ma.isMaskedArray(waveobject.data):
        print("Data not masked!")
        return Stream(waveobject)

    else:
        unmasked_clump = np.ma.clump_unmasked(waveobject.data)
#        waveobject.data = np.array(waveobject.data, copy=False)  # Remove mask
        stats = waveobject.stats
        stream = Stream()

        for unmasked in unmasked_clump:
            start = stats.starttime + (unmasked.start) * stats.delta
            end = stats.starttime + (unmasked.stop-1) * stats.delta

            stream.append(waveobject.slice(start, end))

        return stream


def resample_stream_patched(self, sampling_rate, window='hanning', no_filter=True,
                            strict_length=False):
    """
    Patched resample() method of Stream() class.
    """

    for tr in self:
        tr.resample(sampling_rate, window=native_str(window),
                    no_filter=no_filter, strict_length=strict_length)
    return self


def resample_trace_patched(self, sampling_rate, window='hanning', no_filter=True,
                           strict_length=False):
    """
    Patched resample() method of Trace() class.

    If processing a trace with gaps (masked array), you should separate the clumps
    and process each individually with this function.

    This function will make sure that the borders are windowed to 'zero' withing one second.
    (thus loosing information at the borders)

    To make use of the speed-up using 5-smooth numbers in FFTPACK,
    the resampling factor should be factorizable by 2^i × 3^j × 5^k,
        e.g., 2, 3, 4, 5, 6, 8, 9, 10, etc. --> no 7, 11, etc.!!
    """
#    print('Hello from patched Trace.resample().')

    fft_package = 'numpy'  # ['numpy' or 'scipy']

    factor = self.stats.sampling_rate / float(sampling_rate)
    # check if end time changes and this is not explicitly allowed
    if strict_length:
        if len(self.data) % factor != 0.0:
            msg = "End time of trace would change and strict_length=True."
            raise ValueError(msg)

#    # Plot the "input" spectra (to check if the input data is "okay")
#    plt.figure()
#    flen = next_fast_len(self.stats.npts)
#    glen = 1024
#    xlen = self.stats.npts
#    f = np.fft.rfftfreq(glen, d=self.stats.delta)
#    plt.semilogy(f, np.abs(np.fft.rfft(self.data[:glen], glen).real), label='beginning')
#    plt.semilogy(f, np.abs(np.fft.rfft(self.data[xlen//2-glen//2:xlen//2+glen//2], glen).real), label='middle')
#    plt.semilogy(f, np.abs(np.fft.rfft(self.data[-glen:], glen).real), label='end')
##    f = np.fft.rfftfreq(flen, d=self.stats.delta)
##    plt.semilogy(f[::10000], np.abs(np.fft.rfft(self.data, flen).real)[::10000])
#    plt.semilogy(*scipy.signal.welch(self.data, self.stats.sampling_rate, nperseg=glen), label='all (welch)')
#    plt.title('spectra of input')
#    plt.legend(loc=(1.05,0))
#    plt.xlim([0, self.stats.sampling_rate//2])
    
    # Cast data to float32 (FFTPACK works with float32 if using a length of a 5-smooth number)
    self.data = np.asarray(self.data, dtype=np.float32)  # casts + remove masks

    # Taper input data at borders (b/c we zero-pad to speed up FFT)
    wlen = int(self.stats.sampling_rate)  # 1s
    xwindow = get_window(native_str('hanning'), 2*wlen-1)
    self.data[:wlen] *= xwindow[:wlen]
    self.data[-wlen:] *= xwindow[-wlen:]

    # Speed up FFT by padding to optimal size for FFTPACK
    Nx_padded = next_fast_len(self.stats.npts)  # a 5-smooth number
    
    # Output length: immitate array[::facor] length determination
    num = int(np.ceil(self.stats.npts / factor))
#    num = int(self.stats.npts//factor)  # "intuition"

    # --- Transform into frequency domain (making sure the byteorder is native)
    if fft_package == 'scipy':  # Option 1: scipy rfft

        # Allocate contiguous array (to avoid copying whole array for changing 2 values)
        X = np.zeros(Nx_padded+2, dtype=np.float32)

        # Nx_padded is even, thus scipy's fft outputs:
        # scipy outputs X_r1, X_r2, X_i2, X_r3, X_r3, ..., X_r_n-1, X_i_n-1, X_r_n
        #    (thus leaving out     ^ X_i1,           and                    ^ X_in,
        #     which are both 0)

        # Fill into view of X
        X[1:-1] = scipy.fftpack.rfft(self.data.newbyteorder("="), Nx_padded)
        # (now, a '0' is at the beginning and at the end)

        # "Correct" for scipy's scheme to separate Real & Imag (X_r, X_i)
        X[0] = X[1]
        X[1] = X.dtype.type(0)  # Same dtype as the array to avoid issues with numpy rule 'safe'
        # Now X looks like this:
        #      idx |  0    1   2    3   ...  n-3  n-2  n-1 |   (n = Nx_padded)
        #       X  | X_r,  0, X_r, X_i, ..., X_i, X_r,  0  |

        # Separate real and imaginary parts
        X_r = X[::2]
        X_i = X[1::2]

    else:  # Option 2: numpy rfft (outputs complex128)

        X =  np.fft.rfft(self.data.newbyteorder("="), Nx_padded)

    # --- Resample
    if factor > 1:      # If downsampling

        # Create a narrow window at the edge of nyquist (0.2 tapered cosine)
        W = np.fft.ifftshift(get_window((native_str('tukey'), 0.2), num))
#        # Plot the window b/w up to Nyquist freq.
#        plt.plot(np.fft.ifftshift(get_window(('tukey', 0.2), sampling_rate)))
#        plt.plot(np.fft.ifftshift(get_window('hann', sampling_rate)))
#        plt.xlim([0, int(sampling_rate//2+1)])

        # Only use freqs up to new nyquist - ignore/drop higher freqs
        nqy = num//2

    else:  # If upsampling

        if window is not None:
            if callable(window):
                W = window(np.fft.fftfreq(self.stats.npts))
            elif isinstance(window, np.ndarray):
                if window.shape != (self.stats.npts,):
                    msg = "Window has the wrong shape. Window length must " + \
                          "equal the number of points."
                    raise ValueError(msg)
                W = window
            else:
                W = np.fft.ifftshift(get_window(native_str(window), self.stats.npts))

        # Do simple slicing up to old nyquist (rest is zeros)
        #  (and only apply window to non-zero-padded data)
        nqy = self.stats.npts//2

    # IFFT size
    num_padded = next_fast_len(num)  # Speed up IFFT

    # --- Transform back into time domain (IFFT)
    if fft_package == 'scipy':  # Option 1: scipy irfft

        # Init output array (+2 for scipy scheme correction)
        Y = np.zeros(num_padded+2, dtype=np.float32)

        # Write windowed FFT into output array
        Y[:2*nqy:2] = X_r[:nqy] * W[:nqy]
        Y[1:2*nqy:2] = X_i[:nqy] * W[:nqy]

        # Bring into scipy's fft format (for even lengths)
        Y[1] = Y[0]  # "rescue" the first value
        Y = Y[1:-1]

        y = scipy.fftpack.irfft(Y)

    else:  # Option 2: numpy irfft

        Y = np.zeros(num_padded//2, dtype=np.complex64)
        # (even faster: init np.empty() and later overwrite Y[nqy:] with 0)

        Y.real[:nqy] = X.real[:nqy] * W[:nqy]  # (now as complex64 dtype)
        Y.imag[:nqy] = X.imag[:nqy] * W[:nqy]

        y = np.fft.irfft(Y, num_padded)  # Enforce optimal FFT length

#    # Debug
#    plt.figure()
#    plt.plot(Y.real[:21])
#    plt.plot(Y.imag[:21])
#    plt.figure()
#    plt.plot(Y.real[nqy-10:nqy+11])
#    plt.plot(Y.imag[nqy-10:nqy+11])
#
#    print("npts: %d  Nx_padded: %d  num: %d  num_padded: %d"
#          % (self.stats.npts, Nx_padded, num, num_padded))

    # Correct amplitudes
    self.data = y[slice(0, num)] * (float(num_padded) / float(Nx_padded))
    self.stats.sampling_rate = sampling_rate

#    # Plot the spectra
#    plt.figure()
#    flen = 1024
#    xlen = self.stats.npts
#    f = np.fft.rfftfreq(flen, d=self.stats.delta)
#    plt.semilogy(f, np.abs(np.fft.rfft(self.data[:flen], flen).real), label='beginning')
#    plt.semilogy(f, np.abs(np.fft.rfft(self.data[xlen//2-flen//2:xlen//2+flen//2], flen).real), label='middle')
#    plt.semilogy(f, np.abs(np.fft.rfft(self.data[-flen:], flen).real), label='end')
#    plt.semilogy(*scipy.signal.welch(self.data, sampling_rate, nperseg=flen), label='all (welch)')
#    plt.title('spectra of output; %s sample windows' % flen)
#    plt.legend(loc=(1.05,0))
#    plt.xlim([0, sampling_rate//2])

    return self


# Patch resample()
Stream.resample = resample_stream_patched
Trace.resample = resample_trace_patched


def _waveform_processing_trace(trace, resampling, procType='bandpass', freqMax=None):
    """
    Private function to process a single trace.
    """

    duration = trace.stats.npts * trace.stats.delta

    # Save mask (masks originate from gaps - they can be different among the channels)
    # Not handled by obspy yet: maintaining an array mask during processing
    if np.ma.isMaskedArray(trace.data):
        mask = trace.data.mask
        # Remove mask to avoid obspy raising 'NotImplementedError'
        trace.data = np.asarray(trace.data)
    else:
        mask = []  # create an empty element

    # Detrend to avoid filter responses (but for useful lengths)
    if 1.5 <= duration <= 100:
        trace.detrend("linear")

    # We filter first, and then re/downsample;
    # This costs more performance (more to filter), but the data will be represented "better"
    # It also makes sure that impulses become "periodic" - important for FFT-resampling
    if procType == 'bandpass':
        trace.filter("bandpass", freqmin=config.filtConf[0],
                     freqmax=config.filtConf[1], corners=config.filtConf[2])

    elif procType == 'highpass':
        trace.filter("highpass", freq=1.0)
        # If maximum freq specified, apply also a low-pass filter
        #   (to remove too-high frequencies that disturb the amplitude comparison)
        if freqMax:
            # Determine lowest senseful frequency
            # (at max, only up to Nyquist of original signal, or
            #  if signal will be resampled below/equal Nyquist freq, take the new sampling rate)
            freqmax = min(resampling if resampling else float('Inf'),   # New sampling freq
                          trace.stats.sampling_rate//2,          # Nyguist freq
                          freqMax)                                      # Specified max. freq
            trace.filter("lowpass", freq=freqmax)

    # Apply 50Hz bandSTOP filter if configured
    if config.stop50Hz:
        # ObsPy Standard
        trace.filter("bandstop", freqmin=46, freqmax=54, corners=2, zerophase=True)

    # FYI: if filtered, trace.data.dtype = np.float64

    if (resampling and resampling < trace.stats.sampling_rate):

        if (duration > 3600) and trace.stats.channel == config.channels[0]:
            print("Resampling to %d Hz..." % resampling, flush=True)
#            print("Decimating to %d Hz..." % resampling, flush=True)

        factor = trace.stats.sampling_rate/resampling
        # Note: only INTEGER DOWNsampling supported
        if not float(factor).is_integer() or factor < 1:
            # TODO: make array masking robust if factor is not a natural number (int)
            raise Warning("Sorry, this resampling factor will not be compatible with "
                          "masked data (gaps). Please use an integer factor > 1"
                          " (e.g. %d -> %dHz = %d" % (trace.stats.sampling_rate,
                                                      trace.stats.sampling_rate//2, 2))

        if np.ma.is_mask(mask):
            # Apply mask again (was removed before)
            trace.data = np.ma.masked_array(trace.data, mask=mask)  # (does not copy)
            # Separate each masked clump into an individual (no-gap) trace
            trace = separate_masked_trace(trace)  # is now a stream (without array masks)

            # Raise Warning if too many gaps
            if len(trace) > 50:
                raise Warning("Too many gaps; chunk lengths may not agree among channels.")
                # TODO: Possible solution: triming each sub-trace to indiv. common start/end
                #       Test-case: MATTE 2016-06-29
                #       But where to implement? in main()?

            # Remove traces that are too small (<= 2 * template duration)
            for subtrace in trace:
                if subtrace.stats.npts <= (2 * config.duration) * \
                                          subtrace.stats.sampling_rate:
                    trace.remove(subtrace)

        trace.resample(resampling)  # Using patched resample() (see above)
#        trace.decimate(int(factor))  # (deprecated due to patched .resample())

        if np.ma.is_mask(mask):

            # Bring everything back together as trace (non-masked)
            # Note: '0' is faster than 'None' - we have the mask anyway...
            trace = trace.merge(fill_value=0)[0]

            # Create a "resampled" mask
#            mask = trace.copy().merge(fill_value=None)[0].data.mask
            # ... more intelligently (and faster) without obspy
#            extend_gap = trace.stats.sampling_rate  # 1s [in samples of resampled waveform]
            extend_gap = 0  # We will extend it later (after xCorr) by half the template length
            mask = _process_mask(mask, factor, extend_gap)

            # Make data and mask the same length (because .merge() may lead
            # to changed re-adjusted timing of samples due to gaps)
            len_common = min(trace.stats.npts, len(mask))
            trace.data = trace.data[:len_common]
            mask = mask[:len_common]

    else:
        factor = 1  # used for "resampling" the masks (if present)

    # FYI: trace.data is not masked now

    # To save memory, cast data to single precision (float32) ... standard is float64
    # This is because all signal processing (filter/decimate/resample) results in float64
    # Note: If data was resampled after filtering, dtype is already np.float32
    trace.data = trace.data.astype(np.float32, copy=False)
    trace.stats.mseed.encoding = 'FLOAT32'  # update stats (b/c obspy wouldn't notice)

    if np.ma.is_mask(mask):
        # Propagate/re-apply (possibly processed) array mask (if existing)
        trace.data = np.ma.masked_array(trace.data, mask=mask)  # (does not copy)

    return trace


def waveform_processing(waveobject, procType='bandpass', resampling=-1,
                        freqMax=None, copy=True, trimFilterMargin=True):
    """
    Returns a processed waveform. It can be a Trace or a Stream and returns the same type.
    Will cut 1 second at the beginning and the end of the waveform to remove filter response.

    procType:
        'bandpass'
        'highpass'
        'raw'     --> no processing at all (also ignoring 'resampling') No trimming.
        any other --> no *filter* processing, but still 'resampling', detrending, trimming.

    resampling:
        -1        --> will use default resampling (from config.resampling)
        None/0    --> will NOT resample
        any other --> will resample to that frequency

    i.e., procType='standard' and resampling='None'
              will cut away the waveform margin, but not resample

    trimFilterMargin: Will trim waveform at start and beginning to remove filter response
        (certain filter response still possible despite detrending)
    """

    if procType == 'raw':
        if resampling != -1:
            print("'resampling' specified, but ignored. Returning RAW waveform.")
        return waveobject
#    elif 'pass' not in procType:
#        trimFilterMargin = False  # No trimming if no filter applied

    # To not modify the original waveform (i.e. make a "pure" function)
    if copy:
        waveobject = waveobject.copy()

    # Default value for resampling if nothing is set (but not if 'None')
    if resampling == -1:
        resampling = config.resampling

    if isinstance(waveobject, Trace):

        waveobject = _waveform_processing_trace(waveobject, resampling, procType, freqMax)
        stats = waveobject.stats

    else:  # if Stream

        # TODO: find out why I can't overwrite a trace in a stream loop
        waveobject_new = Stream()

        # TODO: parallize?
        for tr in waveobject:
            waveobject_new.append(_waveform_processing_trace(tr, resampling, procType, freqMax))
            if copy:
                waveobject.remove(tr)  # To save memory
        waveobject = waveobject_new
        stats = waveobject[0].stats

    # Trim the waveform (esp. to avoid filter response)
    if trimFilterMargin:

        duration = stats.npts * stats.delta
        if duration >= 100:
            addTime = 1.0  # Always trim daily waveforms
        else:
            addTime = config.filterBuff  # [s]

        # Triming sustains array mask
        waveobject.trim(stats.starttime + addTime,
                        stats.endtime - addTime)

    return waveobject


def waveform_processing_parallel(waveform, procType='bandpass', resampling=-1,
                                 freqMax=None):
    """
    Waveform filtering - making use of multiple cores (=parallel filtering) if desired

    Only makes sense if you need a COPY of the stream
    (since muliprocessing using processes copies the memory to its processes - no sharing)
    """

    if isinstance(waveform, Trace) or config.parallelFilter is not True:
        # Use the standalone version (and thus, tell to copy the object)
        return waveform_processing(waveform, procType, resampling, freqMax, copy=True)

    nTraces = waveform.count()

#    with ThreadPool(nTraces) as pool:  # Uses threads (shared data) - no improvement/worser
    with Pool(nTraces) as pool:  # Uses processes (copies data)
        waveformFiltered = pool.starmap(
            waveform_processing, zip(waveform, repeat(procType), repeat(resampling),
                                     repeat(freqMax), repeat(False)))

    return Stream(waveformFiltered)


# --- Utils


def get_max_amp(stream, ret='amp'):
    """
    Determine the maximum amplitude among all channels and ...

    if ret == 'amp': return it

    if ret == 'time': return time of maximum *relative* to start of stream [s]

    if ret == 'both': return both
    """

    # Patch (trim to first part - in case of a larger successive event)
#    stream = stream.slice(stream[0].stats.starttime, stream[0].stats.starttime+3)

    # In case of small EQ, filter out disturbing long-period waves
#    stream = stream.copy().detrend('simple').filter('highpass', freq=1)

    for chnlInd in range(stream.count()):
        # Build up the sum of the squared waveform
        if chnlInd == 0:
            waveformPow2 = stream[0].data**2
        else:
            waveformPow2 += stream[chnlInd].data**2

    waveformTotalAmp = waveformPow2**0.5

    max_amp = max(waveformTotalAmp)
    timeshift = np.argmax(waveformTotalAmp) * stream[0].stats.delta  # [s]

    if ret == 'amp':
        return max_amp

    elif ret == 'time':
        return timeshift

    else:
        return max_amp, timeshift


def getTimeShiftOfMaxSimilarity(stream1, stream2):
    """
    Given two streams, finds the time shift so that stream2 has maximum similarity with stream1.
    In other words: the time shift on stream2 aligns it with stream1.
    """

    if not stream1.count() == stream2.count():
        raise Exception("Streams don't have the same number of channels.")

    # Make comparison less dependent on the high frequencies
    stream1 = stream1.copy().filter('bandpass', freqmin=5, freqmax=50)
    stream2 = stream2.copy().filter('bandpass', freqmin=5, freqmax=50)

    # Make sure both have the same sampling rate
    if stream2[0].stats.sampling_rate < stream1[0].stats.sampling_rate:
        stream1.resampling(stream2[0].sampling_rate)
    elif stream1[0].stats.sampling_rate < stream2[0].stats.sampling_rate:
        stream2.resampling(stream1[0].sampling_rate)

    for chnlInd in range(stream1.count()):
        # Do cross-correlation
        xCorr_comp = scipy.signal.fftconvolve(stream1[chnlInd].data,
                                              stream2[chnlInd].data[::-1],
                                              mode='same')
        if chnlInd == 0:
            xCorr = xCorr_comp
        else:
            xCorr += xCorr_comp  # Stacking cross-correlation traces

    # Calculate shift of stream2 relative to stream1 [in samples]
    sampleShift = -(xCorr.argmax() - stream2[0].count()//2)

    # DEBUG
#    stream1.copy().filter('bandpass', freqmin=1, freqmax=50).plot(
#        outfile=directory + "/DEBUG_" + str(stream1[0].stats.starttime).replace(":", "-") + ".png")
#    stream2.copy().filter('bandpass', freqmin=1, freqmax=50).plot(
#        outfile=directory + "/DEBUG_" + str(stream2[0].stats.starttime).replace(":", "-") + ".png")

    # Return relative time shift
    return sampleShift * stream2[0].stats.delta
