from tm_master_slave import TMMaster, Tasks
from chunk_manager import ChunkManager, Chunks
from template_manager import TemplateManager
from scanned_days_manager import ScannedDaysManager
import TM.io as tmio
import TM.main as tmmain
import time
import os

class TemplateMatching(object):
    """
    Main template matching application logic

    The station waveform data is "divided" (virtually) in chunks (1 hour, 1 day,
    whatever + a small overlapping between chunks to not miss events that cross
    chunk boundaries) and we will sequentially read each chunk from the archive
    and store it in the filessystem so that another process can xcorr that chunk
    This class keeps track of what chunk has already been loaded and processed
    by which template
    We probably cannot store all the station waveform data in the filesystem,
    so there will be a rolling window of available chunks
    As new templates might become available during the processing, a second pass
    scan of the chuncks might be required at the end of processing.
    """

    def __init__(self, config, template_set, catalog, slaves):
        self.config         = config
        self.chunks         = Chunks(start_date=config.startdate,
                                     end_date=config.enddate,
                                     chunk_len=config.chunk_len,
                                     overlap_len=config.overlap_len)
        self.chunk_mngr     = ChunkManager(self.chunks, window_len=config.chunk_on_disk)
        self.template_mngr  = TemplateManager(template_set)
        self.scanner_mngr   = ScannedDaysManager()
        self.catalog        = catalog
        self.num_readers    = self.config.num_readers
        self.m_readers      = TMMaster(task=Tasks.LOAD_CHUNK,  tm_slaves=slaves[:self.num_readers])
        self.m_processors   = TMMaster(task=Tasks.XCORR_CHUNK, tm_slaves=slaves[self.num_readers:])
        self.m_template_ldr = TMMaster(task=Tasks.LOAD_TEMPLATE)

    def terminate_slaves(self):
        self.m_processors.terminate()
        self.m_readers.terminate()
        self.m_template_ldr.terminate()

    def run(self):

        # status information updated while running
        self.first_scan       = True
        self.events           = []
        self.failed_chunks    = []
        self.failed_templates = []

        # Check if this run was active before
        # TODO: requires more checks and deals from my old main.py (template_matching())
        if not os.path.exists(self.config.directory):
            os.makedirs(self.config.directory + "/templates_mseed",
                        exist_ok=True)
        #
        # this is the main loop,we never exit until all data has been processed
        #
        while not self.template_mngr.done() or \
              not self.chunk_mngr.done()    or \
              not self.scanner_mngr.done():

            # Reclaim any slave who completed his task and check results
            self.__reclaim_slaves()

            # Load new templates (if any new one to load)
            self.__load_templates()

            # Load next chunks  (if any new one to load)
            self.__load_chunks()

            # Process (xCorr) newly loaded chunks and templates
            self.__process_chunks()

            time.sleep(0.2)


    def __reclaim_slaves(self):
        """
        Reclaim any slave who completed his task and reuse it for new work
        """

        # reclaim template loaders and mark new templates ready to be processed
        for l in self.m_template_ldr.get_completed():
            template, error = self.m_template_ldr.get_data(l)
            if error:
                self.template_mngr.set_unloadable(template)
                self.failed_templates.append(template)
                print('(Error) Loading template %d' % (template) )
            else:
                self.template_mngr.set_loaded(template)
                self.scanner_mngr.add_template_to_process(template)
                print('(done) Loading template %d' % (template) )

        # reclaim readers and mark new chunks ready to be processed
        for r in self.m_readers.get_completed():
            chunk, error = self.m_readers.get_data(r)
            if error:
                self.chunk_mngr.set_unloadable(chunk)
                self.failed_chunks.append(chunk)
                print('(Error) Reading chunk %d' % (chunk) )
            else:
                self.chunk_mngr.set_loaded(chunk)
                self.scanner_mngr.add_chunk_to_process(chunk)
                print('(done) Reading chunk %d' % (chunk) )

        # reclaim processors and mark chunk/template pair as processed
        # and possibly add new templates
        for p in self.m_processors.get_completed():
            chunk, template, xcorr_results = self.m_processors.get_data(p)
            self.scanner_mngr.set_processed(chunk, template)
            if xcorr_results is not None:
                self.events.extend(xcorr_results)
            print('(done) Processing chunk %d template %d - events %d' %
                  (chunk, template, (len(xcorr_results)if xcorr_results else 0) ))

            # Possibly add a new template when the chunk has been xcorred with
            # all templates and still an UNDETECTED event from catalog is found
            if self.first_scan and self.scanner_mngr.done([chunk]):
                self.__detect_new_templates(chunk)

            # Delete from filesystem oldest chunk that have been already processed
            # against all templates.
            self.__delete_oldest_processed_chunk()

    def __delete_oldest_processed_chunk(self):
        """
        Delete from filesystem oldest chunk that have been already processed
        against all templates.
        """

        # if we alreday have windows_len chunks loaded it might be time to
        # unload the oldest chunk
        if self.chunk_mngr.window_full():

            # check the oldest_chunk has been xcorred with all templates
            oldest_chunk = self.chunk_mngr.loaded[0]
            if self.scanner_mngr.done([oldest_chunk]):

                self.chunk_mngr.remove(oldest_chunk)
                tmio.delete_chunk(oldest_chunk)
                print('Deleted chunk %d from file system' % oldest_chunk)

    def __detect_new_templates(self, chunk_id):
        """
        Check if the template set can be updated with the an UNDETECTED event
        from catalog
        """

        starttime, endtime, length = self.chunks.get(chunk_id)

        highCCC = []
        for e in self.events:
            time, amp, ccc = e
            if ccc >= self.config.threshNewTemplate and \
               time >= starttime and time < endtime:
                highCCC.append(e)

        if not highCCC:
            return

        missed = tmmain.missingEvents(highCCC, self.catalog, starttime, endtime)
        if missed:
            new_template = (missed[0][0], 0.0, self.config.duration, missed[0][1])
            #FIXME: check the template is not already present
            self.template_mngr.add_template(new_template)
            print('New template found  (%s)' % new_template)

    def __load_templates(self):
        """
        Load new templates (if any new one to load)
        """
        while self.template_mngr.is_next_available():
            # we don't keep slaves reserved for template loading (they would
            # be idle most of the time). So if there are some available we'll
            # use them otherwise we'll borrow processors' slaves (and give them
            # back later)
            ldrs = self.m_template_ldr.get_avaliable()
            if ldrs:
                l = next(iter(ldrs))
            else:
                l = self.m_processors.move_slave(to_master=self.m_template_ldr)
                if l is not None: print('Borrowing template loader %d from processors' % (l) )
            if l is None:
                break

            template = self.template_mngr.get_next()
            print('Loading template %d' % (template) )

            start, shift, length, mag = self.template_mngr.get_template(template)
            self.m_template_ldr.run(l, data=(template, start, length, mag) )

        # return borrowed template loader slaves when no work to do
        if self.template_mngr.done():
            for l in self.m_template_ldr.get_avaliable():
                print('Giving template loader %d back to processors' % (l) )
                self.m_template_ldr.move_slave(to_master=self.m_processors, slave=l)

    def __load_chunks(self):
        """
        Load next chunks  (if any new one to load)
        """

        # make sure we have reader slaves (if any chunk to load)
        if self.chunk_mngr.is_next_available() and \
           self.m_readers.num_slaves() < self.num_readers:
            r = self.m_processors.move_slave(to_master=self.m_readers)
            if r is not None: print('Taking reader %d  back from processors' % (r) )

        # load chunks
        for r in self.m_readers.get_avaliable():
            chunk = self.chunk_mngr.get_next()
            if chunk is None:
                # lend the reader slave to processors when no more work to do
                # to avoid idle slaves
                print('Giving reader %d to processors' % (r) )
                self.m_readers.move_slave(to_master=self.m_processors, slave=r)
                break
            start, end, length = self.chunks.get(chunk)
            print('Reading chunk %d (start %s end %s length %s)' % (chunk, start, end, length))
            self.m_readers.run(r, data=(chunk, start, length) )

    def __process_chunks(self):
        """
        Process (xCorr) new chunks
        """
        for p in self.m_processors.get_avaliable():
            chunk, template = self.scanner_mngr.next_to_process(self.chunk_mngr.loaded)
            if chunk is None or template is None:
                # If no more chunks to process it could be that we have to start
                # the second pass scanning (otherwise it simply means we have to
                # wait for more chunks to become available/loaded)
                self.__second_pass()
                break
            print('Processing chunk %d template %d' % (chunk, template) )
            self.m_processors.run(p, (chunk, template))

    def __second_pass(self):
        """
        Check if all the chunks have been processed with all templates and
        if so set up the second pass scan
        """

        # check all chunks have been loaded at least once
        if not self.chunk_mngr.done():
            return False

        # check there is still some chunk to be processed
        if self.scanner_mngr.done():
            return False

        # check all currently loaded chunks have been processed
        if not self.scanner_mngr.done(self.chunk_mngr.loaded):
            return False

        # check the chunks that still need to be processed
        missing = sorted( self.scanner_mngr.to_process.keys() )
        if not missing:
            return False
        #
        # we need to start a second pass scan: let's load all the chunks
        # that still need to be processed
        #
        for chunk in missing:
            self.chunk_mngr.add_to_load(missing)

        self.first_scan = False
        return True
