#!python
"""
Plot filter response and the effect of down-sampling

main source: https://github.com/obspy/obspy/issues/318
"""

import numpy as np

from scipy.signal import iirfilter, freqz

import matplotlib.pyplot as plt

from obspy.signal.filter import lowpass_cheby_2


sampling_rate = 1000
decimation_factor = 5

# calculate
new_sample_rate = sampling_rate / decimation_factor
nyquist = sampling_rate * 0.5
new_nyquist = nyquist / decimation_factor

# calculate b, a for filters

# ChebyII as used by Obspy as anti-alias / lowpass when resampling
corners_cheby = 12
#[b_cheby, a_cheby] = lowpass_cheby_2([], new_nyquist, sampling_rate, corners_cheby, ba=True)
[b_cheby, a_cheby] = iirfilter(corners_cheby, new_nyquist/nyquist, rp=1, rs=96, btype='lowpass', ftype='cheby2', output='ba')

# Default bandpass used by obspy
corners_butter = 4
bandpass = np.asarray([5, 30])
[b_BP, a_BP] = iirfilter(corners_butter, bandpass/nyquist, btype='bandpass', ftype='butter', output='ba')

# Another one
corners = 12
[b, a] = iirfilter(corners, new_nyquist/nyquist, rp=1, rs=96, btype='lowpass', ftype='ellip', output='ba')


# calculate frequency response
w, h = freqz(b, a, int(nyquist))
_, h_cheby = freqz(b_cheby, a_cheby, int(nyquist))
_, h_BP = freqz(b_BP, a_BP, int(nyquist))
freq = w/np.pi * nyquist

# Plot
fig = plt.figure(figsize=(8, 5), dpi=300)
plt.axvspan(0, new_sample_rate, color='grey', alpha=0.2, label='new signal')

plt.semilogy(freq, abs(h_cheby), label='Cheby2 LP %d-O' % corners_cheby)
plt.semilogy(freq, abs(h_BP), label='Butterw BP%d-%d %d-O' % (*bandpass, corners_butter))
#plt.semilogy(freq, abs(h), label='Ellip LP %d-O' % corners)

plt.fill_between(freq, np.minimum(abs(h_BP), abs(h_cheby)), 1e-8, where=freq<=new_sample_rate,
                 facecolor='gray', alpha=0.5, interpolate=True, label='new signal filtered')

plt.title('New Nyquist @ %3.1fHz (%3.1fsps)' % (new_nyquist, new_sample_rate))
plt.axvline(nyquist, color='r', label='old nyquist', lw=4)
plt.axvline(new_nyquist, color='k', label='new nyquist', lw=2)

plt.ylim([1e-7, 1])
plt.xlabel('Frequency [Hz]')
plt.ylabel('Amplitude')
plt.legend()

#plt.show()

plt.savefig("freq-response_resamp%dHz_BP%dHz.png" % (new_sample_rate, bandpass[1]),
            dpi=150, bbox_inches='tight')
